(*  Title:      Nitpick/Tools/nitpick_hol.ML
    Author:     Jasmin Blanchette, TU Muenchen
    Copyright   2008, 2009

Auxiliary HOL-related functions used by Nitpick.
*)

signature NITPICK_HOL =
sig
  type const_table = term list Symtab.table
  type special_fun = (styp * int list * term list) * styp
  type better_founded = styp * styp
  type wf_cache = (styp * bool) list

  datatype mood = Optimistic | Realistic | Pessimistic

  type extended_context = {
    thy: theory,
    ctxt: Proof.context,
    user_axioms: bool option,
    debug: bool,
    inductive_mood: mood,
    special_depth: int,
    skolem_depth: int,
    tac_timeout: Time.time option,
    eval: term list,
    def_table: const_table,
    nondef_table: const_table,
    user_nondefs: term list,
    simp_table: const_table ref,
    psimp_table: const_table,
    intro_table: const_table,
    skolems: (string * string list) list ref,
    special_funs: special_fun list ref,
    better_founded_preds: better_founded list ref,
    wfs: wf_cache ref}

  val numeral_prefix : string
  val skolem_prefix : string
  val eval_prefix : string
  val is_TFree : typ -> bool
  val is_higher_order_type : typ -> bool
  val is_fun_type : typ -> bool
  val is_set_type : typ -> bool
  val is_iterator_type : typ -> bool
  val is_boolean_type : typ -> bool
  val is_integer_type : typ -> bool
  val is_noninteger_number_type : typ -> bool
  val is_number_type : typ -> bool
  val is_record_type : typ -> bool
  val iterator_type_for_const : styp -> typ
  val const_for_iterator_type : typ -> styp
  val second_range_type : typ -> typ
  val third_range_type : typ -> typ
  val is_pure_typedef : theory -> typ -> bool
  val is_univ_typedef : theory -> typ -> bool
  val is_datatype_like : theory -> typ -> bool
  val is_record_ctor : styp -> bool
  val is_abs_fun : theory -> styp -> bool
  val is_rep_fun : theory -> styp -> bool
  val is_ctor_like : theory -> styp -> bool
  val instantiate_type : theory -> typ -> typ -> typ -> typ
  val eta_expand : typ list -> term -> int -> term
  val extensionalize : term -> term
  val distinctness_formula : typ -> term list -> term
  val datatype_ctors : theory -> typ -> styp list
  val precise_card_of_type : theory -> int -> (typ * int) list -> typ -> int
  val is_finite_type : theory -> typ -> bool
  val original_name : string -> string
  val user_friendly_const :
    term list -> (string * string list) list -> special_fun list -> styp -> term
  val all_axioms_of : theory -> term list * term list * term list
  val is_built_in_const : styp -> bool
  val const_def_table : Proof.context -> term list -> const_table
  val const_nondef_table : term list -> const_table
  val const_simp_table : Proof.context -> const_table
  val const_psimp_table : Proof.context -> const_table
  val inductive_intro_table : Proof.context -> const_table -> const_table
  val prefix_name : string -> string -> string
  val unprefix_name : string -> string
  val def_of_const : theory -> const_table -> styp -> term option
  val is_inductive_pred : extended_context -> styp -> bool
  val preprocess_term :
    extended_context -> term -> ((term list * term list) * (bool * bool)) * term
  val ground_types_in_type : theory -> typ -> typ list
  val ground_types_in_terms : theory -> term list -> typ list
end;

structure NitpickHOL : NITPICK_HOL =
struct

open NitpickUtil

type const_table = term list Symtab.table
type special_fun = (styp * int list * term list) * styp
type better_founded = styp * styp
type wf_cache = (styp * bool) list

datatype mood = Optimistic | Realistic | Pessimistic

type extended_context = {
  thy: theory,
  ctxt: Proof.context,
  user_axioms: bool option,
  debug: bool,
  inductive_mood: mood,
  special_depth: int,
  skolem_depth: int,
  tac_timeout: Time.time option,
  eval: term list,
  def_table: const_table,
  nondef_table: const_table,
  user_nondefs: term list,
  simp_table: const_table ref,
  psimp_table: const_table,
  intro_table: const_table,
  skolems: (string * string list) list ref,
  special_funs: special_fun list ref,
  better_founded_preds: better_founded list ref,
  wfs: wf_cache ref}

val numeral_prefix = nitpick_prefix ^ "num_"
val set_prefix = nitpick_prefix ^ "set_"
val iterator_prefix = nitpick_prefix ^ "it_"
val wf_prefix = nitpick_prefix ^ "wf_"
val skolem_prefix = nitpick_prefix ^ "sk"
val special_prefix = nitpick_prefix ^ "sp"
val eval_prefix = nitpick_prefix ^ "eval"

(* int -> string *)
fun skolem_prefix_for j = skolem_prefix ^ nat_subscript j ^ "_"
fun special_prefix_for j = special_prefix ^ nat_subscript j ^ "_"

val bound_var_prefix = "b"
val cong_var_prefix = "c"
val iter_var_prefix = "i"

val alternative_table =
  [(@{const_name fold_graph}, @{const_name fold_graph'}),
   (@{const_name zero_rat_inst.zero_rat}, @{const_name zero_frac}),
   (@{const_name one_rat_inst.one_rat}, @{const_name one_frac}),
   (@{const_name plus_rat_inst.plus_rat}, @{const_name plus_frac}),
   (@{const_name times_rat_inst.times_rat}, @{const_name times_frac}),
   (@{const_name uminus_rat_inst.uminus_rat}, @{const_name uminus_frac}),
   (@{const_name number_rat_inst.number_of_rat}, @{const_name number_of_frac}),
   (@{const_name inverse_rat_inst.inverse_rat}, @{const_name inverse_frac}),
   (@{const_name ord_rat_inst.less_eq_rat}, @{const_name less_eq_frac}),
   (@{const_name of_rat}, @{const_name of_frac}),
   (@{const_name Rats}, @{const_name UNIV}),
   (@{const_name zero_real_inst.zero_real}, @{const_name zero_frac}),
   (@{const_name one_real_inst.one_real}, @{const_name one_frac}),
   (@{const_name plus_real_inst.plus_real}, @{const_name plus_frac}),
   (@{const_name times_real_inst.times_real}, @{const_name times_frac}),
   (@{const_name uminus_real_inst.uminus_real}, @{const_name uminus_frac}),
   (@{const_name number_real_inst.number_of_real},
    @{const_name number_of_frac}),
   (@{const_name inverse_real_inst.inverse_real}, @{const_name inverse_frac}),
   (@{const_name ord_real_inst.less_eq_real}, @{const_name less_eq_frac})]

(* When you add constants to these lists, make sure to handle them in
   "NitpickTerm.from_term", and ideally in "NitpickMono.consider_term" as
   well. *)
val built_in_consts =
  [(@{const_name all}, 1),
   (@{const_name "=="}, 2),
   (@{const_name "==>"}, 2),
   (@{const_name Pure.conjunction}, 2),
   (@{const_name Trueprop}, 1),
   (@{const_name Not}, 1),
   (@{const_name False}, 0),
   (@{const_name True}, 0),
   (@{const_name All}, 1),
   (@{const_name Ex}, 1),
   (@{const_name "op ="}, 2),
   (@{const_name "op &"}, 2),
   (@{const_name "op |"}, 2),
   (@{const_name "op -->"}, 2),
   (@{const_name If}, 3),
   (@{const_name Let}, 2),
   (@{const_name Unity}, 0),
   (@{const_name Pair}, 2),
   (@{const_name fst}, 1),
   (@{const_name snd}, 1),
   (@{const_name Id}, 0),
   (@{const_name "op :"}, 2),
   (@{const_name insert}, 2),
   (@{const_name converse}, 1),
   (@{const_name trancl}, 1),
   (@{const_name rtrancl}, 1),
   (@{const_name Set.Un}, 2),
   (@{const_name Set.Int}, 2),
   (@{const_name rel_comp}, 2),
   (@{const_name image}, 2),
   (@{const_name Suc}, 0),
   (@{const_name finite}, 1),
   (@{const_name nat}, 0),
   (@{const_name zero_nat_inst.zero_nat}, 0),
   (@{const_name one_nat_inst.one_nat}, 0),
   (@{const_name plus_nat_inst.plus_nat}, 0),
   (@{const_name minus_nat_inst.minus_nat}, 0),
   (@{const_name times_nat_inst.times_nat}, 0),
   (@{const_name div_nat_inst.div_nat}, 0),
   (@{const_name div_nat_inst.mod_nat}, 0),
   (@{const_name ord_nat_inst.less_nat}, 2),
   (@{const_name ord_nat_inst.less_eq_nat}, 2),
   (@{const_name gcd}, 2),
   (@{const_name lcm}, 2),
   (@{const_name zero_int_inst.zero_int}, 0),
   (@{const_name one_int_inst.one_int}, 0),
   (@{const_name plus_int_inst.plus_int}, 0),
   (@{const_name minus_int_inst.minus_int}, 0),
   (@{const_name times_int_inst.times_int}, 0),
   (@{const_name div_int_inst.div_int}, 0),
   (@{const_name div_int_inst.mod_int}, 0),
   (@{const_name uminus_int_inst.uminus_int}, 0), (* ## needed? *)
   (@{const_name ord_int_inst.less_int}, 2),
   (@{const_name ord_int_inst.less_eq_int}, 2),
   (@{const_name Frac}, 0),
   (@{const_name norm_frac}, 0)]
val built_in_typed_consts =
  [((@{const_name of_nat}, nat_T --> int_T), 0)]
val built_in_set_consts =
  [(@{const_name minus_fun_inst.minus_fun}, 2),
   (@{const_name ord_fun_inst.less_eq_fun}, 2)]

(* typ -> bool *)
fun is_TFree (TFree _) = true
  | is_TFree _ = false
fun is_higher_order_type (Type ("fun", _)) = true
  | is_higher_order_type (Type (_, Ts)) = exists is_higher_order_type Ts
  | is_higher_order_type _ = false
fun is_fun_type (Type ("fun", _)) = true
  | is_fun_type _ = false
fun is_set_type (Type ("fun", [_, @{typ bool}])) = true
  | is_set_type _ = false
fun is_iterator_type (Type (s, _)) = String.isPrefix iterator_prefix s
  | is_iterator_type _ = false
val is_boolean_type = member (op =) [prop_T, bool_T]
val is_integer_type = member (op =) [nat_T, int_T] orf is_iterator_type
val is_noninteger_number_type = member (op =) [@{typ rat}, @{typ real}]
val is_number_type = is_integer_type orf is_noninteger_number_type
val is_record_type = not o null o RecordPackage.dest_recTs

(* styp -> typ *)
fun iterator_type_for_const (s, T) = Type (iterator_prefix ^ s, [domain_type T])
(* typ -> styp *)
fun const_for_iterator_type (Type (s, [T])) =
    (unprefix iterator_prefix s, T --> bool_T)
  | const_for_iterator_type T = raise TYPE ("const_for_iterator_type", [T], [])

(* typ -> typ *)
val second_range_type = range_type o range_type
val third_range_type = range_type o second_range_type

type typedef_info =
  {rep_type: typ, abs_type: typ, Rep_name: string, Abs_name: string,
   set_def: thm option, prop_of_Rep: thm, set_name: string}

(* theory -> string -> typedef_info *)
fun typedef_info thy @{type_name rat} =
    SOME {abs_type = @{typ rat}, rep_type = @{typ "int * int"},
          Abs_name = @{const_name Abs_Frac}, Rep_name = @{const_name Rep_Frac},
          set_def = NONE, prop_of_Rep = @{prop "Rep_Frac x \\<in> Frac"}
                          |> Logic.varify,
          set_name = @{const_name Frac}}
  | typedef_info thy @{type_name real} =
    SOME {abs_type = @{typ real}, rep_type = @{typ "int * int"},
          Abs_name = @{const_name Abs_Frac}, Rep_name = @{const_name Rep_Frac},
          set_def = NONE, prop_of_Rep = @{prop "Rep_Frac x \\<in> Frac"}
                                        |> Logic.varify,
          set_name = @{const_name Frac}}
  | typedef_info thy s =
    case TypedefPackage.get_info thy s of
      SOME {abs_type, rep_type, Abs_name, Rep_name, set_def, Rep, ...} =>
      SOME {abs_type = abs_type, rep_type = rep_type, Abs_name = Abs_name,
            Rep_name = Rep_name, set_def = set_def, prop_of_Rep = prop_of Rep,
            set_name = set_prefix ^ s}
    | NONE => NONE

(* theory -> string -> bool *)
val is_typedef = isSome oo typedef_info
val is_datatype = isSome oo DatatypePackage.get_datatype
(* theory -> typ -> bool *)
fun is_pure_typedef thy (T as Type (s, _)) =
    is_typedef thy s andalso not (is_datatype thy s)
    andalso not (is_record_type T) andalso not (is_integer_type T)
  | is_pure_typedef _ _ = false
fun is_univ_typedef thy (Type (s, _)) =
    (case typedef_info thy s of
       SOME {set_def, prop_of_Rep, ...} =>
       (case set_def of
          SOME thm =>
          try (fst o dest_Const o snd o Logic.dest_equals o prop_of) thm
        | NONE =>
          try (fst o dest_Const o snd o HOLogic.dest_mem
               o HOLogic.dest_Trueprop) prop_of_Rep)
       = SOME @{const_name UNIV}
     | NONE => false)
  | is_univ_typedef _ _ = false
fun is_datatype_like thy (T as Type (s, _)) =
    (is_datatype thy s orelse is_record_type T orelse is_pure_typedef thy T
     orelse T = @{typ ind})
    andalso T <> nat_T
  | is_datatype_like _ _ = false

(* styp -> bool *)
fun is_record_ctor (s, T) =
  String.isSuffix RecordPackage.extN s andalso
  let val body_T = body_type T in
    is_record_type body_T
    andalso s = unsuffix RecordPackage.ext_typeN (fst (dest_Type body_T)) ^
                RecordPackage.extN
  end
fun is_abs_fun thy (s, Type ("fun", [_, Type (s', _)])) =
    (case typedef_info thy s' of
       SOME {Abs_name, ...} => s = Abs_name
     | NONE => false)
  | is_abs_fun _ _ = false
fun is_rep_fun thy (s, Type ("fun", [Type (s', _), _])) =
    (case typedef_info thy s' of
       SOME {Rep_name, ...} => s = Rep_name
     | NONE => false)
  | is_rep_fun _ _ = false

(* theory -> styp -> styp *)
fun mate_of_rep_fun thy (x as (_, Type ("fun", [T1 as Type (s', _), T2]))) =
    (case typedef_info thy s' of
       SOME {Abs_name, ...} => (Abs_name, Type ("fun", [T2, T1]))
     | NONE => raise TERM ("mate_of_rep_fun", [Const x]))
  | mate_of_rep_fun _ x = raise TERM ("mate_of_rep_fun", [Const x])

(* theory -> styp -> bool *)
fun is_ctor_like thy (x as (s, T)) =
  (Refute.is_IDT_constructor thy x orelse is_record_ctor x
   orelse (is_abs_fun thy x andalso is_pure_typedef thy (range_type T))
   orelse s mem [@{const_name Zero_Rep}, @{const_name Suc_Rep}])
  andalso body_type T <> nat_T

(* typ -> typ *)
val varify_type = map_type_tfree (fn (s, S) => TVar ((s, 0), S))

(* theory -> typ -> typ -> typ -> typ *)
fun instantiate_type thy T1 T1' T2 =
  Envir.typ_subst_TVars (Sign.typ_match thy (varify_type T1, T1') Vartab.empty)
                        (varify_type T2)
  handle Type.TYPE_MATCH => raise TYPE ("instantiate_type", [T1, T1'], [])

(* typ list -> term -> int -> term *)
fun eta_expand _ t 0 = t
  | eta_expand Ts t n =
    fold_rev (curry3 Abs ("x\\<^isub>\\<eta>" ^ nat_subscript n))
             (List.take (binder_types (fastype_of1 (Ts, t)), n))
             (list_comb (incr_boundvars n t, map Bound (n - 1 downto 0)))

(* term -> term *)
fun extensionalize t =
  case t of
    (t0 as @{const Trueprop}) $ t1 => t0 $ extensionalize t1
  | Const (@{const_name "op ="}, _) $ t1 $ Abs (s, T, t2) =>
    let val v = Var ((s, maxidx_of_term t + 1), T) in
      extensionalize (HOLogic.mk_eq (t1 $ v, subst_bound (v, t2)))
    end
  | _ => t

(* typ -> term list -> term *)
fun distinctness_formula T ts =
  let
    (* term list -> term list -> term list *)
    fun aux [] = I
      | aux [_] = I
      | aux (t :: ts) =
        append (map (fn t' => @{const Not} $ (HOLogic.eq_const T $ t $ t')) ts)
        #> aux ts
  in
    if length ts <= 1 then @{const True}
    else aux ts [] |> foldl1 (HOLogic.mk_conj o swap)
  end

(* theory -> typ -> styp list *)
fun datatype_ctors thy (T as Type (s, Ts)) =
    if is_datatype_like thy T then
      case DatatypePackage.get_datatype thy s of
        SOME {descr, index, ...} =>
        let val (_, dtyps, ctors) = AList.lookup (op =) descr index |> the in
          map (fn (s', Us) =>
                  (s', map (Refute.typ_of_dtyp descr (dtyps ~~ Ts)) Us ---> T))
              (rev ctors)
         end
       | NONE =>
         if is_record_type T then
           let
             val s' = unsuffix RecordPackage.ext_typeN s ^ RecordPackage.extN
             val T' = (RecordPackage.get_extT_fields thy T
                       |> apsnd single |> uncurry append |> map snd) ---> T
           in [(s', T')] end
         else
           case typedef_info thy s of
             SOME {abs_type, rep_type, Abs_name, ...} =>
             [(Abs_name, instantiate_type thy abs_type T rep_type --> T)]
           | NONE =>
             if T = @{typ ind} then
               [dest_Const @{const Suc_Rep}, dest_Const @{const Zero_Rep}]
             else
               []
    else
      []
  | datatype_ctors _ _ = []

(* theory -> int -> (typ * int) list -> typ -> int *)
fun precise_card_of_type thy def_card asgns =
  let
    (* typ list -> typ -> int *)
    fun aux avoid T =
      if T mem avoid then
        0
      else
        (case T of
           Type ("fun", [T1, T2]) =>
           (case aux avoid T1 of 0 => 0 | k1 => power (aux avoid T2) k1)
         | Type ("*", [T1, T2]) => aux avoid T1 * aux avoid T2
         | Type (@{type_name itself}, _) => 1
         | @{typ prop} => 2
         | @{typ bool} => 2
         | @{typ unit} => 1
         | Type _ =>
           (case datatype_ctors thy T of
              [] => if is_integer_type T then 0 else raise SAME
            | ctors =>
              let
                val ctor_cards = datatype_ctors thy T
                                 |> map (foldl (op *) 1 o map (aux (T :: avoid))
                                         o binder_types o snd)
              in
                if exists (equal 0) ctor_cards then 0
                else foldl1 (op +) ctor_cards
              end)
         | _ => raise SAME)
        handle SAME => AList.lookup (op =) asgns T |> the_default def_card
  in aux [] end

(* theory -> typ -> bool *)
fun is_finite_type thy = not_equal 0 o precise_card_of_type thy 2 []

(* string -> string *)
fun original_name s =
  if String.isPrefix nitpick_prefix s then
    original_name (Substring.string (Substring.triml 1
        (snd (Substring.position "_" (Substring.full s)))))
  else
    s

(* term list -> (string * int) * typ list *)
fun special_bounds ts =
  fold Term.add_vars ts [] |> sort (fast_string_ord o pairself (fst o fst))

(* (string * int) * typ -> term -> term *)
fun abs_var ((s, j), T) body = Abs (s, T, abstract_over (Var ((s, j), T), body))

(* term list -> (string * string list) list -> special_fun list -> styp
   -> term *)
fun user_friendly_const eval skolems special_funs (x as (s, T)) =
  if String.isPrefix special_prefix s then
    let
      val ((s', T'), js, ts) = the_single (AList.find (op =) special_funs x)
      val max_j = List.last js
      val Ts = List.take (binder_types T', max_j + 1)
      val vars = special_bounds ts
      val ts' = map2 (fn T => AList.lookup (op =) (js ~~ ts)
                              #> the_default (Const ("\\<star>", T)))
                     Ts (0 upto max_j)
    in
      list_comb (user_friendly_const eval skolems special_funs (s', T'), ts')
      |> fold_rev abs_var vars
    end
  else if String.isPrefix wf_prefix s then
    lambda (Free (iter_var_prefix, nat_T))
           (Const (original_name s, range_type T))
  else if String.isPrefix skolem_prefix s then
    let
      val (binder_Ts, body_T) = strip_type T
      val ss = the (AList.lookup (op =) skolems s)
      val (Ts, Ts') = chop (length ss) binder_Ts
      val frees = map Free (ss ~~ Ts)
    in fold lambda frees (Const (original_name s, Ts' ---> T)) end
  else if String.isPrefix eval_prefix s then
    nth eval (the (Int.fromString (unprefix eval_prefix s)))
  else
    Const (original_name s, T)

(* term -> bool *)
fun is_arity_type_axiom (Const (@{const_name HOL.type_class}, _)
                         $ Const (@{const_name TYPE}, _)) = true
  | is_arity_type_axiom _ = false
(* theory -> bool -> term -> bool *)
fun is_typedef_axiom thy only_boring (@{const "==>"} $ _ $ t2) =
    is_typedef_axiom thy only_boring t2
  | is_typedef_axiom thy only_boring
        (@{const Trueprop} $ (Const (@{const_name Typedef.type_definition}, _)
         $ Const (_, Type ("fun", [Type (s, _), _])) $ Const _ $ _)) =
    not (only_boring andalso
         s mem [@{type_name unit}, @{type_name "*"}, @{type_name "+"},
                @{type_name rat}, @{type_name real}])
    andalso is_typedef thy s
  | is_typedef_axiom _ _ _ = false

(* Distinguishes between (1) constant definition axioms, (2) type arity and
   typedef axioms, and (3) other axioms, and returns the pair ((1), (3)).
   Typedef axioms are uninteresting to Nitpick, because it can retrieve them
   using "typedef_info". *)
(* theory -> (string * term) list -> string list -> term list * term list *)
fun partition_axioms_by_definitionality thy axioms def_names =
  let
    val axioms = sort (fast_string_ord o pairself fst) axioms
    val defs = OrdList.inter (fast_string_ord o apsnd fst) def_names axioms
    val nondefs =
      OrdList.subtract (fast_string_ord o apsnd fst) def_names axioms
      |> filter_out ((is_arity_type_axiom orf is_typedef_axiom thy true) o snd)
  in pairself (map snd) (defs, nondefs) end

(* theory -> bool *)
fun is_built_in_theory thy =
  exists (curry Theory.subthy thy)
         [@{theory Complex_Main}, @{theory NitpickDefs}]

(* theory -> term list * term list * term list *)
fun all_axioms_of thy =
  let
    (* theory list -> term list *)
    val axioms_of_thys = maps Thm.axioms_of #> map (apsnd prop_of)
    val def_names =
      Defs.all_specifications_of (Theory.defs_of thy) |> maps snd
      |> filter #is_def |> map #name |> OrdList.make fast_string_ord
    val thys = thy :: Theory.ancestors_of thy
    val (built_in_thys, user_thys) = List.partition is_built_in_theory thys
    val built_in_axioms = axioms_of_thys built_in_thys
    val user_axioms = axioms_of_thys user_thys
    val (built_in_defs, built_in_nondefs) =
      partition_axioms_by_definitionality thy built_in_axioms def_names
      |> apsnd (filter (is_typedef_axiom thy false))
    val (user_defs, user_nondefs) =
      partition_axioms_by_definitionality thy user_axioms def_names
  in (user_defs @ built_in_defs, built_in_nondefs, user_nondefs) end

(* styp -> int option *)
fun arity_of_built_in_const (s, T) =
  if s = @{const_name If} then
    if third_range_type T = @{typ bool} then NONE else SOME 3
  else
    case AList.lookup (op =) built_in_consts s of
      SOME n => SOME n
    | NONE =>
      case AList.lookup (op =) built_in_typed_consts (s, T) of
        SOME n => SOME n
      | NONE =>
        if is_fun_type T andalso is_set_type (domain_type T) then
          AList.lookup (op =) built_in_set_consts s
        else
          NONE
(* styp -> bool *)
val is_built_in_const = isSome o arity_of_built_in_const

(* This function is designed to work for both real definition axioms and
   simplification rules (equational specifications). *)
(* term -> term *)
fun term_under_def t =
  case t of
    @{const "==>"} $ _ $ t2 => term_under_def t2
  | Const (@{const_name "=="}, _) $ t1 $ _ => term_under_def t1
  | @{const Trueprop} $ t1 => term_under_def t1
  | Const (@{const_name "op ="}, _) $ t1 $ _ => term_under_def t1
  | Abs (_, _, t') => term_under_def t'
  | t1 $ _ => term_under_def t1
  | _ => t

(* string -> string -> string *)
fun prefix_name s = prefix (List.last (space_explode "." s) ^ ".")
(* string -> string *)
fun unprefix_name s = List.last (space_explode "." s) handle Empty => ""
(* string -> term -> term *)
val prefix_abs_vars = Term.map_abs_vars o prefix_name
(* term -> term *)
val unprefix_abs_vars = Term.map_abs_vars unprefix_name

(* Here we crucially rely on "Refute.specialize_type" performing a preorder
   traversal of the term, without which the wrong occurrence of a constant could
   be matched in the face of overloading. *)
(* theory -> const_table -> styp -> term list *)
fun def_props_for_const thy table (x as (s, _)) =
  if is_built_in_const x then
    []
  else
    these (Symtab.lookup table s)
    |> map_filter (try (Refute.specialize_type thy x))
    |> filter (equal (Const x) o term_under_def)

(* term -> term *)
fun normalized_rhs t =
  let
    (* term -> term *)
    fun aux (v as Var ((s, _), T)) t = Abs (s, T, abstract_over (v, t))
      | aux v t = raise TERM ("normalized_rhs", [v, t])
    val (lhs, rhs) =
      case t of
        Const (@{const_name "=="}, _) $ t1 $ t2 => (t1, t2)
      | @{const Trueprop} $ (Const (@{const_name "op ="}, _) $ t1 $ t2) =>
        (t1, t2)
      | _ => raise TERM ("normalized_rhs", [t])
    val args = Term.strip_comb lhs |> snd
  in fold_rev aux args rhs end

(* theory -> const_table -> styp -> term option *)
fun def_of_const thy table (x as (s, _)) =
  (if is_built_in_const x orelse String.isPrefix nitpick_prefix s then NONE
   else try (normalized_rhs o List.last) (def_props_for_const thy table x))
  |> Option.map (prefix_abs_vars s)

datatype fixpoint_kind = Lfp | Gfp | NoFp

(* term -> fixpoint_kind *)
fun fixpoint_kind_of_rhs (Abs (_, _, t)) = fixpoint_kind_of_rhs t
  | fixpoint_kind_of_rhs (Const (@{const_name lfp}, _) $ Abs _) = Lfp
  | fixpoint_kind_of_rhs (Const (@{const_name gfp}, _) $ Abs _) = Gfp
  | fixpoint_kind_of_rhs _ = NoFp

(* theory -> const_table -> term -> bool *)
fun is_mutually_inductive_pred_def thy table t =
  let
    (* term -> bool *)
    fun is_good_arg (Bound _) = true
      | is_good_arg (Const (s, _)) =
        s mem [@{const_name True}, @{const_name False}, @{const_name undefined}]
      | is_good_arg _ = false
  in
    case t |> strip_abs_body |> strip_comb of
      (Const x, ts as (_ :: _)) =>
      (case def_of_const thy table x of
         SOME t' => fixpoint_kind_of_rhs t' = Lfp andalso forall is_good_arg ts
       | NONE => false)
    | _ => false
  end
(* theory -> const_table -> term -> term *)
fun unfold_mutually_inductive_preds thy table =
  map_aterms (fn t as Const x =>
                 (case def_of_const thy table x of
                    SOME t' =>
                    let val t' = Envir.eta_contract t' in
                      if is_mutually_inductive_pred_def thy table t' then t'
                      else t
                    end
                 | NONE => t)
               | t => t)

(* term -> string * term *)
fun pair_for_prop t =
  case term_under_def t of
    Const (s, _) => (s, t)
  | Free _ => raise NOT_SUPPORTED "local definitions"
  | t => raise TERM ("pair_for_prop", [t, head_of t])

(* We add the "Nitpick" theorems to the tables if the "Nitpick" theory wasn't
   imported by the user. *)
(* (Proof.context -> term list) -> Proof.context -> const_table *)
fun table_for get ctxt =
  get ctxt
  |> (if member Theory.eq_thy
                (Theory.ancestors_of (ProofContext.theory_of ctxt))
                (@{theory NitpickDefs}) then
        I
      else
        append (get (ProofContext.init @{theory NitpickDefs})))
  |> map pair_for_prop |> AList.group (op =) |> Symtab.make

(* Proof.context -> term list -> const_table *)
fun const_def_table ctxt ts =
  table_for (map prop_of o Nitpick_Const_Def_Thms.get) ctxt
  |> fold (fn (s, t) => Symtab.map_default (s, []) (cons t))
          (map pair_for_prop ts)
(* term list -> const_table *)
fun const_nondef_table ts =
  fold (fn t => append (map (fn s => (s, t)) (Term.add_const_names t []))) ts []
  |> AList.group (op =) |> Symtab.make
(* Proof.context -> const_table *)
val const_simp_table = table_for (map prop_of o Nitpick_Const_Simp_Thms.get)
val const_psimp_table = table_for (map prop_of o Nitpick_Const_Psimp_Thms.get)
(* Proof.context -> const_table -> const_table *)
fun inductive_intro_table ctxt def_table =
  table_for (map (unfold_mutually_inductive_preds (ProofContext.theory_of ctxt)
                                                  def_table o prop_of)
             o Nitpick_Ind_Intro_Thms.get) ctxt

(* Similar to "Refute.specialize_type" but returns all matches rather than only
   the first (preorder) match. *)
(* theory -> styp -> term -> term list *)
fun multi_specialize_type thy (x as (s, T)) t =
  let
    (* term -> (typ * term) list -> (typ * term) list *)
    fun aux (Const (s', T')) ys =
        if s = s' then
          (if AList.defined (op =) ys T' then
             I
           else if T = T' then
             cons (T, t)
           else
             cons (T', Refute.monomorphic_term
                           (Sign.typ_match thy (T', T) Vartab.empty) t)
             handle Type.TYPE_MATCH => I) ys
        else
          ys
      | aux _ ys = ys
  in map snd (fold_aterms aux t []) end

(* theory -> const_table -> styp -> term list *)
fun nondef_props_for_const thy table (x as (s, _)) =
  these (Symtab.lookup table s) |> maps (multi_specialize_type thy x)
  handle Refute.REFUTE _ =>
         raise NOT_SUPPORTED ("too much polymorphism in axiom involving " ^
                              quote s)

(* theory -> styp list -> term list *)
fun optimized_typedef_axioms thy (s, Ts) =
  if is_univ_typedef thy (Type (s, Ts)) then
    []
  else
    case typedef_info thy s of
      SOME {abs_type, rep_type, Abs_name, Rep_name, prop_of_Rep, set_name,
            ...} =>
      let
        val abs_T = Type (s, Ts)
        val rep_T = instantiate_type thy abs_type abs_T rep_type
        val abs_const = Const (Abs_name, rep_T --> abs_T)
        val rep_const = Const (Rep_name, abs_T --> rep_T)
        val rep_t = prop_of_Rep |> HOLogic.dest_Trueprop
        val set_T = rep_T --> bool_T
      in
        [HOLogic.mk_eq (Const (set_name, set_T),
             rep_t |> Refute.specialize_type thy (dest_Const rep_const)
                   |> HOLogic.dest_mem |> snd),
         Const (@{const_name ord_fun_inst.less_eq_fun},
                [set_T, set_T] ---> bool_T)
             $ (Const (@{const_name image},
                       [abs_T --> rep_T, abs_T --> bool_T, rep_T] ---> bool_T)
                 $ rep_const $ Const (@{const_name UNIV}, abs_T --> bool_T))
             $ Const (set_name, set_T)]
        |> map HOLogic.mk_Trueprop
      end
    | NONE => []

val redefined_in_NitpickDefs_thy =
  [@{const_name option_case}, @{const_name nat_case}, @{const_name list_case},
   @{const_name list_size}]

(* Constants "c" whose definition is of the form "c == c'", where "c'" is also a
   constant, are said to be trivial. For those, we ignore the simplification
   rules and use the definition instead, to ensure that built-in symbols like
   "ord_nat_inst.less_eq_nat" are picked up correctly. *)
(* theory -> const_table -> styp -> bool *)
fun has_trivial_definition thy table x =
  case def_of_const thy table x of SOME (Const _) => true | _ => false

(* theory -> const_table -> string * typ -> fixpoint_kind *)
fun fixpoint_kind_of_const thy table x =
  if is_built_in_const x then
    NoFp
  else
    fixpoint_kind_of_rhs (the (def_of_const thy table x)) handle Option => NoFp

(* extended_context -> styp -> bool *)
fun is_real_inductive_pred (ext_ctxt as {thy, def_table, intro_table, ...}) x =
  not (null (def_props_for_const thy intro_table x))
  andalso fixpoint_kind_of_const thy def_table x = Lfp
fun is_real_equational_fun {thy, simp_table, psimp_table, ...} x =
  exists (fn table => not (null (def_props_for_const thy table x)))
         [!simp_table, psimp_table]
fun is_inductive_pred ext_ctxt =
  is_real_inductive_pred ext_ctxt andf (not o is_real_equational_fun ext_ctxt)
fun is_equational_fun (ext_ctxt as {thy, def_table, ...}) =
  (not o member (op =) redefined_in_NitpickDefs_thy o fst)
  andf (is_real_equational_fun ext_ctxt orf is_real_inductive_pred ext_ctxt)
  andf (not o has_trivial_definition thy def_table)

exception NO_TRIPLE

(* theory -> styp -> term -> term list * term list * term *)
fun triple_for_intro_rule thy x t =
  let
    val prems = Logic.strip_imp_prems t |> map (ObjectLogic.atomize_term thy)
    val concl = Logic.strip_imp_concl t |> ObjectLogic.atomize_term thy
    val (main, side) = List.partition (fn t => x mem Term.add_consts t []) prems
    (* term -> bool *)
    val is_good_head = equal (Const x) o head_of
  in
    if forall is_good_head main then (side, main, concl) else raise NO_TRIPLE
  end

(* term -> term *)
val tuple_for_args = foldl1 HOLogic.mk_prod o snd o strip_comb

(* indexname * typ -> term list -> term -> term -> term *)
fun wf_constraint_for rel side concl main =
  let
    val core = HOLogic.mk_mem (HOLogic.mk_prod (tuple_for_args main,
                                                tuple_for_args concl), Var rel)
    val t = foldl HOLogic.mk_imp core side
    val vars = filter (not_equal rel) (Term.add_vars t [])
  in
    Library.foldl (fn (t', ((x, j), T)) =>
                      HOLogic.all_const T
                      $ Abs (x, T, abstract_over (Var ((x, j), T), t')))
                  (t, vars)
  end

(* indexname * typ -> term list * term list * term -> term *)
fun wf_constraint_for_triple rel (side, main, concl) =
  map (wf_constraint_for rel side concl) main |> foldl1 HOLogic.mk_conj

(* Proof.context -> Time.time option -> thm
   -> (Proof.context -> tactic -> tactic) -> bool *)
fun terminates_by ctxt timeout goal tac =
  can (SINGLE (Classical.safe_tac (local_claset_of ctxt)) #> the
       #> SINGLE (timeout_tac timeout
                              (tac ctxt (auto_tac (local_clasimpset_of ctxt))))
       #> the #> Goal.finish) goal

(* extended_context -> const_table -> styp -> bool *)
fun is_is_well_founded_inductive_pred {thy, ctxt, debug, tac_timeout,
                                       intro_table, ...} (x as (_, T)) =
  case def_props_for_const thy intro_table x of
    [] => raise TERM ("is_is_well_founded_inductive_pred", [Const x])
  | intro_ts =>
    (case map (triple_for_intro_rule thy x) intro_ts
          |> filter_out (null o #2) of
       [] => true
     | triples =>
       let
         val binders_T = foldl1 HOLogic.mk_prodT (binder_types T)
         val rel_T = HOLogic.mk_prodT (binders_T, binders_T) --> bool_T
         val j = foldl Int.max 0 (map maxidx_of_term intro_ts) + 1
         val rel = (("R", j), rel_T)
         val prop = Const (@{const_name wf}, rel_T --> bool_T) $ Var rel ::
                    map (wf_constraint_for_triple rel) triples
                    |> foldl1 HOLogic.mk_conj |> HOLogic.mk_Trueprop
         val _ = if debug then
                   priority ("Wellfoundedness goal: " ^
                             Syntax.string_of_term ctxt prop)
                 else
                   ()
         val goal = cterm_of thy prop |> Goal.init
         val tacs = [LexicographicOrder.lex_order_tac,
                     ScnpReconstruct.sizechange_tac]
       in silence (exists (terminates_by ctxt tac_timeout goal)) tacs end)
    handle Empty => false
         | NO_TRIPLE => false

(* extended_context -> styp -> bool *)
fun is_well_founded_inductive_pred
        (ext_ctxt as {inductive_mood, intro_table, wfs, ...}) (x as (s, _)) =
  case inductive_mood of
    Optimistic => true
  | Pessimistic => false
  | Realistic =>
    if s = @{const_name fold_graph'} then
      true
    else
      case AList.lookup (op =) (!wfs) x of
        SOME wf => wf
      | NONE =>
        let
          val wf = is_is_well_founded_inductive_pred ext_ctxt x
          val _ = change wfs (cons (x, wf))
        in wf end

(* extended_context -> styp -> term *)
fun better_founded_inductive_pred_const
        (ext_ctxt as {thy, def_table, simp_table, ...}) (x as (s, T)) =
  let
    val iter_T = iterator_type_for_const x
    val x' as (s', _) = (wf_prefix ^ s, iter_T --> T)
    val surrogate = Const x'
                    $ Const (@{const_name zero_nat_inst.zero_nat}, iter_T)
  in
    (if is_equational_fun ext_ctxt x' then
       () (* already done *)
     else
       let
         val def = the (def_of_const thy def_table x)
         val j = maxidx_of_term def + 1
         val (outer, fp_app) = strip_abs def
         val outer_bounds = map Bound (length outer - 1 downto 0)
         val cur = Var ((iter_var_prefix, j + 1), iter_T)
         val next = Const (@{const_name Suc}, iter_T --> iter_T) $ cur
         val rhs = case fp_app of
                     Const _ $ t =>
                     betapply (t, list_comb (Const x', next :: outer_bounds))
                   | _ => raise TERM ("inductive_fun_axiom", [fp_app])
         val (inner, naked_rhs) = strip_abs rhs
         val all = outer @ inner
         val bounds = map Bound (length all - 1 downto 0)
         val vars = map (fn (s, T) => Var ((s, j), T)) all
         val eq = HOLogic.mk_eq (list_comb (Const x', cur :: bounds), naked_rhs)
                  |> HOLogic.mk_Trueprop |> curry subst_bounds (rev vars)
       in
         simp_table := Symtab.update (s',
                           eq :: these (Symtab.lookup (!simp_table) s'))
                           (!simp_table)
       end);
    surrogate
  end

(* extended_context -> styp -> term *)
fun raw_inductive_pred_axiom {thy, def_table, ...} x =
  let
    val def = the (def_of_const thy def_table x)
    val (outer, fp_app) = strip_abs def
    val outer_bounds = map Bound (length outer - 1 downto 0)
    val rhs = case fp_app of
                Const _ $ t => betapply (t, list_comb (Const x, outer_bounds))
              | _ => raise TERM ("raw_inductive_pred_axiom", [fp_app])
    val (inner, naked_rhs) = strip_abs rhs
    val all = outer @ inner
    val bounds = map Bound (length all - 1 downto 0)
    val j = maxidx_of_term def + 1
    val vars = map (fn (s, T) => Var ((s, j), T)) all
  in
    HOLogic.mk_eq (list_comb (Const x, bounds), naked_rhs)
    |> HOLogic.mk_Trueprop |> curry subst_bounds (rev vars)
  end

(* extended_context -> styp -> term list *)
fun raw_equational_fun_axioms
        (ext_ctxt as {thy, simp_table, psimp_table, ...}) (x as (s, _)) =
  if s mem redefined_in_NitpickDefs_thy then
    []
  else
    case def_props_for_const thy (!simp_table) x of
      [] => (case def_props_for_const thy psimp_table x of
               [] => [raw_inductive_pred_axiom ext_ctxt x]
             | psimps => psimps)
    | simps => simps

val equational_fun_axioms = map extensionalize oo raw_equational_fun_axioms

(* int -> term list -> term list -> (term * term list) option *)
fun find_silly_assignment _ _ [] = NONE
  | find_silly_assignment j seen (t :: ts) =
    case t of
      Const (@{const_name "op ="}, _) $ t1 $ t2 =>
      if t1 = Bound j andalso not (loose_bvar1 (t2, j)) then
        SOME (t2, seen @ ts)
      else if t2 = Bound j andalso not (loose_bvar1 (t1, j)) then
        SOME (t1, seen @ ts)
      else
        find_silly_assignment j (t :: seen) ts
    | _ => find_silly_assignment j (t :: seen) ts

(* int -> term -> term -> term *)
fun subst_one_bound j arg t =
  let
    fun aux (Bound i, lev) =
        if i < lev then raise SAME
        else if i = lev then incr_boundvars (lev - j) arg
        else Bound (i - 1)
      | aux (Abs (a, T, body), lev) = Abs (a, T, aux (body, lev + 1))
      | aux (f $ t, lev) =
        (aux (f, lev) $ (aux (t, lev) handle SAME => t)
         handle SAME => f $ aux (t, lev))
      | aux _ = raise SAME
  in aux (t, j) handle SAME => t end

(* term -> term *)
val destroy_equalities =
  let
    (* string list -> typ list -> term list -> term *)
    fun kill [] [] ts = foldr1 HOLogic.mk_conj ts
      | kill (s :: ss) (T :: Ts) ts =
        (case find_silly_assignment (length ss) [] ts of
           SOME (_, []) => @{const True}
         | SOME (arg_t, ts) =>
           kill ss Ts (map (subst_one_bound (length ss)
                                (incr_bv (~1, length ss + 1, arg_t))) ts)
         | NONE =>
           Const (@{const_name Ex}, (T --> bool_T) --> bool_T)
           $ Abs (s, T, kill ss Ts ts))
      | kill _ _ _ = raise UnequalLengths
    (* string list -> typ list -> term -> term *)
    fun gather ss Ts ((t0 as Const (@{const_name Ex}, _)) $ Abs (s1, T1, t1)) =
        gather (ss @ [s1]) (Ts @ [T1]) t1
      | gather [] [] (Abs (s, T, t1)) = Abs (s, T, gather [] [] t1)
      | gather [] [] (t1 $ t2) = gather [] [] t1 $ gather [] [] t2
      | gather [] [] t = t
      | gather ss Ts t = kill ss Ts (HOLogic.dest_conj (gather [] [] t))
  in gather [] [] end

(* styp -> (int * term option) list *)
fun zombie_args_in_term x t =
  let
    (* term list -> term -> term list -> term list list *)
    fun fun_calls accum (Abs (_, _, t)) _ = fun_calls accum t []
      | fun_calls accum (t1 $ t2) arg_ts =
        fun_calls (fun_calls accum t2 []) t1 (t2 :: arg_ts)
      | fun_calls accum t arg_ts = accum |> t = Const x ? cons arg_ts
    (* term list list -> term list list -> term list -> term list list *)
    fun call_sets [] [] vs = [vs]
      | call_sets [] uss vs = vs :: call_sets uss [] []
      | call_sets ([] :: _) _ _ = []
      | call_sets ((t :: ts) :: tss) uss vs =
        OrdList.insert TermOrd.term_ord t vs |> call_sets tss (ts :: uss)
    val sets = call_sets (fun_calls [] t []) [] []
    val indexed_sets = sets ~~ (0 upto length sets - 1)
  in
    fold_rev (fn (set, j) =>
                 case set of
                   [Var _] => AList.lookup (op =) indexed_sets set = SOME j
                              ? cons (j, NONE)
                 | [t as Const _] => cons (j, SOME t)
                 | [t as Free _] => cons (j, SOME t)
                 | _ => I) indexed_sets []
  end
(* styp -> term list -> (int * term option) list *)
fun zombie_args_in_terms x =
  map (zombie_args_in_term x)
  #> fold1 (OrdList.inter (prod_ord int_ord (option_ord TermOrd.term_ord)))

(* term -> term list *)
fun params_in_equation (@{const "==>"} $ _ $ t2) = params_in_equation t2
  | params_in_equation (@{const Trueprop} $ t1) = params_in_equation t1
  | params_in_equation (Const (@{const_name "op ="}, _) $ t1 $ _) =
    snd (strip_comb t1)
  | params_in_equation _ = []

(* styp -> styp -> int list -> term list -> term list -> term -> term *)
fun specialize_fun_axiom x x' fixed_js fixed_args bound_args t =
  let
    val fixed_params = filter_indices fixed_js (params_in_equation t)
    (* term list -> term -> term *)
    fun aux args (Abs (s, T, t)) = list_comb (Abs (s, T, aux [] t), args)
      | aux args (t1 $ t2) = aux (aux [] t2 :: args) t1
      | aux args t =
        if t = Const x then
          list_comb (Const x', bound_args @ filter_out_indices fixed_js args)
        else
          let val j = find_index (equal t) fixed_params in
            list_comb (if j >= 0 then nth fixed_args j else t, args)
          end
  in aux [] t end

(* typ list -> term -> bool *)
fun is_candidate_argument Ts t =
  let val bound_js = loose_bnos t in
    null bound_js
    orelse (is_higher_order_type (fastype_of1 (Ts, t))
            andalso forall (not o is_higher_order_type o nth Ts) bound_js)
  end

(* (int * term option) list -> (int * term) list -> int list *)
fun overlapping_indices [] _ = []
  | overlapping_indices _ [] = []
  | overlapping_indices (ps1 as (j1, t1) :: ps1') (ps2 as (j2, t2) :: ps2') =
    if j1 < j2 then overlapping_indices ps1' ps2
    else if j1 > j2 then overlapping_indices ps1 ps2'
    else overlapping_indices ps1' ps2' |> the_default t2 t1 = t2 ? cons j1

(* extended_context -> int -> term -> term *)
fun specialize_consts_in_term
        (ext_ctxt as {special_depth, simp_table, special_funs, ...}) depth t =
  if depth > special_depth then
    t
  else
    let
      (* ### strong enough in the face of user-defined axioms? *)
      val blacklist = if depth = 0 then []
                      else case term_under_def t of Const x => [x] | _ => []
      (* term list -> typ list -> term -> term *)
      fun aux args Ts (Const (x as (s, T))) =
          ((if not (x mem blacklist) andalso not (null args)
               andalso not (String.isPrefix special_prefix s)
               andalso is_equational_fun ext_ctxt x then
              let
                val candidate_args =
                  filter (is_candidate_argument Ts o snd)
                         ((0 upto length args - 1) ~~ args)
                val _ = not (null candidate_args) orelse raise SAME
                val old_axs = equational_fun_axioms ext_ctxt x
                              |> map destroy_equalities
                val zombie_params = zombie_args_in_terms x old_axs
                val fixed_js = overlapping_indices zombie_params candidate_args
                val _ = not (null fixed_js) orelse raise SAME
                val fixed_args = filter_indices fixed_js args
                val bound_js = fold (fn t => fn js => add_loose_bnos (t, 0, js))
                                    fixed_args []
                               |> sort int_ord
                val variable_args = filter_out_indices fixed_js args
                val extra_args = map Bound bound_js @ variable_args
                val bound_Ts = filter_indices bound_js Ts
                val k = maxidx_of_term t + 1
                (* int -> term *)
                fun var_for_bound_no j =
                  Var ((bound_var_prefix ^
                        nat_subscript (find_index (equal j) bound_js + 1), k),
                       nth Ts j)
                val fixed_args_in_axiom =
                  map (curry subst_bounds
                             (map var_for_bound_no (0 upto length Ts - 1)))
                      fixed_args
              in
                case AList.lookup (op =) (!special_funs)
                                  (x, fixed_js, fixed_args_in_axiom) of
                  SOME x' => list_comb (Const x', extra_args)
                | NONE =>
                  let
                    val bound_args_in_axiom = map var_for_bound_no bound_js
                    val x' =
                      (special_prefix_for (length (!special_funs) + 1) ^ s,
                       bound_Ts @ filter_out_indices fixed_js (binder_types T)
                       ---> body_type T)
                    val new_axs =
                      map (specialize_fun_axiom x x' fixed_js
                               fixed_args_in_axiom bound_args_in_axiom) old_axs
                    val _ =
                      change special_funs
                             (cons ((x, fixed_js, fixed_args_in_axiom), x'))
                    val _ =
                      change simp_table (Symtab.update_new (fst x', new_axs))
                  in list_comb (Const x', extra_args) end
              end
            else
              raise SAME)
           handle SAME => list_comb (Const x, args))
        | aux args Ts (Abs (s, T, t)) =
          list_comb (Abs (s, T, aux [] (T :: Ts) t), args)
        | aux args Ts (t1 $ t2) = aux (aux [] Ts t2 :: args) Ts t1
        | aux args _ t = list_comb (t, args)
    in aux [] [] t end

val unfold_max_depth = 63
val axioms_max_depth = 63

(* extended_context -> term -> term *)
fun unfold_defs_in_term (ext_ctxt as {thy, def_table, ...}) =
  let
    (* int -> typ list -> term -> term *)
    fun aux depth Ts t =
      case t of
        (t0 as Const (@{const_name Int.number_class.number_of},
                      Type ("fun", [_, ran_T]))) $ t1 =>
        ((if is_number_type ran_T then
            let
              val s = numeral_prefix ^ string_for_int (HOLogic.dest_numeral t1)
            in
              if is_integer_type ran_T then
                Const (s, ran_T)
              else
                aux depth Ts (Const (@{const_name of_int}, int_T --> ran_T)
                              $ Const (s, int_T))
            end
            handle TERM _ => raise SAME
          else
            raise SAME)
         handle SAME => betapply (aux depth Ts t0, aux depth Ts t1))
      | (t0 as Const (x as (@{const_name Sigma}, T))) $ t1
        $ (t2 as Abs (_, _, t2')) =>
        betapplys ((loose_bvar1 (t2', 0) ? aux depth Ts) t0,
                   map (aux depth Ts) [t1, t2])
      | Const (x as (@{const_name distinct},
               Type ("fun", [Type (@{type_name list}, [T']), _]))) $ t1 =>
        (t1 |> HOLogic.dest_list |> distinctness_formula T'
         handle TERM _ => aux_const depth Ts x [t1])
      | Const x $ t1 $ t2 $ t3 => aux_const depth Ts x [t1, t2, t3]
      | Const x $ t1 $ t2 => aux_const depth Ts x [t1, t2]
      | Const x $ t1 => aux_const depth Ts x [t1]
      | Const x => aux_const depth Ts x []
      | Free _ => t
      | Var _ => t
      | Bound _ => t
      | Abs (s, T, body) => Abs (s, T, aux depth (T :: Ts) body)
      | t1 $ t2 => betapply (aux depth Ts t1, aux depth Ts t2)
    (* int -> typ list -> styp -> term list -> term *)
    and aux_const depth Ts (x as (s, T)) ts =
      case AList.lookup (op =) alternative_table s of
        SOME s' => aux_const (depth + 1) Ts (s', T) ts
      | NONE =>
        let
          val (arity, const) =
            case arity_of_built_in_const x of
              SOME n =>
              if s = @{const_name finite}
                 andalso is_finite_type thy (domain_type T) then
                (1, Abs ("finite_set", domain_type T, @{const True}))
              else
                (n, Const x)
            | NONE =>
              if is_inductive_pred ext_ctxt x
                 andalso not (is_real_equational_fun ext_ctxt x)
                 andalso not (is_well_founded_inductive_pred ext_ctxt x) then
                (0, better_founded_inductive_pred_const ext_ctxt x)
              else if (is_ctor_like thy orf is_equational_fun ext_ctxt) x then
                (0, Const x)
              else
                case def_of_const thy def_table x of
                  SOME def =>
                  if depth > unfold_max_depth then
                    raise LIMIT ("unfold_defs_in_term",
                                 "too many nested definitions (" ^
                                 string_for_int depth ^ ") while expanding " ^
                                 quote s)
                  else
                    (0, aux (depth + 1) Ts def)
                | NONE => (0, Const x)
            val const_of_ts =
              foldl (betapply o swap) const (map (aux depth Ts) ts)
              |> Envir.beta_norm
          in eta_expand Ts const_of_ts (arity nat_minus length ts) end
  in aux 0 [] end

(* int -> term -> term *)
fun eval_axiom_for_term j t =
  Logic.mk_equals (Const (eval_prefix ^ string_for_int j, fastype_of t), t)

(* extended_context -> styp -> bool *)
fun is_equational_fun_surely_complete ext_ctxt x =
  case raw_equational_fun_axioms ext_ctxt x of
    [@{const Trueprop} $ (Const (@{const_name "op ="}, _) $ t1 $ _)] =>
    strip_comb t1 |> snd |> forall is_Var
  | _ => false

type special = int list * term list * styp

(* styp -> special -> special -> term *)
fun special_congruence_axiom (s, T) (js1, ts1, x1) (js2, ts2, x2) =
  let
    val (bounds1, bounds2) = pairself (map Var o special_bounds) (ts1, ts2)
    val Ts = binder_types T
    val max_j = fold (fold (curry Int.max)) [js1, js2] ~1
    val (eqs, (args1, args2)) =
      fold (fn j => case pairself (fn ps => AList.lookup (op =) ps j)
                                  (js1 ~~ ts1, js2 ~~ ts2) of
                      (SOME t1, SOME t2) => apfst (cons (t1, t2))
                    | (SOME t1, NONE) => apsnd (apsnd (cons t1))
                    | (NONE, SOME t2) => apsnd (apfst (cons t2))
                    | (NONE, NONE) =>
                      let val v = Var ((cong_var_prefix ^ nat_subscript j, 0),
                                       nth Ts j) in
                        apsnd (pairself (cons v))
                      end) (max_j downto 0) ([], ([], []))
  in
    Logic.list_implies (eqs |> filter_out (op =) |> distinct (op =)
                            |> map Logic.mk_equals,
                        Logic.mk_equals (list_comb (Const x1, bounds1 @ args1),
                                         list_comb (Const x2, bounds2 @ args2)))
    |> Refute.close_form
  end

(* extended_context -> styp list -> term list *)
fun special_congruence_axioms (ext_ctxt as {special_funs, ...}) xs =
  let
    val groups =
      !special_funs
      |> map (fn ((x, js, ts), x') => (x, (js, ts, x')))
      |> AList.group (op =)
      |> filter_out (is_equational_fun_surely_complete ext_ctxt o fst)
      |> map (fn (x, zs) => (x, zs |> (x mem xs) ? cons ([], [], x)))
    (* special -> int *)
    fun generality (js, _, _) = ~(length js)
    (* special -> special -> bool *)
    fun is_more_specific (j1, t1, x1) (j2, t2, x2) =
      x1 <> x2 andalso OrdList.subset (prod_ord int_ord TermOrd.term_ord)
                                      (j2 ~~ t2, j1 ~~ t1)
    (* styp -> special list -> special list -> special list -> term list
       -> term list *)
    fun do_pass_1 _ [] [_] [_] = I
      | do_pass_1 x skipped _ [] = do_pass_2 x skipped
      | do_pass_1 x skipped all (z :: zs) =
        case filter (is_more_specific z) all
             |> sort (int_ord o pairself generality) of
          [] => do_pass_1 x (z :: skipped) all zs
        | (z' :: _) => cons (special_congruence_axiom x z z')
                       #> do_pass_1 x skipped all zs
    (* styp -> special list -> term list -> term list *)
    and do_pass_2 _ [] = I
      | do_pass_2 x (z :: zs) =
        fold (cons o special_congruence_axiom x z) zs #> do_pass_2 x zs
  in fold (fn (x, zs) => do_pass_1 x [] zs zs) groups [] end

(* term -> bool *)
fun is_trivial_equation (Const (@{const_name all}, _) $ Abs (_, _, t)) =
    is_trivial_equation t
  | is_trivial_equation t =
    try (op aconv o Logic.dest_equals) t |> the_default false

(* 'a Symtab.table -> 'a list *)
fun all_table_entries table = Symtab.fold (append o snd) table []
(* const_table -> string -> const_table *)
fun extra_table table s = Symtab.make [(s, all_table_entries table)]

(* extended_context -> term -> (term list * term list) * (bool * bool) *)
fun axioms_for_term (ext_ctxt as {thy, user_axioms, eval, def_table,
                                  nondef_table, user_nondefs, ...}) t =
  let
    type accumulator = styp list * (term list * term list)
    (* (term list * term list -> term list)
       -> ((term list -> term list) -> term list * term list
           -> term list * term list)
       -> int -> term -> accumulator -> accumulator *)
    fun add_axiom get app depth t (accum as (xs, axs)) =
      let val t = unfold_defs_in_term ext_ctxt t |> Refute.close_form in
        if is_trivial_equation t then
          accum
        else
          let
            val t = destroy_equalities t
            val t' = specialize_consts_in_term ext_ctxt depth t
          in
            if exists (member (op aconv) (get axs)) [t, t'] then accum
            else add_axioms_for_term (depth + 1) t' (xs, app (cons t') axs)
          end
      end
    (* int -> term -> accumulator -> accumulator *)
    and add_nondef_axiom depth = add_axiom snd apsnd depth
    and add_def_axiom depth t =
      (if head_of t = @{const "==>"} then add_nondef_axiom
       else add_axiom fst apfst) depth t
    (* int -> term -> accumulator -> accumulator *)
    and add_axioms_for_term depth t (accum as (xs, axs)) =
      case t of
        Const (x as (s, T)) =>
        (if x mem xs orelse is_built_in_const x then
           accum
         else
           let val accum as (xs, _) = (x :: xs, axs) in
             if depth > axioms_max_depth then
               raise LIMIT ("axioms_for_term",
                            "too many nested axioms (" ^ string_for_int depth ^
                            ")")
             else if Refute.is_const_of_class thy x then
               let
                 val class = Logic.class_of_const s
                 val inclass = Logic.mk_inclass (TVar (("'a", 0), [class]),
                                                 class)
                 val ax1 = try (Refute.specialize_type thy x) inclass
                 val ax2 = Option.map (Refute.specialize_type thy x o snd)
                                      (Refute.get_classdef thy class)
               in fold (add_def_axiom depth) (map_filter I [ax1, ax2]) accum end
             else if is_equational_fun ext_ctxt x then
               fold (add_def_axiom depth) (equational_fun_axioms ext_ctxt x)
                    accum
             else if s = @{const_name The}
                  orelse s = @{const_name Hilbert_Choice.Eps} then
               raise BAD ("axioms_for_term", "missing \"Nitpick\" theorems")
             else if is_abs_fun thy x
                     andalso not (is_pure_typedef thy (range_type (snd x))) then
               accum |> fold (add_nondef_axiom depth)
                             (nondef_props_for_const thy nondef_table x)
                     |> fold (add_def_axiom depth)
                             (nondef_props_for_const thy
                                                    (extra_table def_table s) x)
             else if is_rep_fun thy x andalso
                     not (is_pure_typedef thy (domain_type (snd x))) then
               accum |> fold (add_nondef_axiom depth)
                             (nondef_props_for_const thy nondef_table x)
                     |> fold (add_def_axiom depth)
                             (nondef_props_for_const thy
                                                    (extra_table def_table s) x)
                     |> add_axioms_for_term depth
                                            (Const (mate_of_rep_fun thy x))
             else if is_ctor_like thy x then
               accum
             else
               accum |> user_axioms <> SOME false
                        ? fold (add_nondef_axiom depth)
                               (nondef_props_for_const thy nondef_table x)
           end)
        |> add_axioms_for_type depth T
      | Free (_, T) => add_axioms_for_type depth T accum
      | Var (_, T) => add_axioms_for_type depth T accum
      | Bound _ => accum
      | Abs (_, T, t) => accum |> add_axioms_for_term depth t
                               |> add_axioms_for_type depth T
      | t1 $ t2 => accum |> fold (add_axioms_for_term depth) [t1, t2]
    (* int -> typ -> accumulator -> accumulator *)
    and add_axioms_for_type depth T =
      case T of
        Type ("fun", Ts) => fold (add_axioms_for_type depth) Ts
      | Type ("*", Ts) => fold (add_axioms_for_type depth) Ts
      | @{typ prop} => I
      | @{typ bool} => I
      | @{typ unit} => I
      | TFree (_, S) => add_axioms_for_sort depth T S
      | TVar (_, S) => add_axioms_for_sort depth T S
      | Type (s, Ts) =>
        fold (add_axioms_for_type depth) Ts
        #> (if is_pure_typedef thy T then
              fold (add_def_axiom depth) (optimized_typedef_axioms thy (s, Ts))
            else
              I)
    (* int -> typ -> sort -> accumulator -> accumulator *)
    and add_axioms_for_sort depth T S =
      let
        val supers = S @ maps (Sign.super_classes thy) S |> distinct (op =)
        val class_axioms =
          maps (fn class => map prop_of (AxClass.get_info thy class |> #axioms
                                         handle ERROR _ => [])) supers
        val monomorphic_class_axioms =
          map (fn t => case Term.add_tvars t [] of
                         [] => t
                       | [(x, S)] =>
                         Refute.monomorphic_term (Vartab.make [(x, (S, T))]) t
                       | _ => raise TERM ("add_axioms_for_sort", [t]))
              class_axioms
      in fold (add_def_axiom depth) monomorphic_class_axioms end
    val (mono_user_nondefs, poly_user_nondefs) =
      List.partition (null o Term.hidden_polymorphism) user_nondefs
    val eval_axioms = map2 eval_axiom_for_term (0 upto length eval - 1) eval
    val (xs, (defs, nondefs)) =
      add_axioms_for_term 1 t ([], ([], []))
      |> fold (add_def_axiom 1) eval_axioms
      |> user_axioms = SOME true ? fold (add_nondef_axiom 1) mono_user_nondefs
    val defs = defs @ special_congruence_axioms ext_ctxt xs
  in
    ((defs, nondefs), (user_axioms = SOME true orelse null mono_user_nondefs,
                       null poly_user_nondefs))
  end

(* (string * string list) list ref -> int -> term -> term *)
fun skolemize_term skolems max_depth =
  let
    (* int list -> int list *)
    val incrs = map (curry (op +) 1)
    (* string list -> typ list -> int list -> int -> bool -> term -> term *)
    fun aux ss Ts js depth neg t =
      let
        (* string -> typ -> string -> typ -> term -> term *)
        fun do_quantifier quant_s quant_T abs_s abs_T t =
          if not (loose_bvar1 (t, 0)) then
            aux ss Ts js depth neg (incr_boundvars ~1 t)
          else if depth <= max_depth
                  andalso neg = (quant_s <> @{const_name Ex}) then
            let
              val j = length (!skolems) + 1
              val sko_s = skolem_prefix_for j ^ abs_s
              val _ = change skolems (cons (sko_s, ss))
              val sko_t = list_comb (Const (sko_s, rev Ts ---> abs_T),
                                     map Bound (rev js))
              val abs_t = Abs (abs_s, abs_T, aux ss Ts (incrs js) depth neg t)
            in
              if null js then betapply (abs_t, sko_t)
              else Const (@{const_name Let}, abs_T --> quant_T) $ sko_t $ abs_t
            end
          else
            Const (quant_s, quant_T)
            $ Abs (abs_s, abs_T,
                   if is_higher_order_type abs_T then
                     t
                   else
                     aux (abs_s :: ss) (abs_T :: Ts) (0 :: incrs js)
                         (depth + 1) neg t)
      in
        case t of
          Const (s0 as @{const_name all}, T0) $ Abs (s1, T1, t1) =>
          do_quantifier s0 T0 s1 T1 t1
        | Const (s0 as @{const_name All}, T0) $ Abs (s1, T1, t1) =>
          do_quantifier s0 T0 s1 T1 t1
        | Const (s0 as @{const_name Ex}, T0) $ Abs (s1, T1, t1) =>
          do_quantifier s0 T0 s1 T1 t1
        | @{const Pure.conjunction} $ t1 $ t2 =>
          @{const Pure.conjunction} $ aux ss Ts js depth neg t1
          $ aux ss Ts js depth neg t2
        | @{const Trueprop} $ t1 =>
          @{const Trueprop} $ aux ss Ts js depth neg t1
        | @{const Not} $ t1 => @{const Not} $ aux ss Ts js depth (not neg) t1
        | @{const "==>"} $ t1 $ t2 =>
          @{const "==>"} $ aux ss Ts js depth (not neg) t1
          $ aux ss Ts js depth neg t2
        | @{const "op &"} $ t1 $ t2 =>
          @{const "op &"} $ aux ss Ts js depth neg t1
          $ aux ss Ts js depth neg t2
        | @{const "op |"} $ t1 $ t2 =>
          @{const "op |"} $ aux ss Ts js depth neg t1
          $ aux ss Ts js depth neg t2
        | @{const "op -->"} $ t1 $ t2 =>
          @{const "op -->"} $ aux ss Ts js depth (not neg) t1
          $ aux ss Ts js depth neg t2
        | (t0 as Const (@{const_name Let}, T0)) $ t1 $ t2 =>
          t0 $ t1 $ aux ss Ts js depth neg t2
        | Abs (s, T, t1) => Abs (s, T, aux ss Ts (incrs js) depth neg t1)
        | _ => t
      end
  in aux [] [] [] 0 false end

(* extended_context -> bool -> term
   -> ((term list * term list) * (bool * bool)) * term *)
fun preprocess_term (ext_ctxt as {skolem_depth, skolems, ...}) =
  unfold_defs_in_term ext_ctxt
  #> destroy_equalities
  #> skolemize_term skolems skolem_depth
  #> specialize_consts_in_term ext_ctxt 0
  #> `(axioms_for_term ext_ctxt)
  #> apfst (apfst (pairself (map unprefix_abs_vars)))
  #> apsnd unprefix_abs_vars

(* theory -> typ -> typ list -> typ list *)
fun add_ground_types thy T accum =
  case T of
    Type ("fun", Ts) => fold (add_ground_types thy) Ts accum
  | Type ("*", Ts) => fold (add_ground_types thy) Ts accum
  | Type (@{type_name itself}, [T1]) => add_ground_types thy T1 accum
  | @{typ prop} => accum
  | @{typ bool} => accum
  | @{typ unit} => accum
  | Type (_, Ts) =>
    if T mem accum then
      accum
    else
      T :: accum
      |> fold (add_ground_types thy)
              (case datatype_ctors thy T of [] => Ts | xs => map snd xs)
  | _ => insert (op =) T accum

(* theory -> typ -> typ list *)
fun ground_types_in_type thy T = add_ground_types thy T []
(* theory -> term list -> typ list *)
fun ground_types_in_terms thy ts =
  foldl (OldTerm.it_term_types (uncurry (add_ground_types thy))) [] ts

end;
