(*  Title:      Nitpick/Tools/nitpick_kodkod.ML
    Author:     Jasmin Blanchette, TU Muenchen
    Copyright   2008, 2009

Kodkod problem generator part of Kodkod.
*)

signature NITPICK_KODKOD =
sig
  type dtype_spec = NitpickScope.dtype_spec
  type kodkod_ctors = NitpickPeephole.kodkod_ctors
  type nut = NitpickNut.nut
  type nfa_transition = Kodkod.rel_expr * typ
  type nfa_entry = typ * nfa_transition list
  type nfa_table = nfa_entry list

  structure NameTable : TABLE

  val univ_card :
    int -> int -> int -> Kodkod.bound list -> Kodkod.formula -> int
  val check_arity : int -> int -> unit
  val kk_tuple : bool -> int -> int list -> Kodkod.tuple
  val sequential_int_bounds : int -> Kodkod.int_bound list
  val bounds_for_built_in_rels_in_formula :
    bool -> int -> int -> int -> int -> Kodkod.formula -> Kodkod.bound list
  val bound_for_plain_rel : nut -> Kodkod.bound
  val bound_for_ctor_rel : dtype_spec list -> nut -> Kodkod.bound
  val merge_bounds : Kodkod.bound list -> Kodkod.bound list
  val axiom_for_plain_rel : bool -> kodkod_ctors -> nut -> Kodkod.formula
  val axiom_for_ctor_rel :
    bool -> kodkod_ctors -> nut NameTable.table -> dtype_spec list -> nut
    -> Kodkod.formula
  val axioms_for_datatypes :
    int Typtab.table -> kodkod_ctors -> nut NameTable.table -> dtype_spec list
    -> Kodkod.formula list
  val kodkod_formula_from_nut :
    Proof.context -> dtype_spec list -> int Typtab.table -> bool -> bool
    -> kodkod_ctors -> nut -> Kodkod.formula
end;

structure NitpickKodkod : NITPICK_KODKOD =
struct

open NitpickUtil
open NitpickHOL
open NitpickScope
open NitpickPeephole
open NitpickRep
open NitpickNut

type nfa_transition = Kodkod.rel_expr * typ
type nfa_entry = typ * nfa_transition list
type nfa_table = nfa_entry list

structure NfaGraph = GraphFun(type key = typ val ord = TermOrd.typ_ord)

(* int -> int -> int -> Kodkod.bound list -> Kodkod.formula -> int *)
fun univ_card nat_card int_card main_j0 bounds formula =
  let
    (* Kodkod.rel_expr -> int -> int *)
    fun rel_expr_func r k =
      Int.max (k, case r of
                    Kodkod.Atom j => j + 1
                  | Kodkod.AtomSeq (k', j0) => j0 + k'
                  | _ => 0)
    (* Kodkod.tuple -> int -> int *)
    fun tuple_func t k =
      case t of
        Kodkod.Tuple js => map (curry (op +) 1) js |> foldl Int.max k
      | _ => k
    (* Kodkod.tuple_set -> int -> int *)
    fun tuple_set_func ts k =
      Int.max (k, case ts of Kodkod.TupleAtomSeq (k', j0) => j0 + k' | _ => 0)
    val expr_F = {formula_func = K I, rel_expr_func = rel_expr_func,
                  int_expr_func = K I}
    val tuple_F = {tuple_func = tuple_func, tuple_set_func = tuple_set_func}
    val formula_card = fold (Kodkod.fold_bound expr_F tuple_F) bounds 1
                       |> Kodkod.fold_formula expr_F formula
  in
    Int.max (main_j0 + foldl Int.max 0 [2, nat_card, int_card], formula_card)
  end

(* string -> rep -> string *)
fun bound_comment nick R = nick ^ " : " ^ string_for_rep R

(* int -> int -> unit *)
fun check_arity univ_card n =
  if n > Kodkod.max_arity univ_card then
    raise LIMIT ("check_arity",
                 "arity " ^ string_for_int n ^ " too large for universe of \
                 \cardinality " ^ string_for_int univ_card)
  else
    ()

(* bool -> int -> int list -> Kodkod.tuple *)
fun kk_tuple debug univ_card js =
  if debug then
    Kodkod.Tuple js
  else
    Kodkod.TupleIndex (length js,
                       fold (fn j => fn accum => accum * univ_card + j) js 0)

(* int -> Kodkod.int_bound list *)
fun sequential_int_bounds n =
  [(NONE,
    map (Kodkod.TupleSet o single o Kodkod.Tuple o single) (0 upto n - 1))]

(* Kodkod.formula -> Kodkod.n_ary_index list *)
fun built_in_rels_in_formula formula =
  let
    (* Kodkod.rel_expr -> Kodkod.n_ary_index list -> Kodkod.n_ary_index list *)
    fun rel_expr_func (Kodkod.Rel (n, j)) rels =
        (case AList.lookup (op =) (#rels initial_pool) n of
           SOME k => (j < k ? insert (op =) (n, j)) rels
         | NONE => rels)
      | rel_expr_func _ rels = rels
    val expr_F = {formula_func = K I, rel_expr_func = rel_expr_func,
                  int_expr_func = K I}
  in Kodkod.fold_formula expr_F formula [] end

val max_table_size = 65536

(* int -> unit *)
fun check_table_size k =
  if k > max_table_size then
    raise LIMIT ("check_table_size",
                 "precomputed table too large (" ^ string_for_int k ^ ")")
  else
    ()

(* bool -> int -> int * int -> (int -> int) -> Kodkod.tuple list *)
fun tabulate_func1 debug univ_card (k, j0) f =
  (check_table_size k;
   map_filter (fn j1 => let val j2 = f j1 in
                          if j2 >= 0 then
                            SOME (kk_tuple debug univ_card [j1 + j0, j2 + j0])
                          else
                            NONE
                        end)
              (0 upto k - 1))
(* bool -> int -> int * int -> int -> (int * int -> int) -> Kodkod.tuple list *)
fun tabulate_op2 debug univ_card (k, j0) res_j0 f =
  (check_table_size (k * k);
   map_filter (fn j => let
                         val j1 = j div k
                         val j2 = j - j1 * k
                         val j3 = f (j1, j2)
                       in
                         if j3 >= 0 then
                           SOME (kk_tuple debug univ_card
                                          [j1 + j0, j2 + j0, j3 + res_j0])
                         else
                           NONE
                       end)
              (0 upto k * k - 1))
(* bool -> int -> int * int -> int -> (int * int -> int * int)
   -> Kodkod.tuple list *)
fun tabulate_op2_2 debug univ_card (k, j0) res_j0 f =
  (check_table_size (k * k);
   map_filter (fn j => let
                         val j1 = j div k
                         val j2 = j - j1 * k
                         val (j3, j4) = f (j1, j2)
                       in
                         if j3 >= 0 andalso j4 >= 0 then
                           SOME (kk_tuple debug univ_card
                                          [j1 + j0, j2 + j0, j3 + res_j0,
                                           j4 + res_j0])
                         else
                           NONE
                       end)
              (0 upto k * k - 1))
(* bool -> int -> int * int -> (int * int -> int) -> Kodkod.tuple list *)
fun tabulate_nat_op2 debug univ_card (k, j0) f =
  tabulate_op2 debug univ_card (k, j0) j0 (atom_for_nat (k, 0) o f)
fun tabulate_int_op2 debug univ_card (k, j0) f =
  tabulate_op2 debug univ_card (k, j0) j0
               (atom_for_int (k, 0) o f o pairself (int_for_atom (k, 0)))
(* bool -> int -> int * int -> (int * int -> int * int) -> Kodkod.tuple list *)
fun tabulate_int_op2_2 debug univ_card (k, j0) f =
  tabulate_op2_2 debug univ_card (k, j0) j0
                 (pairself (atom_for_int (k, 0)) o f
                  o pairself (int_for_atom (k, 0)))

(* int * int -> int *)
fun isa_div (m, n) = m div n handle Div => 0
fun isa_mod (m, n) = m mod n handle Div => m
fun isa_gcd (m, 0) = m
  | isa_gcd (m, n) = isa_gcd (n, isa_mod (m, n))
fun isa_lcm (m, n) = isa_div (m * n, isa_gcd (m, n))
val isa_zgcd = isa_gcd o pairself abs
(* int * int -> int * int *)
fun isa_norm_frac (m, n) =
  if n < 0 then isa_norm_frac (~m, ~n)
  else if m = 0 orelse n = 0 then (0, 1)
  else let val p = isa_zgcd (m, n) in (isa_div (m, p), isa_div (n, p)) end

(* bool -> int -> int -> int -> int -> int * int
   -> string * bool * Kodkod.tuple list *)
fun tabulate_built_in_rel debug univ_card nat_card int_card j0 (x as (n, _)) =
  (check_arity univ_card n;
   if Kodkod.Rel x = not3_rel then
     ("not3", tabulate_func1 debug univ_card (2, j0) (curry (op -) 1))
   else if Kodkod.Rel x = eq3_rel then
     ("eq3", tabulate_op2 debug univ_card (univ_card, 0) j0
                          (int_for_bool o (op =)))
   else if Kodkod.Rel x = suc_rel then
     ("suc", tabulate_func1 debug univ_card (univ_card - j0 - 1, j0)
                            (curry (op +) 1))
   else if Kodkod.Rel x = nat_add_rel then
     ("nat_add", tabulate_nat_op2 debug univ_card (nat_card, j0) (op +))
   else if Kodkod.Rel x = int_add_rel then
     ("int_add", tabulate_int_op2 debug univ_card (int_card, j0) (op +))
   else if Kodkod.Rel x = nat_subtract_rel then
     ("nat_subtract",
      tabulate_op2 debug univ_card (nat_card, j0) j0 (op nat_minus))
   else if Kodkod.Rel x = int_subtract_rel then
     ("int_subtract", tabulate_int_op2 debug univ_card (int_card, j0) (op -))
   else if Kodkod.Rel x = nat_multiply_rel then
     ("nat_multiply", tabulate_nat_op2 debug univ_card (nat_card, j0) (op *))
   else if Kodkod.Rel x = int_multiply_rel then
     ("int_multiply", tabulate_int_op2 debug univ_card (int_card, j0) (op *))
   else if Kodkod.Rel x = nat_divide_rel then
     ("nat_divide", tabulate_nat_op2 debug univ_card (nat_card, j0) isa_div)
   else if Kodkod.Rel x = int_divide_rel then
     ("int_divide", tabulate_int_op2 debug univ_card (int_card, j0) isa_div)
   else if Kodkod.Rel x = nat_modulo_rel then
     ("nat_modulo", tabulate_nat_op2 debug univ_card (nat_card, j0) isa_mod)
   else if Kodkod.Rel x = int_modulo_rel then
     ("int_modulo", tabulate_int_op2 debug univ_card (int_card, j0) isa_mod)
   else if Kodkod.Rel x = int_less_rel then
     ("int_less", tabulate_int_op2 debug univ_card (int_card, j0)
                                   (int_for_bool o op <))
   else if Kodkod.Rel x = gcd_rel then
     ("gcd", tabulate_nat_op2 debug univ_card (nat_card, j0) isa_gcd)
   else if Kodkod.Rel x = lcm_rel then
     ("lcm", tabulate_nat_op2 debug univ_card (nat_card, j0) isa_lcm)
   else if Kodkod.Rel x = norm_frac_rel then
     ("norm_frac", tabulate_int_op2_2 debug univ_card (int_card, j0)
                                      isa_norm_frac)
   else
     raise ARG ("tabulate_built_in_rel", "unknown relation"))

(* bool -> int -> int -> int -> int -> int * int -> Kodkod.rel_expr
   -> Kodkod.bound *)
fun bound_for_built_in_rel debug univ_card nat_card int_card j0 x =
  let
    val (nick, ts) = tabulate_built_in_rel debug univ_card nat_card int_card
                                           j0 x
  in ([(x, nick)], [Kodkod.TupleSet ts]) end

(* bool -> int -> int -> int -> int -> Kodkod.formula -> Kodkod.bound list *)
fun bounds_for_built_in_rels_in_formula debug univ_card nat_card int_card j0 =
  map (bound_for_built_in_rel debug univ_card nat_card int_card j0)
  o built_in_rels_in_formula

(* rep -> Kodkod.tuple_set *)
val upper_bound_for_rep =
  foldl1 Kodkod.TupleProduct o map Kodkod.TupleAtomSeq o atom_schema_of_rep

(* nut -> Kodkod.bound *)
fun bound_for_plain_rel (FreeRel (x, _, R, nick)) =
    ([(x, bound_comment nick R)],
     [Kodkod.TupleSet [], upper_bound_for_rep R])
  | bound_for_plain_rel u = raise NUT ("bound_for_plain_rel", [u])

(* typ -> rep -> int -> Kodkod.tuple_set *)
fun upper_bound_for_self_rec_ctor_value T R delta =
  let
    val binder_Rs = binder_reps R
    val binder_atom_schema = atom_schema_of_reps binder_Rs
    val binder_type_schema = type_schema_of_reps (binder_types T) binder_Rs
    val body_T = body_type T
    val (k, j0) = the_single (atom_schema_of_rep (body_rep R))
    val rec_ts = Kodkod.TupleAtomSeq (delta, j0)
    val dom_tss = map2 (fn x => fn T => if T = body_T then rec_ts
                                        else Kodkod.TupleAtomSeq x)
                       binder_atom_schema binder_type_schema
  in
    foldr Kodkod.TupleProduct (Kodkod.TupleSet [Kodkod.Tuple [delta + j0]])
          dom_tss
  end

(* rep -> ctor_spec -> Kodkod.tuple_set *)
fun upper_bound_for_self_rec_ctor R {typ, delta, epsilon, ...} =
  let val delta = Int.max (1, delta) in
    if delta < epsilon then
      map (upper_bound_for_self_rec_ctor_value typ R) (delta upto epsilon - 1)
      |> foldl1 Kodkod.TupleUnion
    else
      Kodkod.TupleSet []
  end

(* dtype_spec list -> nut -> Kodkod.bound *)
fun bound_for_ctor_rel dtypes (FreeRel (x, T, R, nick)) =
    let
      val ctor as {delta, epsilon, exclusive, explicit_max, total, self_rec,
                   ...} = ctor_spec dtypes (nick, T)
    in
      ([(x, bound_comment nick R)],
       if explicit_max = 0 then
         [Kodkod.TupleSet []]
       else
         let
           val binders_R = Struct (binder_reps R)
           val j0 = snd (the_single (atom_schema_of_rep (body_rep R)))
         in
           if total then
             let
               val dom_jss = all_combinations_for_rep binders_R
               val ran_js = index_seq (delta + j0) (epsilon - delta)
             in
               if length dom_jss = epsilon - delta then
                 [Kodkod.TupleSet (map2 (fn dom_js => fn ran_j =>
                                            Kodkod.Tuple (dom_js @ [ran_j]))
                                        dom_jss ran_js)]
               else
                 raise ARG ("bound_for_ctor_rel",
                            "bad cardinalities (" ^
                            string_for_int (length dom_jss) ^ " vs. " ^
                            string_for_int (epsilon - delta) ^ ")")
             end
           else
             [Kodkod.TupleSet [],
              if self_rec then
                upper_bound_for_self_rec_ctor R ctor
              else
                let
                  val ts = Kodkod.TupleAtomSeq (epsilon - delta, delta + j0)
                in
                  if arity_of_rep binders_R = 0 then ts
                  else Kodkod.TupleProduct (upper_bound_for_rep binders_R, ts)
                end]
         end)
    end
  | bound_for_ctor_rel _ u = raise NUT ("bound_for_ctor_rel", [u])

(* Kodkod.bound -> Kodkod.rel_expr *)
fun first_rel_of_bound (zs, _) = Kodkod.Rel (fst (hd zs))

(* Kodkod.bound list -> Kodkod.bound list *)
fun merge_bounds bs =
  let
    (* Kodkod.bound -> int *)
    fun arity (zs, _) = fst (fst (hd zs))
    (* Kodkod.bound list -> Kodkod.bound -> Kodkod.bound list
       -> Kodkod.bound list *)
    fun add_bound ds b [] = List.revAppend (ds, [b])
      | add_bound ds b (c :: cs) =
        if arity b = arity c andalso snd b = snd c then
          List.revAppend (ds, (fst c @ fst b, snd c) :: cs)
        else
          add_bound (c :: ds) b cs
  in fold (add_bound []) bs [] end

(* int -> int -> Kodkod.rel_expr list *)
fun unary_var_seq j0 n = map (curry Kodkod.Var 1) (index_seq j0 n)

(* int list -> Kodkod.rel_expr *)
val singleton_from_combination = foldl1 Kodkod.Product o map Kodkod.Atom
(* rep -> Kodkod.rel_expr list *)
val all_singletons_for_rep =
  map singleton_from_combination o all_combinations_for_rep

(* Kodkod.rel_expr -> Kodkod.rel_expr list *)
fun unpack_products (Kodkod.Product (r1, r2)) =
    unpack_products r1 @ unpack_products r2
  | unpack_products r = [r]
fun unpack_joins (Kodkod.Join (r1, r2)) = unpack_joins r1 @ unpack_joins r2
  | unpack_joins r = [r]

(* rep -> Kodkod.rel_expr *)
val empty_rel_for_rep = empty_n_ary_rel o arity_of_rep
fun full_rel_for_rep R =
  case atom_schema_of_rep R of
    [] => raise REP ("full_rel_for_rep", [R])
  | schema => foldl1 Kodkod.Product (map Kodkod.AtomSeq schema)

(* int -> int list -> Kodkod.decl list *)
fun decls_for_atom_schema j0 schema =
  map2 (fn j => fn x => Kodkod.DeclOne ((1, j), Kodkod.AtomSeq x))
       (index_seq j0 (length schema)) schema

(* The type constraint below is a workaround for a Poly/ML bug. *)

(* kodkod_ctors -> rep -> Kodkod.rel_expr -> Kodkod.formula *)
fun d_n_ary_function ({kk_all, kk_join, kk_lone, kk_one, ...} : kodkod_ctors)
                     R r =
  let val body_R = body_rep R in
    if is_lone_rep body_R then
      let
        val binder_schema = atom_schema_of_reps (binder_reps R)
        val body_schema = atom_schema_of_rep body_R
        val one = is_one_rep body_R
        val opt_x = case r of Kodkod.Rel x => SOME x | _ => NONE
      in
        if opt_x <> NONE andalso length binder_schema = 1
           andalso length body_schema = 1 then
          (if one then Kodkod.Function else Kodkod.Functional)
              (the opt_x, Kodkod.AtomSeq (hd binder_schema),
               Kodkod.AtomSeq (hd body_schema))
        else
          let
            val decls = decls_for_atom_schema ~1 binder_schema
            val vars = unary_var_seq ~1 (length binder_schema)
            val kk_xone = if one then kk_one else kk_lone
          in kk_all decls (kk_xone (fold kk_join vars r)) end
      end
    else
      Kodkod.True
  end

(* kodkod_ctors -> rep -> rel_expr -> formula *)
fun s_n_ary_function kk R (r as Kodkod.Rel _) =
    if not (is_opt_rep R) then
      if r = suc_rel then
        Kodkod.False
      else if r = nat_add_rel then
        formula_for_bool (card_of_rep (body_rep R) = 1)
      else if r = nat_multiply_rel then
        formula_for_bool (card_of_rep (body_rep R) <= 2)
      else
        d_n_ary_function kk R r
    else if r = nat_subtract_rel then
      Kodkod.True
    else
      d_n_ary_function kk R r
  | s_n_ary_function kk R r = d_n_ary_function kk R r

(* bool -> kodkod_ctors -> rep -> rel_expr -> formula *)
fun kk_n_ary_function peephole_optim =
  if peephole_optim then s_n_ary_function else d_n_ary_function

(* int -> kodkod_ctors -> (Kodkod.rel_expr -> Kodkod.rel_expr)
   -> Kodkod.rel_expr -> Kodkod.rel_expr *)
fun basic_rel_let j kk f r =
  if inline_rel_expr r then
    f r
  else
    let val n = Kodkod.arity_of_rel_expr r in
      #kk_rel_let kk [Kodkod.AssignRelReg ((n, j), r)]
                     (f (Kodkod.RelReg (n, j)))
    end

(* kodkod_ctors -> (Kodkod.rel_expr -> Kodkod.rel_expr) -> Kodkod.rel_expr
   -> Kodkod.rel_expr *)
val single_rel_let = basic_rel_let 0
(* kodkod_ctors -> (Kodkod.rel_expr -> Kodkod.rel_expr -> Kodkod.rel_expr)
   -> Kodkod.rel_expr -> Kodkod.rel_expr -> Kodkod.rel_expr *)
fun double_rel_let kk f r1 r2 =
  single_rel_let kk (fn r1 => basic_rel_let 1 kk (f r1) r2) r1
(* kodkod_ctors
   -> (Kodkod.rel_expr -> Kodkod.rel_expr -> Kodkod.rel_expr -> Kodkod.rel_expr)
   -> Kodkod.rel_expr -> Kodkod.rel_expr -> Kodkod.rel_expr
   -> Kodkod.rel_expr *)
fun triple_rel_let kk f r1 r2 r3 =
  double_rel_let kk (fn r1 => fn r2 => basic_rel_let 2 kk (f r1 r2) r3) r1 r2

(* kodkod_ctors -> int -> Kodkod.formula -> Kodkod.rel_expr *)
fun atom_from_formula {kk_rel_if, ...} j0 f =
  kk_rel_if f (Kodkod.Atom (j0 + 1)) (Kodkod.Atom j0)
(* kodkod_ctors -> rep -> Kodkod.formula -> Kodkod.rel_expr *)
fun rel_expr_from_formula kk R f =
  case unopt_rep R of
    Atom (2, j0) => atom_from_formula kk j0 f
  | _ => raise REP ("rel_expr_from_formula", [R])

(* kodkod_cotrs -> int -> int -> Kodkod.rel_expr -> Kodkod.rel_expr list *)
fun unpack_vect_in_chunks {kk_project_seq, ...} chunk_arity num_chunks r =
  List.tabulate (num_chunks, fn j => kk_project_seq r (j * chunk_arity)
                                                    chunk_arity)

(* kodkod_ctors -> rep -> rep -> Kodkod.rel_expr -> Kodkod.rel_expr
   -> Kodkod.rel_expr *)
fun kk_n_fold_join
        (kk as {kk_intersect, kk_product, kk_join, kk_project_seq, ...}) R1
        res_R r1 r2 =
  case arity_of_rep R1 of
    1 => kk_join r1 r2
  | arity1 =>
    let
      val unpacked_rs1 =
        if inline_rel_expr r1 then unpack_vect_in_chunks kk 1 arity1 r1
        else unpack_products r1
    in
      if length unpacked_rs1 = arity1 then
        fold kk_join unpacked_rs1 r2
      else
        kk_project_seq
            (kk_intersect (kk_product r1 (full_rel_for_rep res_R)) r2)
            arity1 (arity_of_rep res_R)
    end

(* kodkod_ctors -> rep -> rep -> Kodkod.rel_expr -> Kodkod.rel_expr list
   -> Kodkod.rel_expr list -> Kodkod.rel_expr *)
fun kk_case_switch (kk as {kk_union, kk_product, ...}) R1 R2 r rs1 rs2 =
  if rs1 = rs2 then r
  else kk_n_fold_join kk R1 R2 r (fold1 kk_union (map2 kk_product rs1 rs2))

val lone_rep_fallback_max_card = 65536
val some_j0 = 0

(* kodkod_ctors -> rep -> rep -> Kodkod.rel_expr -> Kodkod.rel_expr *)
fun lone_rep_fallback kk new_R old_R r =
  if old_R = new_R then
    r
  else
    let val card = card_of_rep old_R in
      if is_lone_rep old_R andalso is_lone_rep new_R
         andalso card = card_of_rep new_R then
        if card > lone_rep_fallback_max_card then
          raise LIMIT ("lone_rep_fallback",
                       "too high cardinality (" ^ string_for_int card ^ ")")
        else
          kk_case_switch kk old_R new_R r (all_singletons_for_rep old_R)
                         (all_singletons_for_rep new_R)
      else
        raise REP ("lone_rep_fallback", [old_R, new_R])
    end
(* kodkod_ctors -> int * int -> rep -> Kodkod.rel_expr -> Kodkod.rel_expr *)
and atom_from_rel_expr kk (x as (k, j0)) old_R r =
  case old_R of
    Func (R1, R2) =>
    let
      val dom_card = card_of_rep R1
      val R2' = case R2 of Atom _ => R2 | _ => Atom (card_of_rep R2, some_j0)
    in
      atom_from_rel_expr kk x (Vect (dom_card, R2'))
                         (vect_from_rel_expr kk dom_card R2' old_R r)
    end
  | Opt _ => raise REP ("atom_from_rel_expr", [old_R])
  | _ => lone_rep_fallback kk (Atom x) old_R r
(* kodkod_ctors -> rep list -> rep -> Kodkod.rel_expr -> Kodkod.rel_expr *)
and struct_from_rel_expr kk Rs old_R r =
  case old_R of
    Atom _ => lone_rep_fallback kk (Struct Rs) old_R r
  | Struct Rs' =>
    let
      val Rs = filter (not_equal Unit) Rs
      val Rs' = filter (not_equal Unit) Rs'
    in
      if Rs' = Rs then
        r
      else if map card_of_rep Rs' = map card_of_rep Rs then
        let
          val old_arities = map arity_of_rep Rs'
          val old_offsets = offset_list old_arities
          val old_rs = map2 (#kk_project_seq kk r) old_offsets old_arities
        in
          fold1 (#kk_product kk)
                (map3 (rel_expr_from_rel_expr kk) Rs Rs' old_rs)
        end
      else
        lone_rep_fallback kk (Struct Rs) old_R r
    end
  | _ => raise REP ("struct_from_rel_expr", [old_R])
(* kodkod_ctors -> int -> rep -> rep -> Kodkod.rel_expr -> Kodkod.rel_expr *)
and vect_from_rel_expr kk k R old_R r =
  case old_R of
    Atom _ => lone_rep_fallback kk (Vect (k, R)) old_R r
  | Vect (k', R') =>
    if k = k' andalso R = R' then r
    else lone_rep_fallback kk (Vect (k, R)) old_R r
  | Func (R1, Formula Neutral) =>
    if k = card_of_rep R1 then
      fold1 (#kk_product kk)
            (map (fn arg_r =>
                     rel_expr_from_formula kk R (#kk_subset kk arg_r r))
                 (all_singletons_for_rep R1))
    else
      raise REP ("to_vect", [old_R])
  | Func (Unit, R2) => rel_expr_from_rel_expr kk R R2 r
  | Func (R1, R2) =>
    fold1 (#kk_product kk)
          (map (fn arg_r =>
                   rel_expr_from_rel_expr kk R R2
                                          (kk_n_fold_join kk R1 R2 arg_r r))
               (all_singletons_for_rep R1))
  | _ => raise REP ("vect_from_rel_expr", [old_R])
(* kodkod_ctors -> rep -> rep -> rep -> Kodkod.rel_expr -> Kodkod.rel_expr *)
and func_from_no_opt_rel_expr kk R1 R2 (Atom x) r =
    let
      val dom_card = card_of_rep R1
      val R2' = case R2 of Atom _ => R2 | _ => Atom (card_of_rep R2, some_j0)
    in
      func_from_no_opt_rel_expr kk R1 R2 (Vect (dom_card, R2'))
                                (vect_from_rel_expr kk dom_card R2' (Atom x) r)
    end
  | func_from_no_opt_rel_expr kk Unit R2 old_R r =
    (case old_R of
       Vect (k, R') => rel_expr_from_rel_expr kk R2 R' r
     | Func (Unit, R2') => rel_expr_from_rel_expr kk R2 R2' r
     | Func (Atom (1, _), Formula Neutral) =>
       (case unopt_rep R2 of
          Atom (2, j0) => atom_from_formula kk j0 (#kk_some kk r)
        | _ => raise REP ("func_from_no_opt_rel_expr",
                          [old_R, Func (Unit, R2)]))
     | Func (R1', R2') =>
       rel_expr_from_rel_expr kk R2 R2' (#kk_project_seq kk r (arity_of_rep R1')
                              (arity_of_rep R2'))
     | _ => raise REP ("func_from_no_opt_rel_expr", [old_R, Func (Unit, R2)]))
  | func_from_no_opt_rel_expr kk R1 (Formula Neutral) old_R r =
    (case old_R of
       Vect (k, Atom (2, j0)) =>
       let
         val args_rs = all_singletons_for_rep R1
         val vals_rs = unpack_vect_in_chunks kk 1 k r
         (* Kodkod.rel_expr -> Kodkod.rel_expr -> Kodkod.rel_expr *)
         fun empty_or_singleton_set_for arg_r val_r =
           #kk_join kk val_r (#kk_product kk (Kodkod.Atom (j0 + 1)) arg_r)
       in
         fold1 (#kk_union kk) (map2 empty_or_singleton_set_for args_rs vals_rs)
       end
     | Func (R1', Formula Neutral) =>
       if R1 = R1' then
         r
       else
         let
           val schema = atom_schema_of_rep R1
           val r1 = fold1 (#kk_product kk) (unary_var_seq ~1 (length schema))
                    |> rel_expr_from_rel_expr kk R1' R1
         in
           #kk_comprehension kk (decls_for_atom_schema ~1 schema)
                                (#kk_subset kk r1 r)
         end
     | Func (Unit, (Atom (2, j0))) =>
       #kk_rel_if kk (#kk_rel_eq kk r (Kodkod.Atom (j0 + 1)))
                  (full_rel_for_rep R1) (empty_rel_for_rep R1)
     | Func (R1', Atom (2, j0)) =>
       func_from_no_opt_rel_expr kk R1 (Formula Neutral)
           (Func (R1', Formula Neutral)) (#kk_join kk r (Kodkod.Atom (j0 + 1)))
     | _ => raise REP ("func_from_no_opt_rel_expr",
                       [old_R, Func (R1, Formula Neutral)]))
  | func_from_no_opt_rel_expr kk R1 R2 old_R r =
    case old_R of
      Vect (k, R) =>
      let
        val args_rs = all_singletons_for_rep R1
        val vals_rs = unpack_vect_in_chunks kk (arity_of_rep R) k r
                      |> map (rel_expr_from_rel_expr kk R2 R)
      in fold1 (#kk_union kk) (map2 (#kk_product kk) args_rs vals_rs) end
    | Func (R1', Formula Neutral) =>
      (case R2 of
         Atom (x as (2, j0)) =>
         let val schema = atom_schema_of_rep R1 in
           if length schema = 1 then
             #kk_override kk (#kk_product kk (Kodkod.AtomSeq (hd schema))
                                             (Kodkod.Atom j0))
                             (#kk_product kk r (Kodkod.Atom (j0 + 1)))
           else
             let
               val r1 = fold1 (#kk_product kk) (unary_var_seq ~1 (length schema))
                        |> rel_expr_from_rel_expr kk R1' R1
               val r2 = Kodkod.Var (1, ~(length schema) - 1)
               val r3 = atom_from_formula kk j0 (#kk_subset kk r1 r)
             in
               #kk_comprehension kk (decls_for_atom_schema ~1 (schema @ [x]))
                                 (#kk_rel_eq kk r2 r3)
             end
           end
         | _ => raise REP ("func_from_no_opt_rel_expr", [old_R, Func (R1, R2)]))
    | Func (Unit, R2') =>
      let val j0 = some_j0 in
        func_from_no_opt_rel_expr kk R1 R2 (Func (Atom (1, j0), R2'))
                                  (#kk_product kk (Kodkod.Atom j0) r)
      end
    | Func (R1', R2') =>
      if R1 = R1' andalso R2 = R2' then
        r
      else
        let
          val dom_schema = atom_schema_of_rep R1
          val ran_schema = atom_schema_of_rep R2
          val dom_prod = fold1 (#kk_product kk)
                               (unary_var_seq ~1 (length dom_schema))
                         |> rel_expr_from_rel_expr kk R1' R1
          val ran_prod = fold1 (#kk_product kk)
                               (unary_var_seq (~(length dom_schema) - 1)
                                              (length ran_schema))
                         |> rel_expr_from_rel_expr kk R2' R2
          val app = kk_n_fold_join kk R1' R2' dom_prod r
        in
          #kk_comprehension kk (decls_for_atom_schema ~1
                                                      (dom_schema @ ran_schema))
                               (#kk_subset kk ran_prod app)
        end
    | _ => raise REP ("func_from_no_opt_rel_expr", [old_R, Func (R1, R2)])
(* kodkod_ctors -> rep -> rep -> Kodkod.rel_expr -> Kodkod.rel_expr *)
and rel_expr_from_rel_expr kk new_R old_R r =
  let
    val unopt_old_R = unopt_rep old_R
    val unopt_new_R = unopt_rep new_R
  in
    if unopt_old_R <> old_R andalso unopt_new_R = new_R then
      raise REP ("rel_expr_from_rel_expr", [old_R, new_R])
    else if unopt_new_R = unopt_old_R then
      r
    else
      (case unopt_new_R of
         Atom x => atom_from_rel_expr kk x
       | Struct Rs => struct_from_rel_expr kk Rs
       | Vect (k, R') => vect_from_rel_expr kk k R'
       | Func (R1, R2) => func_from_no_opt_rel_expr kk R1 R2
       | _ => raise REP ("rel_expr_from_rel_expr", [old_R, new_R]))
          unopt_old_R r
  end
(* kodkod_ctors -> rep -> rep -> rep -> Kodkod.rel_expr -> Kodkod.rel_expr *)
and rel_expr_to_func kk R1 R2 = rel_expr_from_rel_expr kk (Func (R1, R2))

(* bool -> kodkod_ctors -> nut -> Kodkod.formula *)
fun axiom_for_plain_rel peephole_optim kk (FreeRel (x, _, R as Func _, nick)) =
    kk_n_ary_function peephole_optim kk R (Kodkod.Rel x)
  | axiom_for_plain_rel _ {kk_lone, kk_one, ...} (FreeRel (x, _, R, _)) =
    if is_one_rep R then kk_one (Kodkod.Rel x)
    else if is_lone_rep R andalso card_of_rep R > 1 then kk_lone (Kodkod.Rel x)
    else Kodkod.True
  | axiom_for_plain_rel _ _ u = raise NUT ("axiom_for_plain_rel", [u])

(* nut NameTable.table -> styp -> Kodkod.rel_expr * rep * int *)
fun ctor_triple rel_table (x as (s, T)) =
  case lookup_name rel_table (ConstName (s, T, Any, true)) of
    FreeRel ((n, j), _, R, _) => (Kodkod.Rel (n, j), R, n)
  | _ => raise TERM ("ctor_triple", [Const x])

(* kodkod_ctors -> nut NameTable.table -> styp -> Kodkod.rel_expr *)
fun ctor_range_rel_expr {kk_project, ...} rel_table x =
  let val (r, _, arity) = ctor_triple rel_table x in
    kk_project r [Kodkod.Num (arity - 1)]
  end

(* bool -> kodkod_ctors -> dtype_spec list -> nut -> Kodkod.formula *)
fun axiom_for_ctor_rel peephole_optim kk rel_table dtypes
                       (FreeRel (x as (n, j), T, R, nick)) =
    let
      val {delta, epsilon, exclusive, explicit_max, total, ...} =
        ctor_spec dtypes (nick, T)
      val honors_explicit_max =
        explicit_max < 0 orelse epsilon - delta <= explicit_max
    in
      if total orelse explicit_max = 0 then
        formula_for_bool honors_explicit_max
      else
        let
          val (k, j0) = atom_schema_of_rep (body_rep R) |> the_single
          val decl_axiom = kk_n_ary_function peephole_optim kk R (Kodkod.Rel x)
          val distinct_and_onto_axiom =
            if n = 1 then
              Kodkod.True
            else
              let
                val kk_xone = if exclusive then #kk_one kk else #kk_lone kk
              in
                #kk_all kk
                    [Kodkod.DeclOne ((1, 0),
                         Kodkod.AtomSeq (epsilon - delta, delta + j0))]
                    (kk_xone (#kk_join kk (Kodkod.Rel x) (Kodkod.Var (1, 0))))
              end
          val max_axiom =
            if honors_explicit_max then
              Kodkod.True
            else
              Kodkod.LE (Kodkod.Cardinality
                             (ctor_range_rel_expr kk rel_table (nick, T)),
                         Kodkod.Num explicit_max)
        in
          fold1 (#kk_and kk) [decl_axiom, distinct_and_onto_axiom, max_axiom]
        end
    end
  | axiom_for_ctor_rel _ _ _ _ u = raise NUT ("axiom_for_ctor_rel", [u])

(* kodkod_ctors -> nut NameTable.table -> dtype_spec list -> ctor_spec
   -> nfa_transition list *)
fun nfa_transitions_for_ctor {kk_project, ...} rel_table dtypes
                             {name, typ, ...} =
  let
    val (r, R, arity) = ctor_triple rel_table (name, typ)
    val binder_Rs = binder_reps R
    val binder_atom_schema = atom_schema_of_reps binder_Rs
    val binder_type_schema = type_schema_of_reps (binder_types typ) binder_Rs
    val body_T = body_type typ
  in
    map_filter (fn (T, j) =>
                   if exists (equal T o #typ) dtypes then
                     SOME (kk_project r (map Kodkod.Num [j, arity - 1]), T)
                   else
                     NONE)
               (binder_type_schema ~~ (0 upto length binder_type_schema - 1))
  end

(* kodkod_ctors -> nut NameTable.table -> dtype_spec list -> dtype_spec
   -> nfa_entry *)
fun nfa_entry_for_dtype kk rel_table dtypes {typ, ctors, ...} =
  (typ, maps (nfa_transitions_for_ctor kk rel_table dtypes) ctors)

val empty_rel = Kodkod.Product (Kodkod.None, Kodkod.None)

(* nfa_table -> typ -> typ -> Kodkod.rel_expr list *)
fun direct_path_rel_exprs nfa start final =
  case AList.lookup (op =) nfa final of
    SOME trans => map fst (filter (equal start o snd) trans)
  | NONE => []
(* kodkod_ctors -> nfa_table -> typ list -> typ -> typ -> Kodkod.rel_expr *)
and any_path_rel_expr {kk_union, ...} nfa [] start final =
    fold kk_union (direct_path_rel_exprs nfa start final)
         (if start = final then Kodkod.Iden else empty_rel)
  | any_path_rel_expr kk nfa (q :: qs) start final =
    #kk_union kk (any_path_rel_expr kk nfa qs start final)
                 (knot_path_rel_expr kk nfa qs start q final)
(* kodkod_ctors -> nfa_table -> typ list -> typ -> typ -> typ
   -> Kodkod.rel_expr *)
and knot_path_rel_expr kk nfa qs start knot final =
  #kk_join kk (any_path_rel_expr kk nfa qs start knot)
              (#kk_join kk (#kk_reflexive_closure kk
                                (loop_path_rel_expr kk nfa qs knot))
                           (any_path_rel_expr kk nfa qs knot final))
(* kodkod_ctors -> nfa_table -> typ list -> typ -> Kodkod.rel_expr *)
and loop_path_rel_expr {kk_union, ...} nfa [] start =
    fold kk_union (direct_path_rel_exprs nfa start start) empty_rel
  | loop_path_rel_expr kk nfa (q :: qs) start =
    if start = q then
      #kk_closure kk (loop_path_rel_expr kk nfa qs start)
    else
      #kk_union kk (loop_path_rel_expr kk nfa qs start)
                   (knot_path_rel_expr kk nfa qs start q start)

(* nfa_table -> unit NfaGraph.T *)
fun graph_for_nfa nfa =
  let
    (* typ -> unit NfaGraph.T -> unit NfaGraph.T *)
    fun new_node q = perhaps (try (NfaGraph.new_node (q, ())))
    (* nfa_table -> unit NfaGraph.T -> unit NfaGraph.T *)
    fun add_nfa [] = I
      | add_nfa ((_, []) :: nfa) = add_nfa nfa
      | add_nfa ((q, ((_, q') :: transitions)) :: nfa) =
        add_nfa ((q, transitions) :: nfa) o NfaGraph.add_edge (q, q') o
        new_node q' o new_node q
  in add_nfa nfa NfaGraph.empty end

(* nfa_table -> nfa_table list *)
fun strongly_connected_sub_nfas nfa =
  graph_for_nfa nfa
  |> NfaGraph.strong_conn
  |> map (fn keys => filter (member (op =) keys o fst) nfa)

(* dtype_spec list -> kodkod_ctors -> nfa_table -> typ -> Kodkod.formula *)
fun acyclicity_axiom_for_datatype dtypes kk nfa start =
  #kk_no kk (#kk_intersect kk
                 (loop_path_rel_expr kk nfa (map fst nfa) start) Kodkod.Iden)

(* kodkod_ctors -> (ctor_spec * Kodkod.rel_expr) list -> Kodkod.formula list *)
fun disjointness_axioms_for_ctors _ [] = []
  | disjointness_axioms_for_ctors (kk as {kk_rel_eq, kk_intersect, ...})
                                  ((r, ctor) :: zs) =
    map_filter (fn (r', ctor') =>
                   if #delta ctor >= #epsilon ctor'
                      orelse #delta ctor' >= #epsilon ctor then
                     NONE
                   else
                     SOME (kk_rel_eq (kk_intersect r r') Kodkod.None)) zs @
    disjointness_axioms_for_ctors kk zs

(* ctor_spec -> int *)
fun effective_ctor_max {delta, epsilon, ...} = epsilon - delta

(* int Typtab.table -> kodkod_ctors -> nut NameTable.table -> dtype_spec
   -> Kodkod.formula *)
fun completeness_axiom_for_datatype ofs (kk as {kk_rel_eq, kk_union, ...})
                                    rel_table {typ, card, ctors, ...} =
  if forall #exclusive ctors then
    foldl1 (op +) (map effective_ctor_max ctors) = card |> formula_for_bool
  else
    kk_rel_eq (fold kk_union (map (ctor_range_rel_expr kk rel_table
                                   o #name pairf #typ) ctors) Kodkod.None)
              (Kodkod.AtomSeq (card, offset_of_type ofs typ))

(* int Typtab.table -> kodkod_ctors -> nut NameTable.table -> dtype_spec list
   -> Kodkod.formula list *)
fun axioms_for_datatypes ofs kk rel_table dtypes =
  (map (nfa_entry_for_dtype kk rel_table dtypes) dtypes
   |> strongly_connected_sub_nfas
   |> filter (fn nfas => length nfas > 1)
   |> maps (fn nfa => map (acyclicity_axiom_for_datatype dtypes kk nfa o fst)
                          nfa)) @
  maps (disjointness_axioms_for_ctors kk
        o map (`(ctor_range_rel_expr kk rel_table o #name pairf #typ)) o #ctors)
       dtypes @
  map (completeness_axiom_for_datatype ofs kk rel_table) dtypes

(* Proof.context -> dtype_spec list -> int Typtab.table -> bool -> bool
   -> kodkod_ctors -> nut -> Kodkod.formula *)
fun kodkod_formula_from_nut ctxt dtypes ofs liberal peephole_optim kk u =
  let
    val {kk_all, kk_exist, kk_formula_let, kk_formula_if, kk_or, kk_not, kk_iff,
         kk_implies, kk_and, kk_subset, kk_rel_eq, kk_no, kk_some, kk_rel_let,
         kk_rel_if, kk_union, kk_difference, kk_intersect, kk_product, kk_join,
         kk_closure, kk_comprehension, kk_project, kk_project_seq, kk_not3,
         kk_nat_less, kk_int_less, ...} = kk
    val main_j0 = offset_of_type ofs bool_T
    val bool_j0 = main_j0
    val bool_atom_R = Atom (2, main_j0)
    val false_atom = Kodkod.Atom bool_j0
    val true_atom = Kodkod.Atom (bool_j0 + 1)

    (* occurrence -> int -> Kodkod.rel_expr -> Kodkod.formula *)
    fun formula_from_opt_atom occ j0 r =
      case occ of
        Negative => kk_not (kk_rel_eq r (Kodkod.Atom j0))
      | _ => kk_rel_eq r (Kodkod.Atom (j0 + 1))
    (* int -> Kodkod.rel_expr -> Kodkod.formula *)
    val formula_from_atom = formula_from_opt_atom Positive

    (* Kodkod.formula -> Kodkod.formula -> Kodkod.formula *)
    fun kk_notimplies f1 f2 = kk_and f1 (kk_not f2)
    (* Kodkod.rel_expr -> Kodkod.rel_expr -> Kodkod.rel_expr *)
    val kk_or3 =
      double_rel_let kk
          (fn r1 => fn r2 =>
              kk_rel_if (kk_subset true_atom (kk_union r1 r2)) true_atom
                        (kk_intersect r1 r2))
    val kk_and3 =
      double_rel_let kk
          (fn r1 => fn r2 =>
              kk_rel_if (kk_subset false_atom (kk_union r1 r2)) false_atom
                        (kk_intersect r1 r2))
    fun kk_notimplies3 r1 r2 = kk_and3 r1 (kk_not3 r2)

    (* int -> Kodkod.rel_expr -> Kodkod.formula list *)
    val unpack_formulas =
      map (formula_from_atom bool_j0) oo unpack_vect_in_chunks kk 1
    (* (Kodkod.formula -> Kodkod.formula -> Kodkod.formula) -> int
       -> Kodkod.rel_expr -> Kodkod.rel_expr -> Kodkod.rel_expr *)
    fun kk_vect_set_op connective k r1 r2 =
      fold1 kk_product (map2 (atom_from_formula kk bool_j0 oo connective)
                             (unpack_formulas k r1) (unpack_formulas k r2))
    (* (Kodkod.formula -> Kodkod.formula -> Kodkod.formula) -> int
       -> Kodkod.rel_expr -> Kodkod.rel_expr -> Kodkod.formula *)
    fun kk_vect_set_bool_op connective k r1 r2 =
      fold1 kk_and (map2 connective (unpack_formulas k r1)
                         (unpack_formulas k r2))

    (* nut -> Kodkod.formula *)
    fun to_f u =
      case rep_of u of
        Formula occ =>
        (case u of
           Cst (False, _, _) => Kodkod.False
         | Cst (True, _, _) => Kodkod.True
         | Op1 (Not, _, _, u1) => kk_not (to_f u1)
         | Op1 (Finite, _, _, u1) =>
           let val opt1 = is_opt_rep (rep_of u1) in
             case occ of
               Neutral => if opt1 then raise NUT ("to_f (Finite)", [u])
                          else Kodkod.True
             | Positive => formula_for_bool (not opt1)
             | Negative => Kodkod.True
           end
         | Op1 (Cast, _, _, u1) => to_f_with_occurrence occ u1
         | Op2 (All, _, _, u1, u2) => kk_all (untuple to_decl u1) (to_f u2)
         | Op2 (Exist, _, _, u1, u2) => kk_exist (untuple to_decl u1) (to_f u2)
         | Op2 (Or, _, _, u1, u2) => kk_or (to_f u1) (to_f u2)
         | Op2 (Implies, _, _, u1, u2) => kk_implies (to_f u1) (to_f u2)
         | Op2 (And, _, _, u1, u2) => kk_and (to_f u1) (to_f u2)
         | Op2 (DefEq, _, _, u1, u2) =>
           (case min_rep (rep_of u1) (rep_of u2) of
              Unit => Kodkod.True
            | Formula occ =>
              kk_iff (to_f_with_occurrence occ u1) (to_f_with_occurrence occ u2)
            | min_R =>
              let
                (* nut -> nut list *)
                fun args (Op2 (Apply, _, _, u1, u2)) = u2 :: args u1
                  | args _ = []
                val opt_arg_us = filter (is_opt_rep o rep_of) (args u1)
              in
                fold (kk_or o (kk_no o to_r)) opt_arg_us
                     (kk_rel_eq (to_rep min_R u1) (to_rep min_R u2))
              end)
         | Op2 (Eq, T, R, u1, u2) =>
           (case min_rep (rep_of u1) (rep_of u2) of
              Unit => Kodkod.True
            | Formula occ =>
              kk_iff (to_f_with_occurrence occ u1) (to_f_with_occurrence occ u2)
            | min_R =>
              if is_opt_rep min_R then
                if occ = Neutral then
                  raise NUT ("to_f (Eq)", [u])
                else if is_Cst Biggie u1 then
                  to_could_be_biggie (occ = Negative) u2
                else if is_Cst Biggie u2 then
                  to_could_be_biggie (occ = Negative) u1
                else
                  let
                    val r1 = to_rep min_R u1
                    val r2 = to_rep min_R u2
                  in
                    (if occ = Positive then
                       if exists (not o is_opt_rep o rep_of) [u1, u2] then
                         kk_rel_eq r1 r2
                       else if is_one_rep min_R
                               andalso arity_of_rep min_R = 1 then
                         kk_rel_eq (fold kk_join [r1, r2]
                                         (kk_product Kodkod.Iden false_atom))
                                   false_atom
                       else
                         raise SAME
                     else
                       if is_one_rep min_R andalso arity_of_rep min_R = 1 then
                         kk_subset (kk_product r1 r2) Kodkod.Iden
                       else
                         raise SAME)
                    handle SAME =>
                           formula_from_opt_atom occ bool_j0
                               (to_guard [u1, u2] bool_atom_R
                                         (rel_expr_from_formula kk bool_atom_R
                                                            (kk_rel_eq r1 r2)))
                  end
              else
                let
                  val r1 = to_rep min_R u1
                  val r2 = to_rep min_R u2
                in
                  if is_one_rep min_R then
                    let
                      val rs1 = unpack_products r1
                      val rs2 = unpack_products r2
                    in
                      if length rs1 = length rs2
                         andalso map Kodkod.arity_of_rel_expr rs1
                                 = map Kodkod.arity_of_rel_expr rs2 then
                        fold1 kk_and (map2 kk_subset rs1 rs2)
                      else
                        kk_subset r1 r2
                    end
                  else
                    kk_rel_eq r1 r2
                end)
         | Op2 (Less, T, Formula occ, u1, u2) =>
           formula_from_opt_atom occ bool_j0
               (to_r (Op2 (Less, T, Opt bool_atom_R, u1, u2)))
         | Op2 (Subset, _, _, u1, u2) =>
           let
             val dom_T = domain_type (type_of u1)
             val R1 = rep_of u1
             val R2 = rep_of u2
             val set_R as Func (dom_R, ran_R) =
               case min_rep R1 R2 of
                 Func (Unit, R') =>
                 let val j0 = offset_of_type ofs (domain_type (type_of u1)) in
                   Func (Atom (1, j0), R')
                 end
               | R as Func _ => R
               | R => Func (Atom (card_of_domain_from_rep 2 R,
                                  offset_of_type ofs dom_T),
                            if is_opt_rep R then Opt bool_atom_R
                            else Formula Neutral)
           in
             if not (is_opt_rep ran_R) then
               to_set_bool_op kk_implies kk_subset u1 u2
             else if occ = Neutral then
               raise NUT ("to_f (Subset)", [u])
             else
               let
                 (* bool -> nut -> Kodkod.rel_expr *)
                 fun set_to_r widen u =
                   if widen then
                     kk_difference (full_rel_for_rep dom_R)
                                   (kk_join (to_rep set_R u) false_atom)
                   else
                     kk_join (to_rep set_R u) true_atom
                 val widen1 = (occ = Positive andalso is_opt_rep R1)
                 val widen2 = (occ = Negative andalso is_opt_rep R2)
               in kk_subset (set_to_r widen1 u1) (set_to_r widen2 u2) end
           end
         | Op2 (Apply, T, _, u1, u2) =>
           to_f_with_occurrence occ
               (Op2 (Apply, T, Opt (Atom (2, offset_of_type ofs T)), u1, u2))
         | Op3 (Let, _, _, u1, u2, u3) =>
           kk_formula_let [to_expr_assign u1 u2] (to_f u3)
         | Op3 (If, _, _, u1, u2, u3) =>
           kk_formula_if (to_f u1) (to_f u2) (to_f u3)
         | FormulaReg (j, _, _) => Kodkod.FormulaReg j
         | _ => raise NUT ("to_f", [u]))
      | Atom (2, j0) => formula_from_atom j0 (to_r u)
      | _ => raise NUT ("to_f", [u])
    (* occurrence -> nut -> Kodkod.formula *)
    and to_f_with_occurrence occ u =
      case rep_of u of
        Formula _ => to_f u
      | Atom (2, j0) => formula_from_atom j0 (to_r u)
      | Opt (Atom (2, j0)) => formula_from_opt_atom occ j0 (to_r u)
      | _ => raise NUT ("to_f_with_occurrence", [u])
    (* nut -> Kodkod.rel_expr *)
    and to_r u =
      case u of
        Cst (False, _, Atom _) => false_atom
      | Cst (True, _, Atom _) => true_atom
      | Cst (Iden, T, Func (Struct [R1, R2], Formula Neutral)) =>
        if R1 = R2 andalso arity_of_rep R1 = 1 then
          kk_intersect Kodkod.Iden (kk_product (full_rel_for_rep R1)
                                               Kodkod.Univ)
        else
          let
            val schema1 = atom_schema_of_rep R1
            val schema2 = atom_schema_of_rep R2
            val arity1 = length schema1
            val arity2 = length schema2
            val r1 = fold1 kk_product (unary_var_seq 0 arity1)
            val r2 = fold1 kk_product (unary_var_seq arity1 arity2)
            val min_R = min_rep R1 R2
          in
            kk_comprehension
                (decls_for_atom_schema 0 (schema1 @ schema2))
                (kk_rel_eq (rel_expr_from_rel_expr kk min_R R1 r1)
                           (rel_expr_from_rel_expr kk min_R R2 r2))
          end
      | Cst (Iden, T, Func (Atom (1, j0), Formula Neutral)) => Kodkod.Atom j0
      | Cst (Iden, T, Func (R1, R2)) =>
        to_rep (Func (R1, R2)) (Cst (Iden, T, Func (R1, Formula Neutral)))
      | Cst (Num j, @{typ int}, R) =>
         (case atom_for_int (card_of_rep R, offset_of_type ofs int_T) j of
            ~1 => if is_opt_rep R then Kodkod.None
                  else raise NUT ("to_r (Num)", [u])
          | j' => Kodkod.Atom j')
      | Cst (Num j, T, R) =>
         if j < card_of_rep R then Kodkod.Atom (j + offset_of_type ofs T)
         else if is_opt_rep R then Kodkod.None
         else raise NUT ("to_r", [u])
      | Cst (Biggie, _, R) => empty_rel_for_rep R
      | Cst (Suc, _, Func (Atom x, _)) =>
        kk_intersect suc_rel (kk_product Kodkod.Univ (Kodkod.AtomSeq x))
      | Cst (Add, Type ("fun", [@{typ nat}, _]), _) => nat_add_rel
      | Cst (Add, Type ("fun", [@{typ int}, _]), _) => int_add_rel
      | Cst (Subtract, Type ("fun", [@{typ nat}, _]), _) => nat_subtract_rel
      | Cst (Subtract, Type ("fun", [@{typ int}, _]), _) => int_subtract_rel
      | Cst (Multiply, Type ("fun", [@{typ nat}, _]), _) => nat_multiply_rel
      | Cst (Multiply, Type ("fun", [@{typ int}, _]), _) => int_multiply_rel
      | Cst (Divide, Type ("fun", [@{typ nat}, _]), _) => nat_divide_rel
      | Cst (Divide, Type ("fun", [@{typ int}, _]), _) => int_divide_rel
      | Cst (Modulo, Type ("fun", [@{typ nat}, _]), _) => nat_modulo_rel
      | Cst (Modulo, Type ("fun", [@{typ int}, _]), _) => int_modulo_rel
      | Cst (Gcd, _, _) => gcd_rel
      | Cst (Lcm, _, _) => lcm_rel
      | Cst (Fracs, _, Func (Atom (1, _), _)) => Kodkod.None
      | Cst (Fracs, _, Func (Struct _, _)) =>
        kk_project_seq norm_frac_rel 2 2
      | Cst (NormFrac, _, _) => norm_frac_rel
      | Cst (NatToInt, _, Func (Atom _, Atom _)) => Kodkod.Iden
      | Cst (NatToInt, _,
             Func (Atom (nat_k, nat_j0), Opt (Atom (int_k, int_j0)))) =>
        if nat_j0 = int_j0 then
          kk_intersect Kodkod.Iden
              (kk_product (Kodkod.AtomSeq (max_int_for_card int_k + 1, nat_j0))
                          Kodkod.Univ)
        else
          raise BAD ("to_r (NatToInt)", "\"nat_j0 <> int_j0\"")
      | Cst (IntToNat, _, Func (Atom (int_k, int_j0), nat_R)) =>
        let
          val abs_card = max_int_for_card int_k + 1
          val (nat_k, nat_j0) = the_single (atom_schema_of_rep nat_R)
          val overlap = Int.min (nat_k, abs_card)
        in
          if nat_j0 = int_j0 then
            kk_union (kk_product (Kodkod.AtomSeq (int_k - abs_card,
                                                  int_j0 + abs_card))
                                 (Kodkod.Atom nat_j0))
                     (kk_intersect Kodkod.Iden
                          (kk_product (Kodkod.AtomSeq (overlap, int_j0))
                                      Kodkod.Univ))
          else
            raise BAD ("to_r (IntToNat)", "\"nat_j0 <> int_j0\"")
        end
      | Op1 (Not, _, R, u1) => kk_not3 (to_rep R u1)
      | Op1 (Finite, _, Opt (Atom _), _) => Kodkod.None
      | Op1 (Converse, T, R, u1) =>
        let
          val (b_T, a_T) = HOLogic.dest_prodT (domain_type T)
          val (b_R, a_R) =
            case R of
              Func (Struct [R1, R2], _) => (R1, R2)
            | Func (R1, _) =>
              if card_of_rep R1 <> 1 then raise REP ("to_r (Converse)", [R])
              else pairself (Atom o pair 1 o offset_of_type ofs) (b_T, a_T)
          val body_R = body_rep R
          val a_arity = arity_of_rep a_R
          val b_arity = arity_of_rep b_R
          val ab_arity = a_arity + b_arity
          val body_arity = arity_of_rep body_R
        in
          kk_project (to_rep (Func (Struct [a_R, b_R], body_R)) u1)
              (map Kodkod.Num ((a_arity upto ab_arity - 1) @
                               (0 upto a_arity - 1) @
                               (ab_arity upto ab_arity + body_arity - 1)))
          |> rel_expr_from_rel_expr kk R (Func (Struct [b_R, a_R], body_R))
        end
      | Op1 (Closure, _, R, u1) =>
        if is_opt_rep R then
          let
            val T1 = type_of u1
            val R' = rep_to_binary_rel_rep ofs T1 (unopt_rep (rep_of u1))
            val R'' = opt_rep ofs T1 R'
            val r1 = kk_join (to_rep R'' u1) true_atom
          in
            rel_expr_from_rel_expr kk R R''
                                   (kk_product (kk_closure r1) true_atom)
          end
        else
          let val R' = rep_to_binary_rel_rep ofs (type_of u1) (rep_of u1) in
            rel_expr_from_rel_expr kk R R' (kk_closure (to_rep R' u1))
          end
      | Op1 (SingletonSet, _, Func (R1, Opt _), Cst (Biggie, _, _)) =>
        (if R1 = Unit then I else kk_product (full_rel_for_rep R1)) false_atom
      | Op1 (SingletonSet, _, R, u1) =>
        (case R of
           Func (R1, Formula Neutral) => to_rep R1 u1
         | Func (Unit, Opt R) => to_guard [u1] R true_atom
         | Func (R1, R2 as Opt _) =>
           single_rel_let kk
               (fn r => kk_rel_if (kk_no r) (empty_rel_for_rep R)
                            (rel_expr_to_func kk R1 bool_atom_R
                                              (Func (R1, Formula Neutral)) r))
               (to_opt R1 u1)
         | _ => raise NUT ("to_r (SingletonSet)", [u]))
      | Op1 (First, T, R, u1) => to_nth_pair_dtor 0 T R u1
      | Op1 (Second, T, R, u1) => to_nth_pair_dtor 1 T R u1
      | Op1 (RepOfAbs, T, Func (R1, R2), u1) =>
        let val (arity1, arity2) = pairself arity_of_rep (R1, R2) in
          kk_project (to_r u1)
                     (map Kodkod.Num ((arity2 upto arity2 + arity1 - 1) @
                                      (0 upto arity2 - 1)))
        end
      | Op1 (Cast, _, R, u1) =>
        ((case rep_of u1 of
            Formula _ =>
            (case unopt_rep R of
               Atom (2, j0) => atom_from_formula kk j0 (to_f u1)
             | _ => raise SAME)
          | _ => raise SAME)
         handle SAME => rel_expr_from_rel_expr kk R (rep_of u1) (to_r u1))
      | Op2 (All, T, R as Opt _, u1, u2) =>
        to_r (Op1 (Not, T, R,
                   Op2 (Exist, T, R, u1, Op1 (Not, T, rep_of u2, u2))))
      | Op2 (Exist, T, Opt _, u1, u2) =>
        let val rs1 = untuple to_decl u1 in
          if not (is_opt_rep (rep_of u2)) then
            kk_rel_if (kk_exist rs1 (to_f u2)) true_atom Kodkod.None
          else
            let val r2 = to_r u2 in
              kk_rel_if (kk_exist rs1 (kk_rel_eq r2 true_atom)) true_atom
                        (kk_rel_if (kk_all rs1 (kk_rel_eq r2 false_atom))
                                   false_atom Kodkod.None)
            end
        end
      | Op2 (Or, _, _, u1, u2) =>
        if is_opt_rep (rep_of u1) then kk_rel_if (to_f u2) true_atom (to_r u1)
        else kk_rel_if (to_f u1) true_atom (to_r u2)
      | Op2 (Implies, _, _, u1, u2) =>
        if is_opt_rep (rep_of u1) then
          kk_rel_if (to_f u2) true_atom (kk_not3 (to_r u1))
        else
          kk_rel_if (to_f u1) (to_r u2) true_atom
      | Op2 (And, _, _, u1, u2) =>
        if is_opt_rep (rep_of u1) then kk_rel_if (to_f u2) (to_r u1) false_atom
        else kk_rel_if (to_f u1) (to_r u2) false_atom
      | Op2 (Less, _, _, u1, u2) =>
        if type_of u1 = nat_T then
          if is_Cst Biggie u1 then to_compare_with_biggie u2 false_atom
          else if is_Cst Biggie u2 then to_compare_with_biggie u1 true_atom
          else kk_nat_less (to_integer u1) (to_integer u2)
        else
          kk_int_less (to_integer u1) (to_integer u2)
      | Op2 (Triad, T, Opt (Atom (2, j0)), u1, u2) =>
        let
          val f1 = to_f u1
          val f2 = to_f u2
        in
          if f1 = f2 then atom_from_formula kk j0 f1
          else kk_rel_if f1 true_atom (kk_rel_if f2 Kodkod.None false_atom)
        end
      | Op2 (Union, _, R, u1, u2) =>
        to_set_op kk_or kk_or3 kk_union kk_union kk_intersect false R u1 u2
      | Op2 (SetDifference, _, R, u1, u2) =>
        to_set_op kk_notimplies kk_notimplies3 kk_difference kk_intersect
                  kk_union true R u1 u2
      | Op2 (Intersect, _, R, u1, u2) =>
        to_set_op kk_and kk_and3 kk_intersect kk_intersect kk_union false R
                  u1 u2
      | Op2 (Composition, _, R, u1, u2) =>
        let
          val (a_T, b_T) = HOLogic.dest_prodT (domain_type (type_of u2))
          val (_, c_T) = HOLogic.dest_prodT (domain_type (type_of u1))
          val ab_k = card_of_domain_from_rep 2 (rep_of u2)
          val bc_k = card_of_domain_from_rep 2 (rep_of u1)
          val ac_k = card_of_domain_from_rep 2 R
          val a_k = exact_root 2 (ac_k * ab_k div bc_k)
          val b_k = exact_root 2 (ab_k * bc_k div ac_k)
          val c_k = exact_root 2 (bc_k * ac_k div ab_k)
          val a_R = Atom (a_k, offset_of_type ofs a_T)
          val b_R = Atom (b_k, offset_of_type ofs b_T)
          val c_R = Atom (c_k, offset_of_type ofs c_T)
          val body_R = body_rep R
        in
          (case body_R of
             Formula Neutral =>
             kk_join (to_rep (Func (Struct [a_R, b_R], Formula Neutral)) u2)
                     (to_rep (Func (Struct [b_R, c_R], Formula Neutral)) u1)
           | Opt (Atom (2, _)) =>
             let
               (* Kodkod.rel_expr -> rep -> rep -> nut -> Kodkod.rel_expr *)
               fun do_nut r R1 R2 u =
                 kk_join (to_rep (Func (Struct [R1, R2], body_R)) u) r
               (* Kodkod.rel_expr -> Kodkod.rel_expr *)
               fun do_term r =
                 kk_product (kk_join (do_nut r a_R b_R u2)
                                     (do_nut r b_R c_R u1)) r
             in kk_union (do_term true_atom) (do_term false_atom) end
           | _ => raise NUT ("to_r (Composition)", [u]))
          |> rel_expr_from_rel_expr kk R (Func (Struct [a_R, c_R], body_R))
        end
      | Op2 (Product, T, R, u1, u2) =>
        let
          val (a_T, b_T) = HOLogic.dest_prodT (domain_type T)
          val a_k = card_of_domain_from_rep 2 (rep_of u1)
          val b_k = card_of_domain_from_rep 2 (rep_of u2)
          val a_R = Atom (a_k, offset_of_type ofs a_T)
          val b_R = Atom (b_k, offset_of_type ofs b_T)
          val body_R = body_rep R
        in
          (case body_R of
             Formula Neutral =>
             kk_product (to_rep (Func (a_R, Formula Neutral)) u1)
                        (to_rep (Func (b_R, Formula Neutral)) u2)
           | Opt (Atom (2, _)) =>
             let
               (* Kodkod.rel_expr -> rep -> nut -> Kodkod.rel_expr *)
               fun do_nut r R u = kk_join (to_rep (Func (R, body_R)) u) r
               (* Kodkod.rel_expr -> Kodkod.rel_expr *)
               fun do_term r =
                 kk_product (kk_product (do_nut r a_R u1) (do_nut r b_R u2)) r
             in kk_union (do_term true_atom) (do_term false_atom) end
           | _ => raise NUT ("to_r (Product)", [u]))
          |> rel_expr_from_rel_expr kk R (Func (Struct [a_R, b_R], body_R))
        end
      | Op2 (Image, T, R, u1, u2) =>
        (case (rep_of u1, rep_of u2) of
           (Func (R11, R12), Func (R21, Formula Neutral)) =>
           if R21 = R11 andalso is_lone_rep R12 then
             let
               (* Kodkod.rel_expr -> Kodkod.rel_expr *)
               fun big_join r = kk_n_fold_join kk R21 R12 r (to_r u1)
               val core_r = big_join (to_r u2)
               val core_R = Func (R12, Formula Neutral)
             in
               if is_opt_rep R12 then
                 let
                   val schema = atom_schema_of_rep R21
                   val decls = decls_for_atom_schema ~1 schema
                   val vars = unary_var_seq ~1 (length decls)
                   val f = kk_some (big_join (fold1 kk_product vars))
                 in
                   kk_rel_if (kk_all decls f)
                             (rel_expr_from_rel_expr kk R core_R core_r)
                             (rel_expr_from_rel_expr kk R (opt_rep ofs T core_R)
                                              (kk_product core_r true_atom))
                 end
               else
                 rel_expr_from_rel_expr kk R core_R core_r
             end
           else
             raise NUT ("to_r (Image)", [u1, u2])
         | _ => raise NUT ("to_r (Image)", [u1, u2]))
      | Op2 (Apply, @{typ nat}, _,
             Op2 (Apply, _, _, Cst (Subtract, _, _), u1), u2) =>
        if is_Cst Biggie u2 andalso not (is_opt_rep (rep_of u1)) then
          Kodkod.Atom (offset_of_type ofs nat_T)
        else
          fold kk_join [to_integer u1, to_integer u2] nat_subtract_rel
      | Op2 (Apply, _, R, u1, u2) =>
        if is_Cst Biggie u2 andalso is_set_type (type_of u1)
           andalso not (is_opt_rep (rep_of u1)) then
          false_atom
        else
          to_apply R u1 u2
      | Op2 (Abstraction, T, R as Opt (Atom (1, j0)), u1, u2) =>
        to_guard [u1, u2] R (Kodkod.Atom j0)
      | Op2 (Abstraction, T, Func (_, Formula Neutral), u1, u2) =>
        kk_comprehension (untuple to_decl u1) (to_f u2)
      | Op2 (Abstraction, T, Func (_, R2), u1, u2) =>
        let
          val dom_decls = untuple to_decl u1
          val ran_schema = atom_schema_of_rep R2
          val ran_decls = decls_for_atom_schema ~1 ran_schema
          val ran_vars = unary_var_seq ~1 (length ran_decls)
        in
          kk_comprehension (dom_decls @ ran_decls)
                           (kk_subset (fold1 kk_product ran_vars)
                                      (to_rep R2 u2))
        end
      | Op3 (Let, _, R, u1, u2, u3) =>
        kk_rel_let [to_expr_assign u1 u2] (to_rep R u3)
      | Op3 (If, _, R, u1, u2, u3) =>
        if is_opt_rep (rep_of u1) then
          triple_rel_let kk
              (fn r1 => fn r2 => fn r3 =>
                  kk_rel_if (kk_some r1)
                            (kk_rel_if (formula_from_atom bool_j0 r1) r2 r3)
                            (kk_rel_if (kk_rel_eq r2 r3) r2
                                       (empty_rel_for_rep R)))
              (to_r u1) (to_rep R u2) (to_rep R u3)
        else
          kk_rel_if (to_f u1) (to_rep R u2) (to_rep R u3)
      | Tuple (_, R, us) =>
        (case unopt_rep R of
           Struct Rs => to_product Rs us
         | Vect (k, R) => to_product (replicate k R) us
         | Atom (1, j0) =>
           (case filter (not_equal Unit o rep_of) us of
              [] => Kodkod.Atom j0
            | us' =>
              kk_rel_if (kk_some (fold1 kk_product (map to_r us')))
                        (Kodkod.Atom j0) Kodkod.None)
         | _ => raise NUT ("to_r (Tuple)", [u]))
      | BoundRel (x, _, _, _) => Kodkod.Var x
      | FreeRel (x, _, _, _) => Kodkod.Rel x
      | RelReg (j, _, R) => Kodkod.RelReg (arity_of_rep R, j)
      | u => raise NUT ("to_r", [u])
    (* nut -> Kodkod.decl *)
    and to_decl (BoundRel (x, _, R, _)) =
        Kodkod.DeclOne (x, Kodkod.AtomSeq (the_single (atom_schema_of_rep R)))
      | to_decl u = raise NUT ("to_decl", [u])
    (* nut -> Kodkod.expr_assign *)
    and to_expr_assign (FormulaReg (j, _, R)) u =
        Kodkod.AssignFormulaReg (j, to_f u)
      | to_expr_assign (RelReg (j, _, R)) u =
        Kodkod.AssignRelReg ((arity_of_rep R, j), to_r u)
      | to_expr_assign u1 _ = raise NUT ("to_expr_assign", [u1])
    (* int * int -> nut -> Kodkod.rel_expr *)
    and to_atom (x as (k, j0)) u =
      case rep_of u of
        Formula _ => atom_from_formula kk j0 (to_f u)
      | Unit => if k = 1 then Kodkod.Atom j0 else raise NUT ("to_atom", [u])
      | R => atom_from_rel_expr kk x R (to_r u)
    (* rep list -> nut -> Kodkod.rel_expr *)
    and to_struct Rs u = struct_from_rel_expr kk Rs (rep_of u) (to_r u)
    (* int -> rep -> nut -> Kodkod.rel_expr *)
    and to_vect k R u = vect_from_rel_expr kk k R (rep_of u) (to_r u)
    (* rep -> rep -> nut -> Kodkod.rel_expr *)
    and to_func R1 R2 u = rel_expr_to_func kk R1 R2 (rep_of u) (to_r u)
    (* rep -> nut -> Kodkod.rel_expr *)
    and to_opt R u =
      let val old_R = rep_of u in
        if is_opt_rep old_R then
          rel_expr_from_rel_expr kk (Opt R) old_R (to_r u)
        else
          to_rep R u
      end
    (* rep -> nut -> Kodkod.rel_expr *)
    and to_rep (Atom x) u = to_atom x u
      | to_rep (Struct Rs) u = to_struct Rs u
      | to_rep (Vect (k, R)) u = to_vect k R u
      | to_rep (Func (R1, R2)) u = to_func R1 R2 u
      | to_rep (Opt R) u = to_opt R u
      | to_rep R _ = raise REP ("to_rep", [R])
    (* nut -> Kodkod.rel_expr *)
    and to_integer u = to_opt (optable_rep ofs (type_of u) (rep_of u)) u
    (* nut list -> rep -> Kodkod.rel_expr -> Kodkod.rel_expr *)
    and to_guard guard_us R r =
      let
        val unpacked_rs = unpack_joins r
        val plain_guard_rs =
          map to_r (filter (is_Opt o rep_of) guard_us)
          |> filter_out (member (op =) unpacked_rs)
        val func_guard_us =
          filter ((is_Func andf is_opt_rep) o rep_of) guard_us
        val func_guard_rs = map to_r func_guard_us
        val guard_fs =
            map kk_no plain_guard_rs @
            map2 (kk_not oo kk_n_ary_function peephole_optim kk)
                 (map (unopt_rep o rep_of) func_guard_us) func_guard_rs
      in
        if null guard_fs then
          r
        else
          kk_rel_if (fold1 kk_or guard_fs) (empty_rel_for_rep R) r
      end
    (* rep -> rep -> Kodkod.rel_expr -> int -> Kodkod.rel_expr *)
    and to_project new_R old_R r j0 =
      rel_expr_from_rel_expr kk new_R old_R
                             (kk_project_seq r j0 (arity_of_rep old_R))
    (* rep list -> nut list -> Kodkod.rel_expr *)
    and to_product Rs us =
      case map (uncurry to_opt) (filter (not_equal Unit o fst) (Rs ~~ us)) of
        [] => raise REP ("to_product", Rs)
      | rs => fold1 kk_product rs
    (* int -> typ -> rep -> nut -> Kodkod.rel_expr *)
    and to_nth_pair_dtor n res_T res_R u =
      case u of
        Tuple (_, _, us) => to_rep res_R (nth us n)
      | _ => let
               val R = rep_of u
               val (a_T, b_T) = HOLogic.dest_prodT (type_of u)
               val Rs as [a_R, b_R] =
                 case unopt_rep R of
                   Struct (Rs as [_, _]) => Rs
                 | _ =>
                   let
                     val res_card = card_of_rep res_R
                     val other_card = card_of_rep R div res_card
                     val (a_card, b_card) = (res_card, other_card)
                                            |> n = 1 ? swap
                   in
                     [Atom (a_card, offset_of_type ofs a_T),
                      Atom (b_card, offset_of_type ofs b_T)]
                   end
               val nth_R = nth Rs n
               val j0 = if n = 0 then 0 else arity_of_rep a_R
             in
               case arity_of_rep nth_R of
                 0 => to_guard [u] res_R
                               (to_rep res_R (Cst (Unity, res_T, Unit)))
               | arity => to_project res_R nth_R (to_rep (Opt (Struct Rs)) u) j0
             end
    (* (Kodkod.formula -> Kodkod.formula -> Kodkod.formula)
       -> (Kodkod.rel_expr -> Kodkod.rel_expr -> Kodkod.formula) -> nut -> nut
       -> Kodkod.formula *)
    and to_set_bool_op connective set_oper u1 u2 =
      let
        val min_R = min_rep (rep_of u1) (rep_of u2)
        val r1 = to_rep min_R u1
        val r2 = to_rep min_R u2
      in
        case min_R of
          Vect (k, Atom _) => kk_vect_set_bool_op connective k r1 r2
        | Func (R1, Formula Neutral) => set_oper r1 r2
        | Func (Unit, Atom (2, j0)) =>
          connective (formula_from_atom j0 r1) (formula_from_atom j0 r2)
        | Func (R1, Atom _) => set_oper (kk_join r1 true_atom)
                                        (kk_join r2 true_atom)
        | _ => raise REP ("to_set_bool_op", [min_R])
      end
    (* (Kodkod.formula -> Kodkod.formula -> Kodkod.formula)
       -> (Kodkod.rel_expr -> Kodkod.rel_expr -> Kodkod.rel_expr)
       -> (Kodkod.rel_expr -> Kodkod.rel_expr -> Kodkod.formula)
       -> (Kodkod.rel_expr -> Kodkod.rel_expr -> Kodkod.formula)
       -> (Kodkod.rel_expr -> Kodkod.rel_expr -> Kodkod.formula) -> bool -> rep
       -> nut -> nut -> Kodkod.rel_expr *)
    and to_set_op connective connective3 set_oper true_set_oper false_set_oper
                  neg_second R u1 u2 =
      let
        val min_R = min_rep (rep_of u1) (rep_of u2)
        val r1 = to_rep min_R u1
        val r2 = to_rep min_R u2
        val unopt_R = unopt_rep R
      in
        to_guard [u1, u2] unopt_R
            (rel_expr_from_rel_expr kk unopt_R (unopt_rep min_R)
                 (case min_R of
                   Opt (Vect (k, Atom _)) => kk_vect_set_op connective k r1 r2
                 | Vect (k, Atom _) => kk_vect_set_op connective k r1 r2
                 | Func (_, Formula Neutral) => set_oper r1 r2
                 | Func (Unit, _) => connective3 r1 r2
                 | Func (R1, _) =>
                   double_rel_let kk
                       (fn r1 => fn r2 =>
                           kk_union
                               (kk_product
                                    (true_set_oper (kk_join r1 true_atom)
                                         (kk_join r2 (atom_for_bool bool_j0
                                                          (not neg_second))))
                                    true_atom)
                               (kk_product
                                    (false_set_oper (kk_join r1 false_atom)
                                         (kk_join r2 (atom_for_bool bool_j0
                                                          neg_second)))
                                    false_atom))
                       r1 r2
                 | _ => raise REP ("to_set_op", [min_R])))
      end
    (* rep -> rep -> Kodkod.rel_expr -> nut -> Kodkod.rel_expr *)
    and to_apply res_R func_u arg_u =
      case unopt_rep (rep_of func_u) of
        Unit =>
        let val j0 = offset_of_type ofs (type_of func_u) in
          to_guard [arg_u] res_R
                   (rel_expr_from_rel_expr kk res_R (Atom (1, j0))
                                           (Kodkod.Atom j0))
        end
      | Atom (1, j0) =>
        to_guard [arg_u] res_R
                 (rel_expr_from_rel_expr kk res_R (Atom (1, j0)) (to_r func_u))
      | Atom (k, j0) =>
        let
          val dom_card = card_of_rep (rep_of arg_u)
          val ran_R = Atom (exact_root dom_card k,
                            offset_of_type ofs (range_type (type_of func_u)))
        in
          to_apply_vect dom_card ran_R res_R (to_vect dom_card ran_R func_u)
                        arg_u
        end
      | Vect (1, R') =>
        to_guard [arg_u] res_R
                 (rel_expr_from_rel_expr kk res_R R' (to_r func_u))
      | Vect (k, R') => to_apply_vect k R' res_R (to_r func_u) arg_u
      | Func (R, Formula Neutral) =>
        to_guard [arg_u] res_R (rel_expr_from_formula kk res_R
                                    (kk_subset (to_opt R arg_u) (to_r func_u)))
      | Func (Unit, R2) =>
        to_guard [arg_u] res_R
                 (rel_expr_from_rel_expr kk res_R R2 (to_r func_u))
      | Func (R1, R2) =>
        rel_expr_from_rel_expr kk res_R R2
            (kk_n_fold_join kk R1 R2 (to_opt R1 arg_u) (to_r func_u))
        |> body_rep R2 = Formula Neutral ? to_guard [arg_u] res_R
      | _ => raise NUT ("to_apply", [func_u])
    (* int -> rep -> rep -> Kodkod.rel_expr -> nut *)
    and to_apply_vect k R' res_R func_r arg_u =
      let
        val arg_R = optable_rep ofs (type_of arg_u) (unopt_rep (rep_of arg_u))
        val vect_r = vect_from_rel_expr kk k res_R (Vect (k, R')) func_r
        val vect_rs = unpack_vect_in_chunks kk (arity_of_rep res_R) k vect_r
      in
        kk_case_switch kk arg_R res_R (to_opt arg_R arg_u)
                       (all_singletons_for_rep arg_R) vect_rs
      end
    (* occurrence -> nut -> Kodkod.formula *)
    and to_could_be_biggie neg u =
      if neg andalso is_opt_rep (rep_of u) then kk_no (to_r u)
      else Kodkod.False
    (* nut -> Kodkod.rel_expr -> Kodkod.rel_expr *)
    and to_compare_with_biggie u r =
      if is_opt_rep (rep_of u) then kk_rel_if (kk_some (to_r u)) r Kodkod.None
      else r
  in to_f_with_occurrence Positive u end

end;
