(*  Title:      Nitpick/Tools/nitpick_model.ML
    Author:     Jasmin Blanchette, TU Muenchen
    Copyright   2009

Model reconstruction for Nitpick.
*)

signature NITPICK_MODEL =
sig
  type special_fun = NitpickHOL.special_fun
  type scope = NitpickScope.scope
  type rep = NitpickRep.rep
  type nut = NitpickNut.nut

  type params = {
    show_skolems: bool,
    show_datatypes: bool,
    show_consts: bool}

  structure NameTable : TABLE

  val tuple_list_for_name :
    nut NameTable.table -> Kodkod.raw_bound list -> nut -> int list list
  val reconstruct_hol_model :
    params -> scope -> term list -> (string * string list) list
    -> special_fun list -> nut list -> nut list -> nut list
    -> nut NameTable.table -> Kodkod.raw_bound list -> Pretty.T
  val prove_hol_model :
    scope -> Time.time option -> nut list -> nut list -> nut NameTable.table
    -> Kodkod.raw_bound list -> term -> bool option
end;

structure NitpickModel : NITPICK_MODEL =
struct

open NitpickUtil
open NitpickHOL
open NitpickScope
open NitpickPeephole
open NitpickRep
open NitpickNut

type params = {
  show_skolems: bool,
  show_datatypes: bool,
  show_consts: bool}

val unknown = "\\<top>"
val maybe_mixfix = "_\\<^sup>?"
val abs_mixfix = "\\<guillemotleft>_\\<guillemotright>"

(* string -> typ -> int -> string *)
fun atom_name prefix (T as Type (s, _)) j =
    prefix ^ substring (unprefix_name s, 0, 1) ^ nat_subscript (j + 1)
  | atom_name prefix (T as TFree (s, _)) j =
    prefix ^ perhaps (try (unprefix "'")) s ^ nat_subscript (j + 1)
  | atom_name _ T _ = raise TYPE ("atom_name", [T], [])

(* nut NameTable.table -> Kodkod.raw_bound list -> nut -> int list list *)
fun tuple_list_for_name rel_table bounds name =
  the (AList.lookup (op =) bounds (lookup_rel rel_table name))
  handle NUT _ => [[]]

(* string * string -> scope -> nut list -> nut list -> nut list
   -> nut NameTable.table -> Kodkod.raw_bound list -> term *)
fun reconstruct_term (maybe_const_name, abs_const_name)
        {ctxt, card_assigns, datatypes, ofs} ctor_names rel_table bounds =
  let
    val thy = ProofContext.theory_of ctxt
    val for_auto = (maybe_const_name = "")

    (* typ -> typ -> term list -> term list -> term *)
    fun make_fun T1 T2 ts1 ts2 =
      let
        (* term -> string *)
        fun truth_const_sort_key @{const True} = "0"
          | truth_const_sort_key @{const False} = "2"
          | truth_const_sort_key _ = "1"
        (* term list -> term list -> term *)
        fun aux_general [] [] =
            if for_auto andalso T2 = bool_T then Abs ("x", T1, @{const False})
            else Const (@{const_name undefined}, T1 --> T2)
          | aux_general (t1 :: ts1) (t2 :: ts2) =
            Const (@{const_name fun_upd}, [T1 --> T2, T1, T2] ---> (T1 --> T2))
            $ aux_general ts1 ts2 $ t1 $ t2
          | aux_general _ _ = raise UnequalLengths
        (* (term * term) list -> term *)
        fun aux_set [] = Const (@{const_name Set.empty}, T1 --> T2)
          | aux_set ((t1, t2) :: zs) =
            aux_set zs
            |> t2 <> @{const False}
               ? curry (op $) (Const (@{const_name insert},
                                      [T1, T1 --> T2] ---> T1 --> T2)
                               $ (t1 |> t2 <> @{const True}
                                        ? curry (op $) (Const (maybe_const_name,
                                                               T1 --> T1))))
      in
        if for_auto orelse T2 <> bool_T then aux_general (rev ts1) (rev ts2)
        else aux_set (sort_wrt (truth_const_sort_key o snd) (ts1 ~~ ts2))
      end

    (* typ -> int -> term *)
    fun term_for_atom (T as Type ("fun", [T1, T2])) j =
        let
          val k1 = card_of_type card_assigns T1
          val k2 = card_of_type card_assigns T2
        in
          term_for_rep T (Vect (k1, Atom (k2, 0)))
                       [nth_combination (replicate k1 (k2, 0)) j]
          handle Subscript =>
                 raise ARG ("term_for_atom",
                            string_for_int j ^ " for " ^
                            string_for_rep (Vect (k1, Atom (k2, 0))))
        end
      | term_for_atom (Type ("*", [T1, T2])) j =
        let val k1 = card_of_type card_assigns T1 in
          list_comb (HOLogic.pair_const T1 T2,
                     map2 term_for_atom [T1, T2] [j div k1, j mod k1])
        end
      | term_for_atom @{typ prop} j =
        HOLogic.mk_Trueprop (term_for_atom bool_T j)
      | term_for_atom @{typ bool} j =
        if j = 0 then @{const False} else @{const True}
      | term_for_atom @{typ unit} _ = @{const Unity}
      | term_for_atom T j =
        if T = nat_T then
          HOLogic.mk_number nat_T j
        else if T = int_T then
          HOLogic.mk_number int_T
              (int_for_atom (card_of_type card_assigns int_T, 0) j)
        else if is_iterator_type T then
          HOLogic.mk_number nat_T (card_of_type card_assigns T - j - 1)
        else
          case datatype_spec datatypes T of
            NONE => if for_auto then Free (atom_name nitpick_prefix T j, T)
                    else Const (atom_name "" T j, T)
          | SOME {ctors, ...} =>
            let
              val j0 = offset_of_type ofs T
              val real_j = j + j0
              val (x as (s, T), jss) =
                map (fn {name, typ, ...} =>
                        ((name, typ), tuple_list_for_name rel_table bounds
                                          (ConstName (name, typ, Any, true))))
                    ctors
                |> filter (exists (equal real_j o List.last) o snd)
                |> the_single
              val R = get_first (fn ConstName (s', T', R, _) =>
                                    if (s', T') = x then SOME R else NONE
                                  | u => raise NUT ("term_for_atom", [u]))
                                ctor_names |> the
              val js = get_first (fn js => let val (js, j') = split_last js in
                                             if j' = real_j then SOME js
                                             else NONE
                                           end) jss |> the
              val Ts = binder_types T
              val Rs = binder_reps R
              val arities = map arity_of_rep Rs
              val jsss = map single (chunk_list_unevenly arities js)
              val ts = map3 term_for_rep Ts Rs jsss
            in
              if s = @{const_name Abs_Frac} then
                let
                  val num_T = body_type T
                  (* int -> term *)
                  val mk_num = HOLogic.mk_number num_T
                in
                  case ts of
                    [Const (@{const_name Pair}, _) $ t1 $ t2] =>
                    (case snd (HOLogic.dest_number t1) of
                       0 => mk_num 0
                     | n1 => case HOLogic.dest_number t2 |> snd of
                               1 => mk_num n1
                             | n2 => Const (@{const_name HOL.divide},
                                            [num_T, num_T] ---> num_T)
                                     $ mk_num n1 $ mk_num n2)
                  | _ => raise TERM ("term_for_atom (Abs_Frac)", ts)
                end
              else if not for_auto andalso is_abs_fun thy x then
                Const (abs_const_name, T) $ the_single ts
              else
                list_comb (Const x, ts)
            end
    (* int -> rep -> typ -> typ -> int list -> term *)
    and term_for_vect k R T1 T2 js =
      make_fun T1 T2 (map (term_for_atom T1) (0 upto k - 1))
               (map (term_for_rep T2 R o single)
                    (batch_list (arity_of_rep R) js))
    (* typ -> rep -> int list list -> term *)
    and term_for_rep T Unit [[]] = term_for_atom T 0
      | term_for_rep T (R as Atom (k, j0)) [[j]] =
        if j >= j0 andalso j < j0 + k then term_for_atom T (j - j0)
        else raise REP ("term_for_rep", [R])
      | term_for_rep (Type ("*", [T1, T2])) (Struct [R1, R2]) [js] =
        let
          val arity1 = arity_of_rep R1
          val (js1, js2) = chop arity1 js
        in
          list_comb (HOLogic.pair_const T1 T2,
                     map3 term_for_rep [T1, T2] [R1, R2] [[js1], [js2]])
        end
      | term_for_rep (Type ("fun", [T1, T2])) (R as Vect (k, R')) [js] =
        term_for_vect k R' T1 T2 js
      | term_for_rep (Type ("fun", [T1, T2])) (Func (R1, Formula Neutral)) jss =
        let
          val jss1 = all_combinations_for_rep R1
          val ts1 = map (term_for_rep T1 R1 o single) jss1
          val ts2 =
            map (fn js => term_for_rep T2 (Atom (2, 0))
                                       [[int_for_bool (js mem jss)]]) jss1
        in make_fun T1 T2 ts1 ts2 end
      | term_for_rep (Type ("fun", [T1, T2])) (Func (R1, R2)) jss =
        let
          val arity1 = arity_of_rep R1
          val jss1 = all_combinations_for_rep R1
          val ts1 = map (term_for_rep T1 R1 o single) jss1
          val grouped_jss2 = AList.group (op =) (map (chop arity1) jss)
          val ts2 = map (term_for_rep T2 R2 o the_default []
                         o AList.lookup (op =) grouped_jss2) jss1
        in make_fun T1 T2 ts1 ts2 end
      | term_for_rep T (Opt R) jss =
        if null jss then Const (unknown, T) else term_for_rep T R jss
      | term_for_rep T R jss =
        raise ARG ("term_for_rep",
                   Refute.string_of_typ T ^ " " ^ string_for_rep R ^ " " ^
                   string_for_int (length jss))
  in term_for_rep end

(* scope -> nut list -> nut NameTable.table -> Kodkod.raw_bound list -> nut
   -> term *)
fun term_for_name scope ctor_names rel_table bounds name =
  tuple_list_for_name rel_table bounds name
  |> reconstruct_term ("", "") scope ctor_names rel_table bounds (type_of name)
                      (rep_of name)

(* Proof.context -> string * string * Proof.context *)
fun add_wacky_syntax ctxt =
  let
    val thy = ProofContext.theory_of ctxt |> Context.reject_draft
    val (maybe_t, thy) =
      Sign.declare_const [] ((@{binding nitpick_maybe}, @{typ "'a => 'a"}),
                             Mixfix (maybe_mixfix, [40], 40)) thy
    val (abs_t, thy) =
      Sign.declare_const [] ((@{binding nitpick_abs}, @{typ "'a => 'b"}),
                             Mixfix (abs_mixfix, [40], 40)) thy
  in (pairself (fst o dest_Const) (maybe_t, abs_t),
      ProofContext.transfer_syntax thy ctxt) end

(* params -> scope -> term list -> (string * string list) list
   -> special_fun list -> nut list -> nut list -> nut list -> nut list
   -> nut NameTable.table -> Pretty.T *)
fun reconstruct_hol_model {show_skolems, show_datatypes, show_consts}
        {ctxt, card_assigns, datatypes, ofs} eval skolems special_funs
        free_names ctor_names nonctor_names rel_table bounds =
  let
    val (wacky_const_names, ctxt) = add_wacky_syntax ctxt
    val scope = {ctxt = ctxt, card_assigns = card_assigns,
                 datatypes = datatypes, ofs = ofs}
    val term_for_rep = reconstruct_term wacky_const_names scope ctor_names
                                        rel_table bounds
    (* string -> Pretty.T *)
    fun pretty_for_assign name =
      let
        val jss = tuple_list_for_name rel_table bounds name
        val t1 = case name of
                   FreeName (s, T, _) => Free (s, T)
                 | ConstName (s, T, _, _) =>
                   user_friendly_const eval skolems special_funs (s, T)
                 | _ => raise NUT ("pretty_for_assign", [name])
        val t2 = term_for_rep (type_of name) (rep_of name) jss
      in
        Pretty.block (Pretty.breaks
            [(setmp show_question_marks false o setmp_show_all_types)
                 (Syntax.pretty_term ctxt) t1,
             Pretty.str "=", Syntax.pretty_term ctxt t2])
      end
    (* dtype_spec -> Pretty.T *)
    fun pretty_for_datatype {typ, card, precise, ...} =
      Pretty.block (Pretty.breaks
          [Syntax.pretty_typ ctxt typ, Pretty.str "=",
           Pretty.enum "," "{" "}"
               (map (Syntax.pretty_term ctxt
                     o term_for_rep typ (Atom (card, 0)) o single o single)
                    (0 upto card - 1) @
                (if precise then [] else [Pretty.str unknown]))])
    (* typ -> dtype_spec list *)
    fun pseudo_datatype T =
      case card_of_type card_assigns T of
        ~1 => []
      | k => [{typ = T, card = k, precise = false, ctors = []}]
    val datatypes = pseudo_datatype nat_T @ pseudo_datatype int_T @ datatypes
    val block_of_datatypes =
      if show_datatypes andalso not (null datatypes) then
        [Pretty.big_list ("Datatype" ^ plural_s_for_list datatypes ^ ":")
                         (map pretty_for_datatype datatypes)]
      else
        []
    (* bool -> string -> nut list -> Pretty.T list *)
    fun block_of_names show title names =
      if show andalso not (null names) then
        Pretty.str (title ^ plural_s_for_list names ^ ":")
        :: map (Pretty.indent indent_size o pretty_for_assign)
               (sort_wrt nickname_of names)
      else
        []
    val (skolem_names, nonskolem_nonctor_names) =
      List.partition is_skolem_name nonctor_names
    val (eval_names, real_const_names) =
      List.partition (String.isPrefix eval_prefix o nickname_of)
                     nonskolem_nonctor_names
    val chunks = block_of_names true "Free variable" free_names @
                 block_of_names show_skolems "Skolem constant" skolem_names @
                 block_of_names true "Evaluated term" eval_names @
                 block_of_datatypes @
                 block_of_names show_consts "Constant" real_const_names
  in
    Pretty.chunks (if null chunks then [Pretty.str "Empty assignment"]
                   else chunks)
  end

(* scope -> Time.time option -> nut list -> nut list -> nut NameTable.table
  -> Kodkod.raw_bound list -> term -> bool option *)
fun prove_hol_model (scope as {ctxt, card_assigns, ...}) auto_timeout free_names
                    ctor_names rel_table bounds prop =
  let
    val thy = ProofContext.theory_of ctxt
    (* typ * int -> term *)
    fun free_type_assm (T, k) =
      let
        (* int -> term *)
        fun atom j = Free (atom_name nitpick_prefix T j, T)
        fun equation_for_atom j = HOLogic.eq_const T $ Bound 0 $ atom j
        val eqs = map equation_for_atom (0 upto k - 1)
        val compreh_assm =
          Const (@{const_name All}, (T --> bool_T) --> bool_T)
              $ Abs ("x", T, foldl1 HOLogic.mk_disj eqs)
        val distinct_assm = distinctness_formula T (map atom (0 upto k - 1))
      in HOLogic.mk_conj (compreh_assm, distinct_assm) end
    (* nut -> term *)
    fun free_name_assm name =
      HOLogic.mk_eq (Free (nickname_of name, type_of name),
                     term_for_name scope ctor_names rel_table bounds name)

    val freeT_assms = map free_type_assm (filter (is_TFree o fst) card_assigns)
    val model_assms = map free_name_assm free_names
    val assm = foldr HOLogic.mk_conj @{const True} (freeT_assms @ model_assms)
    (* bool -> bool *)
    fun try_out negate =
      let
        val concl = (negate ? curry (op $) @{const Not})
                    (ObjectLogic.atomize_term thy prop)
        val goal = HOLogic.mk_Trueprop (HOLogic.mk_imp (assm, concl))
                   |> map_types (map_type_tfree
                          (fn (s, []) => TFree (s, HOLogic.typeS)
                            | x => TFree x))
                   |> cterm_of thy |> Goal.init
      in
        (goal |> SINGLE (timeout_tac auto_timeout
                                     (auto_tac (local_clasimpset_of ctxt)))
              |> the |> Goal.finish; true)
        handle THM _ => false
             | TimeLimit.TimeOut => false
      end
  in
    if silence try_out false then SOME true
    else if silence try_out true then SOME false
    else NONE
  end

end;
