(*  Title:      Nitpick/Tools/nitpick_nut.ML
    Author:     Jasmin Blanchette, TU Muenchen
    Copyright   2008, 2009

Nitpick underlying terms (nuts).
*)

signature NITPICK_NUT =
sig
  type special_fun = NitpickHOL.special_fun
  type scope = NitpickScope.scope
  type name_pool = NitpickPeephole.name_pool
  type rep = NitpickRep.rep

  datatype cst =
    Unity |
    False |
    True |
    Iden |
    Num of int |
    Biggie |
    Suc |
    Add |
    Subtract |
    Multiply |
    Divide |
    Modulo |
    Gcd |
    Lcm |
    Fracs |
    NormFrac |
    NatToInt |
    IntToNat

  datatype op1 =
    Not |
    Finite |
    Converse |
    Closure |
    SingletonSet |
    First |
    Second |
    RepOfAbs |
    Cast

  datatype op2 =
    All |
    Exist |
    Or |
    Implies |
    And |
    Less |
    Subset |
    DefEq |
    Eq |
    Triad |
    Union |
    SetDifference |
    Intersect |
    Composition |
    Product |
    Image |
    Apply |
    Abstraction

  datatype op3 =
    Let |
    If

  datatype nut =
    Cst of cst * typ * rep |
    Op1 of op1 * typ * rep * nut |
    Op2 of op2 * typ * rep * nut * nut |
    Op3 of op3 * typ * rep * nut * nut * nut |
    Tuple of typ * rep * nut list |
    BoundName of int * typ * rep * string |
    FreeName of string * typ * rep |
    ConstName of string * typ * rep * bool |
    BoundRel of Kodkod.n_ary_index * typ * rep * string |
    FreeRel of Kodkod.n_ary_index * typ * rep * string |
    RelReg of int * typ * rep |
    FormulaReg of int * typ * rep

  structure NameTable : TABLE

  exception NUT of string * nut list

  val string_for_nut : Proof.context -> nut -> string
  val inline_nut : nut -> bool
  val type_of : nut -> typ
  val rep_of : nut -> rep
  val nickname_of : nut -> string
  val is_const_name : (theory -> styp -> bool) -> theory -> nut -> bool
  val is_skolem_name : nut -> bool
  val is_Cst : cst -> nut -> bool
  val is_mate_of_rep_fun_nut : theory -> nut -> nut -> bool
  val fold_nut : (nut -> 'a -> 'a) -> nut -> 'a -> 'a
  val untuple : (nut -> 'a) -> nut -> 'a list
  val add_free_and_const_names :
    nut -> nut list * nut list -> nut list * nut list
  val name_ord : (nut * nut) -> order
  val lookup_name : 'a NameTable.table -> nut -> 'a
  val lookup_rel : nut NameTable.table -> nut -> Kodkod.n_ary_index
  val nut_from_term : theory -> special_fun list -> op2 -> term -> nut
  val choose_reps_for_free_vars :
    scope -> nut list -> rep NameTable.table -> nut list * rep NameTable.table
  val choose_reps_for_consts :
    scope -> bool -> nut list -> rep NameTable.table
    -> nut list * rep NameTable.table
  val choose_reps_for_all_ctors :
    scope -> rep NameTable.table -> nut list * rep NameTable.table
  val choose_reps_in_nut :
    scope -> bool -> rep NameTable.table -> bool -> nut -> nut
  val rename_free_vars :
    nut list -> name_pool -> nut NameTable.table
    -> nut list * name_pool * nut NameTable.table
  val rename_rep_funs :
    (nut * nut) list -> nut NameTable.table -> nut NameTable.table
  val rename_vars_in_nut : name_pool -> nut NameTable.table -> nut -> nut
end;

structure NitpickNut : NITPICK_NUT =
struct

open NitpickUtil
open NitpickHOL
open NitpickScope
open NitpickPeephole
open NitpickRep

datatype cst =
  Unity |
  False |
  True |
  Iden |
  Num of int |
  Biggie |
  Suc |
  Add |
  Subtract |
  Multiply |
  Divide |
  Modulo |
  Gcd |
  Lcm |
  Fracs |
  NormFrac |
  NatToInt |
  IntToNat

datatype op1 =
  Not |
  Finite |
  Converse |
  Closure |
  SingletonSet |
  First |
  Second |
  RepOfAbs |
  Cast

datatype op2 =
  All |
  Exist |
  Or |
  Implies |
  And |
  Less |
  Subset |
  DefEq |
  Eq |
  Triad |
  Union |
  SetDifference |
  Intersect |
  Composition |
  Product |
  Image |
  Apply |
  Abstraction

datatype op3 =
  Let |
  If

datatype nut =
  Cst of cst * typ * rep |
  Op1 of op1 * typ * rep * nut |
  Op2 of op2 * typ * rep * nut * nut |
  Op3 of op3 * typ * rep * nut * nut * nut |
  Tuple of typ * rep * nut list |
  BoundName of int * typ * rep * string |
  FreeName of string * typ * rep |
  ConstName of string * typ * rep * bool |
  BoundRel of Kodkod.n_ary_index * typ * rep * string |
  FreeRel of Kodkod.n_ary_index * typ * rep * string |
  RelReg of int * typ * rep |
  FormulaReg of int * typ * rep

exception NUT of string * nut list

(* cst -> string *)
fun string_for_cst Unity = "Unity"
  | string_for_cst False = "False"
  | string_for_cst True = "True"
  | string_for_cst Iden = "Iden"
  | string_for_cst (Num j) = "Num " ^ string_for_int j
  | string_for_cst Biggie = "Biggie"
  | string_for_cst Suc = "Suc"
  | string_for_cst Add = "Add"
  | string_for_cst Subtract = "Subtract"
  | string_for_cst Multiply = "Multiply"
  | string_for_cst Divide = "Divide"
  | string_for_cst Modulo = "Modulo"
  | string_for_cst Gcd = "Gcd"
  | string_for_cst Lcm = "Lcm"
  | string_for_cst Fracs = "Fracs"
  | string_for_cst NormFrac = "NormFrac"
  | string_for_cst NatToInt = "NatToInt"
  | string_for_cst IntToNat = "IntToNat"

(* op1 -> string *)
fun string_for_op1 Not = "Not"
  | string_for_op1 Finite = "Finite"
  | string_for_op1 Converse = "Converse"
  | string_for_op1 Closure = "Closure"
  | string_for_op1 SingletonSet = "SingletonSet"
  | string_for_op1 First = "First"
  | string_for_op1 Second = "Second"
  | string_for_op1 RepOfAbs = "RepOfAbs"
  | string_for_op1 Cast = "Cast"

(* op2 -> string *)
fun string_for_op2 All = "All"
  | string_for_op2 Exist = "Exist"
  | string_for_op2 Or = "Or"
  | string_for_op2 Implies = "Implies"
  | string_for_op2 And = "And"
  | string_for_op2 Less = "Less"
  | string_for_op2 Subset = "Subset"
  | string_for_op2 DefEq = "DefEq"
  | string_for_op2 Eq = "Eq"
  | string_for_op2 Triad = "Triad"
  | string_for_op2 Union = "Union"
  | string_for_op2 SetDifference = "SetDifference"
  | string_for_op2 Intersect = "Intersect"
  | string_for_op2 Composition = "Composition"
  | string_for_op2 Product = "Product"
  | string_for_op2 Image = "Image"
  | string_for_op2 Apply = "Apply"
  | string_for_op2 Abstraction = "Abstraction"

(* op3 -> string *)
fun string_for_op3 Let = "Let"
  | string_for_op3 If = "If"

(* int -> Proof.context -> nut -> string *)
fun basic_string_for_nut indent ctxt u =
  let
    (* nut -> string *)
    val sub = basic_string_for_nut (indent + 1) ctxt
  in
    (if indent = 0 then "" else "\n" ^ StringCvt.padLeft #" " (2 * indent) "") ^
    "(" ^
    (case u of
       Cst (c, T, R) =>
       "Cst " ^ string_for_cst c ^ " " ^ Syntax.string_of_typ ctxt T ^ " " ^
       string_for_rep R
     | Op1 (oper, T, R, u1) =>
       "Op1 " ^ string_for_op1 oper ^ " " ^ Syntax.string_of_typ ctxt T ^ " " ^
       string_for_rep R ^ " " ^ sub u1
     | Op2 (oper, T, R, u1, u2) =>
       "Op2 " ^ string_for_op2 oper ^ " " ^ Syntax.string_of_typ ctxt T ^ " " ^
       string_for_rep R ^ " " ^ sub u1 ^ " " ^ sub u2
     | Op3 (oper, T, R, u1, u2, u3) =>
       "Op3 " ^ string_for_op3 oper ^ " " ^ Syntax.string_of_typ ctxt T ^ " " ^
       string_for_rep R ^ " " ^ sub u1 ^ " " ^ sub u2 ^ " " ^ sub u3
     | Tuple (T, R, us) =>
       "Tuple " ^ Syntax.string_of_typ ctxt T ^ " " ^ string_for_rep R ^
       implode (map sub us)
     | BoundName (j, T, R, nick) =>
       "BoundName " ^ string_for_int j ^ " " ^ Syntax.string_of_typ ctxt T ^
       " " ^ string_for_rep R ^ " " ^ nick
     | FreeName (s, T, R) =>
       "FreeName " ^ s ^ " " ^ Syntax.string_of_typ ctxt T ^ " " ^
       string_for_rep R
     | ConstName (s, T, R, ctive) =>
       "ConstName " ^ s ^ " " ^ Syntax.string_of_typ ctxt T ^ " " ^
       string_for_rep R ^ " " ^ (if ctive then "C" else "NC")
     | BoundRel ((n, j), T, R, nick) =>
       "BoundRel " ^ string_for_int n ^ "." ^ string_for_int j ^ " " ^
       Syntax.string_of_typ ctxt T ^ " " ^ string_for_rep R ^ " " ^ nick
     | FreeRel ((n, j), T, R, nick) =>
       "FreeRel " ^ string_for_int n ^ "." ^ string_for_int j ^ " " ^
       Syntax.string_of_typ ctxt T ^ " " ^ string_for_rep R ^ " " ^ nick
     | RelReg (j, T, R) =>
       "RelReg " ^ string_for_int j ^ " " ^ Syntax.string_of_typ ctxt T ^ " " ^
       string_for_rep R
     | FormulaReg (j, T, R) =>
       "FormulaReg " ^ string_for_int j ^ " " ^ Syntax.string_of_typ ctxt T ^
       " " ^ string_for_rep R) ^
    ")"
  end
(* Proof.context -> nut -> string *)
val string_for_nut = basic_string_for_nut 0

(* nut -> bool *)
fun inline_nut (Op1 _) = false
  | inline_nut (Op2 _) = false
  | inline_nut (Op3 _) = false
  | inline_nut (Tuple (_, _, us)) = forall inline_nut us
  | inline_nut _ = true

(* nut -> typ *)
fun type_of (Cst (_, T, _)) = T
  | type_of (Op1 (_, T, _, _)) = T
  | type_of (Op2 (_, T, _, _, _)) = T
  | type_of (Op3 (_, T, _, _, _, _)) = T
  | type_of (Tuple (T, _, _)) = T
  | type_of (BoundName (_, T, _, _)) = T
  | type_of (FreeName (_, T, _)) = T
  | type_of (ConstName (_, T, _, _)) = T
  | type_of (BoundRel (_, T, _, _)) = T
  | type_of (FreeRel (_, T, _, _)) = T
  | type_of (RelReg (_, T, _)) = T
  | type_of (FormulaReg (_, T, _)) = T

(* nut -> rep *)
fun rep_of (Cst (_, _, R)) = R
  | rep_of (Op1 (_, _, R, _)) = R
  | rep_of (Op2 (_, _, R, _, _)) = R
  | rep_of (Op3 (_, _, R, _, _, _)) = R
  | rep_of (Tuple (_, R, _)) = R
  | rep_of (BoundName (_, _, R, _)) = R
  | rep_of (FreeName (_, _, R)) = R
  | rep_of (ConstName (_, _, R, _)) = R
  | rep_of (BoundRel (_, _, R, _)) = R
  | rep_of (FreeRel (_, _, R, _)) = R
  | rep_of (RelReg (_, _, R)) = R
  | rep_of (FormulaReg (_, _, R)) = R

(* nut -> string *)
fun nickname_of (BoundName (_, _, _, nick)) = nick
  | nickname_of (FreeName (s, _, _)) = s
  | nickname_of (ConstName (s, _, _, _)) = s
  | nickname_of (BoundRel (_, _, _, nick)) = nick
  | nickname_of (FreeRel (_, _, _, nick)) = nick
  | nickname_of u = raise NUT ("nickname_of", [u])

(* (theory -> styp -> bool) -> theory -> nut -> bool *)
fun is_const_name is_x thy (ConstName (s, T, _, _)) = is_x thy (s, T)
  | is_const_name _ _ _ = false
(* nut -> bool *)
fun is_skolem_name (ConstName (s, _, _, _)) = String.isPrefix skolem_prefix s
  | is_skolem_name _ = false
(* cst -> nut -> bool *)
fun is_Cst cst (Cst (cst', _, _)) = (cst = cst')
  | is_Cst _ _ = false

(* theory -> nut -> nut -> bool *)
fun is_mate_of_rep_fun_nut thy rep_u abs_u =
  case (type_of rep_u, type_of abs_u) of
    (Type ("fun", [T11, T12]), abs_T as Type ("fun", [T21, T22])) =>
    (T11, T12) = (T22, T21) andalso is_abs_fun thy (nickname_of abs_u, abs_T)
  | _ => false

(* theory -> nut -> bool *)
fun is_biggie thy (Op2 (Apply, _, _, u1, u2)) =
    is_biggie thy u2 andalso is_const_name is_ctor_like thy u1
  | is_biggie _ u = is_Cst Biggie u

(* (nut -> 'a -> 'a) -> nut -> 'a -> 'a *)
fun fold_nut f u =
  case u of
    Op1 (_, _, _, u1) => fold_nut f u1
  | Op2 (_, _, _, u1, u2) => fold_nut f u1 #> fold_nut f u2
  | Op3 (_, _, _, u1, u2, u3) => fold_nut f u1 #> fold_nut f u2 #> fold_nut f u3
  | Tuple (_, _, us) => fold (fold_nut f) us
  | _ => f u

(* nut * nut -> order *)
fun name_ord (BoundName (j1, _, _, _), BoundName (j2, _, _, _)) =
    int_ord (j1, j2)
  | name_ord (BoundName _, _) = LESS
  | name_ord (_, BoundName _) = GREATER
  | name_ord (FreeName (s1, T1, _), FreeName (s2, T2, _)) =
    (case fast_string_ord (s1, s2) of
       EQUAL => TermOrd.typ_ord (T1, T2)
     | ord => ord)
  | name_ord (FreeName _, _) = LESS
  | name_ord (_, FreeName _) = GREATER
  | name_ord (ConstName (s1, T1, _, _), ConstName (s2, T2, _, _)) =
    (case fast_string_ord (s1, s2) of
       EQUAL => TermOrd.typ_ord (T1, T2)
     | ord => ord)
  | name_ord (u1, u2) = raise NUT ("name_ord", [u1, u2])

(* nut -> nut -> int *)
fun num_occurrences_in_nut needle_u stack_u =
  fold_nut (fn u => if u = needle_u then curry (op +) 1 else I) stack_u 0
(* nut -> nut -> bool *)
val is_subterm_of = not_equal 0 oo num_occurrences_in_nut

(* nut -> nut *)
fun optimize_unit u =
  if rep_of u = Unit then Cst (Unity, type_of u, Unit) else u
(* theory -> nut -> nut *)
fun optimize_biggie thy u =
  if is_biggie thy u then Cst (Biggie, type_of u, rep_of u) else u

(* theory -> nut -> nut -> nut -> nut *)
fun substitute_in_nut thy needle_u needle_u' =
  let
    (* nut -> nut *)
    fun aux stack_u =
      if stack_u = needle_u then
        needle_u'
      else
        case stack_u of
          Op1 (oper, T, R, u1) => Op1 (oper, T, R, aux u1)
        | Op2 (oper, T, R, u1, u2) =>
          optimize_biggie thy (Op2 (oper, T, R, aux u1, aux u2))
        | Op3 (oper, T, R, u1, u2, u3) =>
          Op3 (oper, T, R, aux u1, aux u2, aux u3)
        | Tuple (T, R, us) => Tuple (T, R, map aux us)
        |  _ => stack_u
  in aux end

(* theory -> nut -> nut -> nut -> nut *)
fun optimized_let thy u1 u2 u3 =
  if inline_nut u2 orelse num_occurrences_in_nut u1 u3 < 2 then
    substitute_in_nut thy u1 u2 u3
  else
    Op3 (Let, type_of u3, rep_of u3, u1, u2, u3)

(* (nut -> 'a) -> nut -> 'a list *)
fun untuple f (Tuple (_, _, us)) = maps (untuple f) us
  | untuple f u = if rep_of u = Unit then [] else [f u]

(* nut -> nut list * nut list -> nut list * nut list *)
val add_free_and_const_names =
  fold_nut (fn u => case u of
                      FreeName _ => apfst (insert (op =) u)
                    | ConstName _ => apsnd (insert (op =) u)
                    | _ => I)

(* nut -> rep -> nut *)
fun change_name_rep (BoundName (j, T, _, nick)) R = BoundName (j, T, R, nick)
  | change_name_rep (FreeName (s, T, _)) R = FreeName (s, T, R)
  | change_name_rep (ConstName (s, T, _, ctive)) R = ConstName (s, T, R, ctive)
  | change_name_rep u _ = raise NUT ("change_name_rep", [u])

structure NameTable = TableFun(type key = nut val ord = name_ord)

(* 'a NameTable.table -> nut -> 'a *)
fun lookup_name table name =
  case NameTable.lookup table name of
    SOME u => u
  | NONE => raise NUT ("lookup_name", [name])
(* nut NameTable.table -> nut -> Kodkod.n_ary_index *)
fun lookup_rel table name =
  case lookup_name table name of
    FreeRel (x, _, _, _) => x
  | u => raise NUT ("lookup_rel", [u])

(* theory -> special_fun list -> op2 -> term -> nut *)
fun nut_from_term thy special_funs eq =
  let
    (* string list -> typ list -> term -> nut *)
    fun aux eq ss Ts t =
      let
        (* term -> nut *)
        val sub = aux eq ss Ts
        (* string -> typ -> term -> nut *)
        fun sub_abs s T = aux eq (s :: ss) (T :: Ts)
        (* typ -> term -> term -> nut *)
        fun sub_equals T t1 t2 =
          let
            val (binder_Ts, body_T) = strip_type (domain_type T)
            val n = length binder_Ts
          in
            if eq = Eq andalso n > 0 then
              let
                val t1 = incr_boundvars n t1
                val t2 = incr_boundvars n t2
                val xs = map Bound (n - 1 downto 0)
                val equation = Const (@{const_name "op ="},
                                      body_T --> body_T --> bool_T)
                                   $ betapplys (t1, xs) $ betapplys (t2, xs)
                val (t, _) =
                  foldr (fn (T, (t, j)) =>
                            (Const (@{const_name All}, T --> bool_T)
                             $ Abs ("x" ^ nat_subscript j, T, t),
                             j - 1))
                        (equation, n) binder_Ts
              in sub t end
            else
              Op2 (eq, bool_T, Any, aux Eq ss Ts t1, aux Eq ss Ts t2)
          end
        (* op2 -> string -> typ -> term -> nut *)
        fun do_quantifier quant s T t1 =
          let
            val bound_u = BoundName (length Ts, T, Any, s)
            val body_u = sub_abs s T t1
          in
            if is_subterm_of bound_u body_u then
              Op2 (quant, bool_T, Any, bound_u, body_u)
            else
              body_u
          end
      in
        case t of
          Const (@{const_name all}, _) $ Abs (s, T, t1) =>
          do_quantifier All s T t1
        | (t0 as Const (@{const_name all}, T)) $ t1 =>
          sub (t0 $ eta_expand Ts t1 1)
        | Const (@{const_name "=="}, T) $ t1 $ t2 => sub_equals T t1 t2
        | Const (@{const_name "==>"}, _) $ t1 $ t2 =>
          Op2 (Implies, bool_T, Any, sub t1, sub t2)
        | Const (@{const_name Pure.conjunction}, _) $ t1 $ t2 =>
          Op2 (And, bool_T, Any, sub t1, sub t2)
        | Const (@{const_name Trueprop}, _) $ t1 => sub t1
        | Const (@{const_name Not}, _) $ t1 =>
          (case sub t1 of
             Op1 (Not, _, _, u11) => u11
           | u1 => Op1 (Not, bool_T, Any, u1))
        | Const (@{const_name False}, T) => Cst (False, T, Any)
        | Const (@{const_name True}, T) => Cst (True, T, Any)
        | Const (@{const_name All}, _) $ Abs (s, T, t1) =>
          do_quantifier All s T t1
        | (t0 as Const (@{const_name All}, T)) $ t1 =>
          sub (t0 $ eta_expand Ts t1 1)
        | Const (@{const_name Ex}, _) $ Abs (s, T, t1) =>
          do_quantifier Exist s T t1
        | (t0 as Const (@{const_name Ex}, T)) $ t1 =>
          sub (t0 $ eta_expand Ts t1 1)
        | Const (@{const_name "op ="}, T) $ t1 $ t2 => sub_equals T t1 t2
        | Const (@{const_name "op &"}, _) $ t1 $ t2 =>
          Op2 (And, bool_T, Any, sub t1, sub t2)
        | Const (@{const_name "op |"}, _) $ t1 $ t2 =>
          Op2 (Or, bool_T, Any, sub t1, sub t2)
        | Const (@{const_name "op -->"}, _) $ t1 $ t2 =>
          Op2 (Implies, bool_T, Any, sub t1, sub t2)
        | Const (@{const_name If}, T) $ t1 $ t2 $ t3 =>
          Op3 (If, third_range_type T, Any, sub t1, sub t2, sub t3)
        | Const (@{const_name Let}, T) $ t1 $ Abs (s, T', t2) =>
          Op3 (Let, second_range_type T, Any, BoundName (length Ts, T', Any, s),
               sub t1, sub_abs s T' t2)
        | (t0 as Const (@{const_name Let}, T)) $ t1 $ t2 =>
          sub (t0 $ t1 $ eta_expand Ts t2 1)
        | @{const Unity} => Cst (Unity, @{typ unit}, Any)
        | Const (@{const_name Pair}, T) $ t1 $ t2 =>
          Tuple (second_range_type T, Any, map (sub) [t1, t2])
        | Const (@{const_name fst}, T) $ t1 =>
          Op1 (First, range_type T, Any, sub t1)
        | Const (@{const_name snd}, T) $ t1 =>
          Op1 (Second, range_type T, Any, sub t1)
        | Const (@{const_name Id}, T) => Cst (Iden, T, Any)
        | Const (@{const_name "op :"}, _) $ t1 $ t2 => sub (t2 $ t1)
        | Const (@{const_name insert}, T) $ t1 $ t2 =>
          (case t2 of
             Const (@{const_name Set.empty}, _) =>
             Op1 (SingletonSet, second_range_type T, Any, sub t1)
           | _ =>
             Op2 (Union, second_range_type T, Any,
                  Op1 (SingletonSet, second_range_type T, Any, sub t1), sub t2))
        | Const (@{const_name converse}, T) $ t1 =>
          Op1 (Converse, range_type T, Any, sub t1)
        | Const (@{const_name trancl}, T) $ t1 =>
          Op1 (Closure, range_type T, Any, sub t1)
        | Const (@{const_name rtrancl}, T) $ t1 =>
          let val ran_T = range_type T in
            Op2 (Union, ran_T, Any, Op1 (Closure, ran_T, Any, sub t1),
                 Cst (Iden, ran_T, Any))
          end
        | Const (@{const_name Set.Un}, T) $ t1 $ t2 =>
          Op2 (Union, second_range_type T, Any, sub t1, sub t2)
        | Const (@{const_name Set.Int}, T) $ t1 $ t2 =>
          Op2 (Intersect, second_range_type T, Any, sub t1, sub t2)
        | Const (@{const_name rel_comp}, T) $ t1 $ t2 =>
          Op2 (Composition, second_range_type T, Any, sub t1, sub t2)
        | Const (@{const_name Sigma}, T) $ t1 $ Abs (s, T', t2') =>
          Op2 (Product, second_range_type T, Any, sub t1, sub_abs s T' t2')
        | Const (@{const_name image}, T) $ t1 $ t2 =>
          Op2 (Image, second_range_type T, Any, sub t1, sub t2)
        | Const (@{const_name Suc}, T) => Cst (Suc, T, Any)
        | Const (@{const_name finite}, T) $ t1 =>
          Op1 (Finite, bool_T, Any, sub t1)
        | Const (@{const_name nat}, T) => Cst (IntToNat, T, Any)
        | Const (@{const_name zero_nat_inst.zero_nat}, T) => Cst (Num 0, T, Any)
        | Const (@{const_name one_nat_inst.one_nat}, T) => Cst (Num 1, T, Any)
        | Const (@{const_name plus_nat_inst.plus_nat}, T) => Cst (Add, T, Any)
        | Const (@{const_name minus_nat_inst.minus_nat}, T) =>
          Cst (Subtract, T, Any)
        | Const (@{const_name times_nat_inst.times_nat}, T) =>
          Cst (Multiply, T, Any)
        | Const (@{const_name div_nat_inst.div_nat}, T) => Cst (Divide, T, Any)
        | Const (@{const_name div_nat_inst.mod_nat}, T) => Cst (Modulo, T, Any)
        | Const (@{const_name ord_nat_inst.less_nat}, T) $ t1 $ t2 =>
          Op2 (Less, bool_T, Any, sub t1, sub t2)
        | Const (@{const_name ord_nat_inst.less_eq_nat}, T) $ t1 $ t2 =>
          Op1 (Not, bool_T, Any, Op2 (Less, bool_T, Any, sub t2, sub t1))
        | Const (@{const_name gcd}, T) => Cst (Gcd, T, Any)
        | Const (@{const_name lcm}, T) => Cst (Lcm, T, Any)
        | Const (@{const_name zero_int_inst.zero_int}, T) =>
          Cst (Num 0, T, Any)
        | Const (@{const_name one_int_inst.one_int}, T) =>
          Cst (Num 1, T, Any)
        | Const (@{const_name plus_int_inst.plus_int}, T) =>
          Cst (Add, T, Any)
        | Const (@{const_name minus_int_inst.minus_int}, T) =>
          Cst (Subtract, T, Any)
        | Const (@{const_name times_int_inst.times_int}, T) =>
          Cst (Multiply, T, Any)
        | Const (@{const_name div_int_inst.div_int}, T) => Cst (Divide, T, Any)
        | Const (@{const_name div_int_inst.mod_int}, T) => Cst (Modulo, T, Any)
        | Const (@{const_name uminus_int_inst.uminus_int}, T) =>
          Op2 (Apply, int_T --> int_T, Any,
               Cst (Subtract, [int_T, int_T] ---> int_T, Any),
               Cst (Num 0, int_T, Any))
        | Const (@{const_name ord_int_inst.less_int}, T) $ t1 $ t2 =>
          Op2 (Less, bool_T, Any, sub t1, sub t2)
        | Const (@{const_name ord_int_inst.less_eq_int}, T) $ t1 $ t2 =>
          Op1 (Not, bool_T, Any, Op2 (Less, bool_T, Any, sub t2, sub t1))
        | Const (@{const_name Frac}, T) => Cst (Fracs, T, Any)
        | Const (@{const_name norm_frac}, T) => Cst (NormFrac, T, Any)
        | Const (@{const_name of_nat}, T as @{typ "nat => int"}) =>
          Cst (NatToInt, T, Any)
        | (t0 as Const (@{const_name minus_fun_inst.minus_fun}, T)) $ t1 $ t2 =>
          Op2 (SetDifference, second_range_type T, Any, sub t1, sub t2)
        | (t0 as Const (@{const_name ord_fun_inst.less_eq_fun}, T)) $ t1 $ t2 =>
          Op2 (Subset, bool_T, Any, sub t1, sub t2)
        | Const (@{const_name Rep_Frac}, _)
          $ (Const (@{const_name Abs_Frac}, _) $ t1) => sub t1
        | Const (x as (s, T)) =>
          if String.isPrefix numeral_prefix s then
            Cst (Num (the (Int.fromString (unprefix numeral_prefix s))), T, Any)
          else
            ConstName (s, T, Any, is_ctor_like thy x)
        | Free (s, T) => FreeName (s, T, Any)
        | Var _ => raise TERM ("nut_from_term", [t])
        | Bound j => BoundName (length Ts - j - 1, nth Ts j, Any, nth ss j)
        | Abs (s, T, t1) =>
          Op2 (Abstraction, T --> fastype_of1 (T :: Ts, t1), Any,
               BoundName (length Ts, T, Any, s), sub_abs s T t1)
        | t1 $ t2 =>
          Op2 (Apply, range_type (fastype_of1 (Ts, t1)), Any, sub t1, sub t2)
      end
  in aux eq [] [] end

(* scope -> typ -> rep *)
fun rep_for_abs_fun scope T =
  let val (R1, R2) = best_non_opt_symmetric_reps_for_fun_type scope T in
    Func (R1, (card_of_rep R1 <> card_of_rep R2 ? Opt) R2)
  end

(* scope -> nut -> nut list * rep NameTable.table
   -> nut list * rep NameTable.table *)
fun choose_rep_for_free_var scope v (vs, table) =
  let
    val R = best_non_opt_set_rep_for_type scope (type_of v)
    val v = change_name_rep v R
  in (v :: vs, NameTable.update (v, R) table) end
(* scope -> bool -> nut -> nut list * rep NameTable.table
   -> nut list * rep NameTable.table *)
fun choose_rep_for_const (scope as {ctxt, datatypes, ...}) all_precise v
                         (vs, table) =
  let
    val thy = ProofContext.theory_of ctxt
    val s = nickname_of v
    val T = type_of v
    val R = (if is_abs_fun thy (s, T) then
               rep_for_abs_fun
             else if is_rep_fun thy (s, T) then
               Func oo best_non_opt_symmetric_reps_for_fun_type
             else if all_precise orelse is_skolem_name v
                     orelse s mem [@{const_name hd}, @{const_name tl}] then
               best_non_opt_set_rep_for_type
             else if s mem [@{const_name set}, @{const_name ord_class.less},
                             @{const_name ord_class.less_eq}] then
               best_set_rep_for_type
             else
               best_opt_set_rep_for_type) scope T
    val v = change_name_rep v R
  in (v :: vs, NameTable.update (v, R) table) end

(* scope -> nut list -> rep NameTable.table -> nut list * rep NameTable.table *)
fun choose_reps_for_free_vars scope vs table =
  fold (choose_rep_for_free_var scope) vs ([], table)
(* scope -> bool -> nut list -> rep NameTable.table
   -> nut list * rep NameTable.table *)
fun choose_reps_for_consts scope all_precise vs table =
  fold (choose_rep_for_const scope all_precise) vs ([], table)

(* scope -> ctor_spec -> nut list * rep NameTable.table
   -> nut list * rep NameTable.table *)
fun choose_rep_for_ctor (scope as {ctxt, ...}) {name, typ, total, ...}
                        (vs, table) =
  let
    val thy = ProofContext.theory_of ctxt
    val R = if is_abs_fun thy (name, typ) then rep_for_abs_fun scope typ
            else best_opt_set_rep_for_type scope typ |> total ? unopt_rep
    val v = ConstName (name, typ, R, true)
  in (v :: vs, NameTable.update (v, R) table) end
(* scope -> dtype_spec -> nut list * rep NameTable.table
   -> nut list * rep NameTable.table *)
fun choose_rep_for_ctors_of_datatype scope {ctors, ...} (vs, table) =
  fold (choose_rep_for_ctor scope) (rev ctors) (vs, table)
(* scope -> rep NameTable.table -> nut list * rep NameTable.table *)
fun choose_reps_for_all_ctors (scope as {datatypes, ...}) table =
  fold (choose_rep_for_ctors_of_datatype scope) (rev datatypes) ([], table)

(* scope -> nut -> rep NameTable.table -> rep NameTable.table *)
fun choose_rep_for_bound_var scope v table =
  let val R = best_one_rep_for_type scope (type_of v) in
    NameTable.update (v, R) table
  end

(* A nut is said to be constructive if whenever it evaluates to unknown in our
   three-valued logic, it would evaluate to a non-representable value (a
   "biggie") according to the HOL semantics. For example, "Suc n" is
   constructive if n is representable or biggie, because unknown implies
   biggie. *)
(* theory -> nut -> bool *)
fun is_constructive thy u =
  (not (is_fun_type (type_of u))
   andalso (not (is_opt_rep (rep_of u)) orelse is_Cst Biggie u))
  orelse case u of
           Cst (Num _, _, _) => true
         | Cst (cst, T, _) =>
           cst = Suc orelse (body_type T = nat_T andalso cst = Add)
         | Op2 (Apply, _, _, u1, u2) => forall (is_constructive thy) [u1, u2]
         | Op3 (If, _, _, u1, u2, u3) =>
           not (is_opt_rep (rep_of u1))
           andalso forall (is_constructive thy) [u2, u3]
         | ConstName (_, _, _, ctive) => ctive
         | _ => false

(* op2 -> nut -> nut list *)
fun phrases All (Op2 (All, _, _, _, u2)) = phrases All u2
  | phrases All (Op2 (Implies, _, _, u1, u2)) =
    phrases Exist u1 @ phrases All u2
  | phrases Exist (Op2 (Exist, _, _, _, u2)) = phrases Exist u2
  | phrases Exist (Op2 (And, _, _, u1, u2)) = maps (phrases Exist) [u1, u2]
  | phrases _ u = [u]

(* theory -> op2 -> nut -> nut -> bool *)
fun is_subterm_equation thy oper bound_u body_u =
  rep_of bound_u <> Unit andalso
  let
    (* nut -> nut -> bool *)
    fun good_eq u1 u2 =
      not (is_opt_rep (rep_of u2)) andalso not (is_subterm_of bound_u u2)
      andalso is_subterm_of bound_u u1 andalso is_constructive thy u1
  in
    case body_u of
      Op2 (Eq, _, _, u1, u2) => good_eq u1 u2 orelse good_eq u2 u1
    | Op2 (DefEq, _, _, u1, u2) => good_eq u1 u2 orelse good_eq u2 u1
    | _ => false
  end

(* theory -> op2 -> nut -> nut -> bool *)
fun has_subterm_property thy oper bound_u body_u =
  exists (is_subterm_equation thy oper bound_u) (phrases oper body_u)

(* scope -> bool -> rep NameTable.table -> bool -> nut -> nut *)
fun choose_reps_in_nut (scope as {ctxt, card_assigns, datatypes, ofs}) liberal
                       table def =
  let
    val thy = ProofContext.theory_of ctxt
    val bool_atom_R = Atom (2, offset_of_type ofs bool_T)

    (* occurrence -> bool -> rep *)
    fun bool_rep occ opt =
      if occ = Neutral andalso opt then Opt bool_atom_R else Formula occ
    (* nut -> nut -> nut *)
    fun triad u1 u2 = Op2 (Triad, type_of u1, Opt bool_atom_R, u1, u2)
    (* (occurrence -> nut) -> nut *)
    fun triad_fn f = triad (f Positive) (f Negative)

    (* rep NameTable.table -> bool -> occurrence -> nut -> nut *)
    fun aux table def occ u =
      let
        (* bool -> occurrence -> nut -> nut *)
        val gsub = aux table
        (* nut -> nut *)
        val sub = gsub false Neutral
      in
        (case u of
           Cst (False, T, _) => Cst (False, T, Formula Neutral)
         | Cst (True, T, _) => Cst (True, T, Formula Neutral)
         | Cst (Num j, T, _) =>
           (case spec_of_type scope T of
              (1, j0) => if j = 0 then Cst (Unity, T, Unit)
                         else Cst (Biggie, T, Opt (Atom (1, j0)))
            | (k, j0) =>
              let
                val ok = (if T = int_T then atom_for_int (k, j0) j <> ~1
                          else j < k)
              in
                if ok then Cst (Num j, T, Atom (k, j0))
                else Cst (Biggie, T, Opt (Atom (k, j0)))
              end)
         | Cst (Biggie, T, _) =>
           Cst (Biggie, T, best_opt_set_rep_for_type scope T)
         | Cst (Suc, T as Type ("fun", [T1, _]), _) =>
           let val R = Atom (spec_of_type scope T1) in
             Cst (Suc, T, Func (R, Opt R))
           end
         | Cst (Fracs, T, _) =>
           Cst (Fracs, T, best_non_opt_set_rep_for_type scope T)
         | Cst (NormFrac, T, _) =>
           let val R1 = Atom (spec_of_type scope (domain_type T)) in
             Cst (NormFrac, T, Func (R1, Func (R1, Opt (Struct [R1, R1]))))
           end
         | Cst (cst, T, _) =>
           if cst mem [Add, Subtract, Multiply, Divide, Modulo, Gcd, Lcm] then
             let
               val T1 = domain_type T
               val R1 = Atom (spec_of_type scope T1)
               val total =
                 T1 = nat_T andalso cst mem [Subtract, Divide, Modulo, Gcd]
             in Cst (cst, T, Func (R1, Func (R1, (not total ? Opt) R1))) end
           else if cst mem [NatToInt, IntToNat] then
             let
               val (nat_card, nat_j0) = spec_of_type scope nat_T
               val (int_card, int_j0) = spec_of_type scope int_T
             in
               if cst = NatToInt then
                 let val total = (max_int_for_card int_card >= nat_card + 1) in
                   Cst (cst, T,
                        Func (Atom (nat_card, nat_j0),
                              (not total ? Opt) (Atom (int_card, int_j0))))
                 end
               else
                 let val total = (max_int_for_card int_card < nat_card) in
                   Cst (cst, T, Func (Atom (int_card, int_j0),
                        (not total ? Opt) (Atom (nat_card, nat_j0))))
                 end
             end
           else
             Cst (cst, T, best_set_rep_for_type scope T)
         | Op1 (Not, T, _, u1) =>
           (case gsub def (flip_occurrence occ) u1 of
              Op2 (Triad, T, R, u11, u12) =>
              triad (Op1 (Not, T, Formula Positive, u12))
                    (Op1 (Not, T, Formula Negative, u11))
            | u1' =>
              if is_opt_rep (rep_of u1') then
                triad_fn (fn occ => gsub def occ (Op1 (Not, T, Any, u1)))
              else
                Op1 (Not, T, flip_rep_occurrence (rep_of u1'), u1'))
         | Op1 (oper, T, _, u1) =>
           let
             val u1 = sub u1
             val R1 = rep_of u1
             val R = case oper of
                       Finite => bool_rep occ (is_opt_rep R1)
                     | _ => (if is_opt_rep R1 then best_opt_set_rep_for_type
                             else best_non_opt_set_rep_for_type) scope T
           in Op1 (oper, T, R, u1) end
         | Op2 (Less, T, _, u1, u2) =>
           let
             val u1 = sub u1
             val u2 = sub u2
             val R = bool_rep occ (exists (is_opt_rep o rep_of) [u1, u2])
           in Op2 (Less, T, R, u1, u2) end
         | Op2 (DefEq, T, _, u1, u2) =>
           Op2 (DefEq, T, Formula Neutral, sub u1, sub u2)
         | Op2 (Eq, T, _, u1, u2) =>
           let
             val u1' = sub u1
             val u2' = sub u2

             (* unit -> nut *)
             fun non_opt_case () = Op2 (Eq, T, Formula occ, u1', u2')
             (* unit -> nut *)
             fun opt_opt_case () =
               if occ = Neutral then
                 triad_fn (fn occ => Op2 (Eq, T, Formula occ, u1', u2'))
               else
                 non_opt_case ()
             (* nut -> nut *)
             fun hybrid_case u =
               if is_constructive thy u then
                 (* hackish optimization *)
                 Op2 (Eq, T, Formula Positive, u1', u2')
               else
                 opt_opt_case ()
           in
             case (is_opt_rep (rep_of u1'), is_opt_rep (rep_of u2')) of
               (true, true) => opt_opt_case ()
             | (true, false) => hybrid_case u1'
             | (false, true) => hybrid_case u2'
             | (false, false) => non_opt_case ()
           end
         | Op2 (Subset, T, _, u1, u2) =>
           let
             val u1 = sub u1
             val u2 = sub u2
             val R = bool_rep occ (exists (is_opt_rep o rep_of) [u1, u2])
           in
             if is_opt_rep R then
               triad_fn (fn occ => Op2 (Subset, T, Formula occ, u1, u2))
             else
               Op2 (Subset, T, R, u1, u2)
           end
         | Op2 (Image, T, _, u1, u2) =>
           let
             val u1' = sub u1
             val u2' = sub u2
           in
             (case (rep_of u1', rep_of u2') of
                (Func (R11, R12), Func (R21, Formula Neutral)) =>
                if R21 = R11 andalso is_lone_rep R12 then
                  let
                    val R =
                      best_non_opt_set_rep_for_type scope T
                      |> exists (is_opt_rep o rep_of) [u1', u2'] ? opt_rep ofs T
                  in Op2 (Image, T, R, u1', u2') end
                else
                  raise SAME
              | _ => raise SAME)
             handle SAME =>
                    let
                      val T1 = type_of u1
                      val dom_T = domain_type T1
                      val ran_T = range_type T1
                      val x_u = BoundName (~1, dom_T, Any, "image.x")
                      val y_u = BoundName (~2, ran_T, Any, "image.y")
                    in
                      Op2 (Abstraction, T, Any, y_u,
                           Op2 (Exist, bool_T, Any, x_u,
                                Op2 (And, bool_T, Any,
                                     Op2 (Subset, bool_T, Any,
                                          Op1 (SingletonSet, dom_T --> bool_T,
                                               Any, x_u), u2),
                                     Op2 (Eq, bool_T, Any, y_u,
                                          Op2 (Apply, ran_T, Any, u1, x_u)))))
                      |> sub
                    end
           end
         | Op2 (Apply, T, _, u1, u2) =>
           let
             val u1 = sub u1
             val u2 = sub u2
             val T1 = type_of u1
             val R1 = rep_of u1
             val opt = exists (is_opt_rep o rep_of) [u1, u2]
             val ran_R =
               if is_boolean_type T then
                 bool_rep occ opt
               else
                 (opt ? opt_rep ofs T)
                     (range_rep ofs T1
                                (fn () => card_of_type card_assigns T) R1)
           in Op2 (Apply, T, ran_R, u1, u2) |> optimize_biggie thy end
         | Op2 (Abstraction, T, _, u1, u2) =>
           (case best_set_rep_for_type scope T of
              Unit => Cst (Unity, T, Unit)
            | R as Func (R1, _) =>
              let
                val table' = NameTable.update (u1, R1) table
                val u1 = aux table' false Neutral u1
                val u2 = aux table' false Neutral u2
                val R =
                  R |> (if is_opt_rep (rep_of u2)
                           orelse (range_type T = bool_T
                                   andalso not (is_Cst False u2)) then
                          opt_rep ofs T
                        else
                          unopt_rep)
              in Op2 (Abstraction, T, R, u1, u2) end
            | R => raise NUT ("choose_reps_in_nut", [u]))
         | Op2 (oper, T, _, u1, u2) =>
           if oper mem [All, Exist] then
             let
               val table' = fold (choose_rep_for_bound_var scope) (untuple I u1)
                                 table
               val u1' = aux table' def occ u1
               val u2' = aux table' def occ u2
             in
               if occ = Neutral andalso is_opt_rep (rep_of u2') then
                 triad_fn (fn occ => gsub def occ (Op2 (oper, T, Any, u1, u2)))
               else
                 let val quant_u = Op2 (oper, T, Formula occ, u1', u2') in
                   if def
                      orelse (liberal andalso (occ = Positive) = (oper = All))
                      orelse is_precise_type datatypes (type_of u1)
                      orelse has_subterm_property thy oper u1' u2' then
                     quant_u
                   else
                     let
                       val connective = if oper = All then And else Or
                       val R1'' = opt_rep ofs (type_of u1') (rep_of u1')
                       val table'' = NameTable.update (u1, R1'') table
                       val u1'' = aux table'' def occ u1
                       val u2'' = aux table'' def occ u2
                       val biggied_u =
                         substitute_in_nut thy u1''
                             (Cst (Biggie, type_of u1'', R1'')) u2''
                     in
                       Op2 (connective, T,
                            min_rep (rep_of quant_u) (rep_of biggied_u),
                            quant_u, biggied_u)
                     end
                 end
             end
           else if oper mem [Or, Implies, And] then
             let
               val u1' = gsub def ((oper = Implies ? flip_occurrence) occ) u1
               val u2' = gsub def occ u2
             in
               (if occ = Neutral then
                  case (is_opt_rep (rep_of u1'), is_opt_rep (rep_of u2')) of
                    (true, true) => triad_fn (fn occ => gsub def occ u)
                  | (true, false) =>
                    Op2 (oper, T, Opt bool_atom_R,
                         triad_fn (fn occ => gsub def occ u1), u2')
                  | (false, true) =>
                    Op2 (oper, T, Opt bool_atom_R, u1',
                         triad_fn (fn occ => gsub def occ u2))
                  | (false, false) => raise SAME
                else
                  raise SAME)
               handle SAME => Op2 (oper, T, Formula occ, u1', u2')
             end
           else
             let
               val u1 = sub u1
               val u2 = sub u2
               val R =
                 best_non_opt_set_rep_for_type scope T
                 |> exists (is_opt_rep o rep_of) [u1, u2] ? opt_rep ofs T
             in Op2 (oper, T, R, u1, u2) end
         | Op3 (Let, T, _, u1, u2, u3) =>
           let
             val u2 = sub u2
             val R2 = rep_of u2
             val table' = NameTable.update (u1, R2) table
             val u1 = change_name_rep u1 R2
             val u3 = aux table' false occ u3
           in optimized_let thy u1 u2 u3 end
         | Op3 (If, T, _, u1, u2, u3) =>
           let
             val u1 = sub u1
             val u2 = gsub def occ u2
             val u3 = gsub def occ u3
             val min_R = min_rep (rep_of u2) (rep_of u3)
             val R = min_R |> is_opt_rep (rep_of u1) ? opt_rep ofs T
           in Op3 (If, T, R, u1, u2, u3) end
         | Tuple (T, _, us) =>
           let
             val Rs = map (best_one_rep_for_type scope o type_of) us
             val us = map sub us
             val R = if forall (equal Unit) Rs then Unit else Struct Rs
             val R' = (exists (is_opt_rep o rep_of) us ? opt_rep ofs T) R
           in Tuple (T, R', us) end
         | _ =>
           let val u = change_name_rep u (lookup_name table u) in
             if occ = Neutral orelse not (is_boolean_type (type_of u))
                orelse not (is_opt_rep (rep_of u)) then
               u
             else
               Op1 (Cast, type_of u, Formula occ, u)
           end)
        |> optimize_unit
      end
  in aux table def Positive end

(* Kodkod output *)

(* int -> Kodkod.n_ary_index list -> Kodkod.n_ary_index list
   -> int * Kodkod.n_ary_index list *)
fun fresh_n_ary_index n [] ys = (0, (n, 1) :: ys)
  | fresh_n_ary_index n ((m, j) :: xs) ys =
    if m = n then (j, ys @ ((m, j + 1) :: xs))
    else fresh_n_ary_index n xs ((m, j) :: ys)
(* int -> name_pool -> int * name_pool *)
fun fresh_rel n {rels, vars, formula_reg, rel_reg} =
  let val (j, rels') = fresh_n_ary_index n rels [] in
    (j, {rels = rels', vars = vars, formula_reg = formula_reg,
         rel_reg = rel_reg})
  end
(* int -> name_pool -> int * name_pool *)
fun fresh_var n {rels, vars, formula_reg, rel_reg} =
  let val (j, vars') = fresh_n_ary_index n vars [] in
    (j, {rels = rels, vars = vars', formula_reg = formula_reg,
         rel_reg = rel_reg})
  end
(* int -> name_pool -> int * name_pool *)
fun fresh_formula_reg {rels, vars, formula_reg, rel_reg} =
  (formula_reg, {rels = rels, vars = vars, formula_reg = formula_reg + 1,
                 rel_reg = rel_reg})
(* int -> name_pool -> int * name_pool *)
fun fresh_rel_reg {rels, vars, formula_reg, rel_reg} =
  (rel_reg, {rels = rels, vars = vars, formula_reg = formula_reg,
             rel_reg = rel_reg + 1})

(* nut -> nut list * name_pool * nut NameTable.table
   -> nut list * name_pool * nut NameTable.table *)
fun rename_plain_var v (ws, pool, table) =
  let
    val is_formula = (rep_of v = Formula Neutral)
    val fresh = if is_formula then fresh_formula_reg else fresh_rel_reg
    val (j, pool) = fresh pool
    val ctor = if is_formula then FormulaReg else RelReg
    val w = ctor (j, type_of v, rep_of v)
  in (w :: ws, pool, NameTable.update (v, w) table) end

(* typ -> rep -> nut list -> nut *)
fun shape_tuple (T as Type ("*", [T1, T2])) (R as Struct [R1, R2]) us =
    let val arity1 = arity_of_rep R1 in
      Tuple (T, R, [shape_tuple T1 R1 (List.take (us, arity1)),
                    shape_tuple T2 R2 (List.drop (us, arity1))])
    end
  | shape_tuple (T as Type ("fun", [_, T2])) (R as Vect (k, R')) us =
    Tuple (T, R, map (shape_tuple T2 R') (batch_list (length us div k) us))
  | shape_tuple T R [u] =
    if type_of u = T then u else raise NUT ("shape_tuple", [u])
  | shape_tuple T Unit [] = Cst (Unity, T, Unit)
  | shape_tuple _ _ us = raise NUT ("shape_tuple", us)

(* bool -> nut -> nut list * name_pool * nut NameTable.table
   -> nut list * name_pool * nut NameTable.table *)
fun rename_n_ary_var rename_free v (ws, pool, table) =
  let
    val T = type_of v
    val R = rep_of v
    val arity = arity_of_rep R
    val nick = nickname_of v
    val (ctor, fresh) = if rename_free then (FreeRel, fresh_rel)
                                       else (BoundRel, fresh_var)
  in
    if not (rename_free) andalso arity > 1 then
      let
        val atom_schema = atom_schema_of_rep R
        val type_schema = type_schema_of_rep T R
        val (js, pool) = funpow arity (fn (js, pool) =>
                                          let val (j, pool) = fresh 1 pool in
                                            (j :: js, pool)
                                          end)
                                ([], pool)
        val ws' = map3 (fn j => fn x => fn T =>
                           ctor ((1, j), T, Atom x,
                                 nick ^ " [" ^ string_for_int j ^ "]"))
                       (rev js) atom_schema type_schema
      in (ws' @ ws, pool, NameTable.update (v, shape_tuple T R ws') table) end
    else
      let
        val (j, pool) = fresh arity pool
        val w = ctor ((arity, j), T, R, nick)
      in (w :: ws, pool, NameTable.update (v, w) table) end
  end

(* nut list -> name_pool -> nut NameTable.table
  -> nut list * name_pool * nut NameTable.table *)
fun rename_free_vars vs pool table =
  let
    val vs = filter (not_equal Unit o rep_of) vs
    val (vs, pool, table) = fold (rename_n_ary_var true) vs ([], pool, table)
  in (rev vs, pool, table) end

(* nut * nut -> nut NameTable.table -> nut NameTable.table *)
fun rename_rep_fun (abs_rel, rep_name) =
  NameTable.update (rep_name,
                    Op1 (RepOfAbs, type_of rep_name, rep_of rep_name, abs_rel))
(* (nut * nut) list -> nut NameTable.table -> nut NameTable.table *)
val rename_rep_funs = fold rename_rep_fun

(* name_pool -> nut NameTable.table -> nut -> nut *)
fun rename_vars_in_nut pool table u =
  case u of
    Cst _ => u
  | Op1 (oper, T, R, u1) => Op1 (oper, T, R, rename_vars_in_nut pool table u1)
  | Op2 (oper, T, R, u1, u2) =>
    if oper mem [All, Exist, Abstraction] then
      let
        val (_, pool, table) = fold (rename_n_ary_var false) (untuple I u1)
                                    ([], pool, table)
      in
        Op2 (oper, T, R, rename_vars_in_nut pool table u1,
             rename_vars_in_nut pool table u2)
      end
    else
      Op2 (oper, T, R, rename_vars_in_nut pool table u1,
           rename_vars_in_nut pool table u2)
  | Op3 (Let, T, R, u1, u2, u3) =>
    if rep_of u2 = Unit orelse inline_nut u2 then
      let
        val u2 = rename_vars_in_nut pool table u2
        val table = NameTable.update (u1, u2) table
      in rename_vars_in_nut pool table u3 end
    else
      let
        val bs = untuple I u1
        val (_, pool, table') = fold rename_plain_var bs ([], pool, table)
        val u11 = rename_vars_in_nut pool table' u1
      in
        Op3 (Let, T, R, rename_vars_in_nut pool table' u1,
             rename_vars_in_nut pool table u2,
             rename_vars_in_nut pool table' u3)
      end
  | Op3 (oper, T, R, u1, u2, u3) =>
    Op3 (oper, T, R, rename_vars_in_nut pool table u1,
         rename_vars_in_nut pool table u2, rename_vars_in_nut pool table u3)
  | Tuple (T, R, us) => Tuple (T, R, map (rename_vars_in_nut pool table) us)
  | _ => lookup_name table u

end;
