(*  Title:      Nitpick/Tools/nitpick_scope.ML
    Author:     Jasmin Blanchette, TU Muenchen
    Copyright   2008, 2009

Scope enumerator for Nitpick.
*)

signature NITPICK_SCOPE =
sig
  type ctor_spec = {
    name: string,
    typ: typ,
    delta: int,
    epsilon: int,
    exclusive: bool,
    explicit_max: int,
    total: bool,
    self_rec: bool}

  type dtype_spec = {
    typ: typ,
    card: int,
    precise: bool,
    ctors: ctor_spec list}

  type scope = {
    ctxt: Proof.context,
    card_assigns: (typ * int) list,
    datatypes: dtype_spec list,
    ofs: int Typtab.table}

  val datatype_spec : dtype_spec list -> typ -> dtype_spec option
  val ctor_spec : dtype_spec list -> styp -> ctor_spec
  val card_of_type : (typ * int) list -> typ -> int
  val maxed_card_of_type : int -> (typ * int) list -> typ -> int
  val is_precise_type : dtype_spec list -> typ -> bool
  val offset_of_type : int Typtab.table -> typ -> int
  val spec_of_type : scope -> typ -> int * int
  val pretties_for_scope : scope -> bool -> Pretty.T list
  val multiline_string_for_scope : scope -> string
  val scopes_equivalent : scope -> scope -> bool
  val all_scopes :
    Proof.context -> int -> (typ option * int list) list
    -> (styp option * int list) list -> (styp option * int list) list
    -> typ list -> typ list -> scope list
end;

structure NitpickScope : NITPICK_SCOPE =
struct

open NitpickUtil
open NitpickHOL

type ctor_spec = {
  name: string,
  typ: typ,
  delta: int,
  epsilon: int,
  exclusive: bool,
  explicit_max: int,
  total: bool,
  self_rec: bool}

type dtype_spec = {
  typ: typ,
  card: int,
  precise: bool,
  ctors: ctor_spec list}

type scope = {
  ctxt: Proof.context,
  card_assigns: (typ * int) list,
  datatypes: dtype_spec list,
  ofs: int Typtab.table}

datatype entry = Card of typ | Max of styp

type row = entry * int list
type block = row list

(* dtype_spec list -> typ -> dtype_spec option *)
fun datatype_spec dtypes T = List.find (equal T o #typ) dtypes

(* dtype_spec list -> styp -> ctor_spec *)
fun ctor_spec [] z = raise TERM ("ctor_spec", [Const z])
  | ctor_spec ({ctors, ...} :: dtypes) z =
    (case List.find (equal z o (#name pairf #typ)) ctors of
       SOME c => c
     | NONE => ctor_spec dtypes z)

(* (typ * int) list -> typ -> int *)
fun card_of_type asgns (Type ("fun", [T1, T2])) =
    power (card_of_type asgns T2) (card_of_type asgns T1)
  | card_of_type asgns (Type ("*", [T1, T2])) =
    card_of_type asgns T1 * card_of_type asgns T2
  | card_of_type _ (Type (@{type_name itself}, _)) = 2
  | card_of_type _ @{typ prop} = 2
  | card_of_type _ @{typ bool} = 2
  | card_of_type _ @{typ unit} = 1
  | card_of_type asgns @{typ ind} = card_of_type asgns nat_T
  | card_of_type asgns T =
    case AList.lookup (op =) asgns T of
      SOME k => k
    | NONE => if is_integer_type T then ~1
              else raise TYPE ("card_of_type", [T], [])

(* int -> (typ * int) list -> typ -> int *)
fun maxed_card_of_type max asgns (Type ("fun", [T1, T2])) =
    let
      val k1 = maxed_card_of_type max asgns T1
      val k2 = maxed_card_of_type max asgns T2
    in if k1 = max orelse k2 = max then max else Int.min (max, power k2 k1) end
  | maxed_card_of_type max asgns (Type ("*", [T1, T2])) =
    let
      val k1 = maxed_card_of_type max asgns T1
      val k2 = maxed_card_of_type max asgns T2
    in if k1 = max orelse k2 = max then max else Int.min (max, k1 * k2) end
  | maxed_card_of_type max asgns T = Int.min (max, card_of_type asgns T)

(* dtype_spec list -> typ -> bool *)
fun is_precise_type dtypes (Type ("fun", Ts)) =
    forall (is_precise_type dtypes) Ts
  | is_precise_type dtypes (Type ("*", Ts)) = forall (is_precise_type dtypes) Ts
  | is_precise_type dtypes T =
    not (is_integer_type T) andalso #precise (the (datatype_spec dtypes T))
    handle Option => true

(* int Typtab.table -> typ -> int *)
fun offset_of_type ofs T =
  case Typtab.lookup ofs T of
    SOME j0 => j0
  | NONE => case Typtab.lookup ofs dummyT of
              SOME j0 => j0
            | NONE => 0

(* scope -> typ -> int * int *)
fun spec_of_type {card_assigns, ofs, ...} T =
  (card_of_type card_assigns T, offset_of_type ofs T)

(* (string -> string) -> scope
   -> string list * string list * string list * string list *)
fun quadruple_for_scope quote {ctxt, card_assigns, datatypes, ...} =
  let
    val thy = ProofContext.theory_of ctxt
    val (iter_asgns, card_asgns) =
      List.partition (is_iterator_type o fst) card_assigns
    val (unimportant_card_asgns, important_card_asgns) =
      List.partition ((is_datatype_like thy orf is_integer_type) o fst)
                     card_asgns
    val cards =
      map (fn (T, k) =>
              quote (Syntax.string_of_typ ctxt T) ^ " = " ^ string_for_int k)
    val maxes =
      maps (map_filter
                (fn {name, typ, explicit_max, ...} =>
                    if explicit_max < 0 then
                      NONE
                    else
                      SOME (Syntax.string_of_term ctxt (Const (name, typ)) ^
                            " = " ^ string_for_int explicit_max))
                 o #ctors)
    val iters =
      map (fn (T, k) =>
              quote (Syntax.string_of_term ctxt
                         (Const (const_for_iterator_type T))) ^ " = " ^
              string_for_int (k - 1))
  in
    setmp_show_all_types
        (fn () => (cards important_card_asgns, cards unimportant_card_asgns,
                   maxes datatypes, iters iter_asgns)) ()
  end

(* scope -> bool -> Pretty.T list *)
fun pretties_for_scope scope verbose =
  let
    val (important_cards, unimportant_cards, maxes, iters) =
      quadruple_for_scope maybe_quote scope
    val ss = map (prefix "card ") important_cards @
             (if verbose then
                map (prefix "card ") unimportant_cards @
                map (prefix "max ") maxes @
                map (prefix "iter ") iters
              else
                [])
  in
    if null ss then []
    else serial_commas "and" ss |> map Pretty.str |> Pretty.breaks
  end

(* scope -> string *)
fun multiline_string_for_scope scope =
  let
    val (important_cards, unimportant_cards, maxes, iters) =
      quadruple_for_scope I scope
    val cards = important_cards @ unimportant_cards
  in
    case (if null cards then [] else ["card: " ^ commas cards]) @
         (if null maxes then [] else ["max: " ^ commas maxes]) @
         (if null iters then [] else ["iter: " ^ commas iters]) of
      [] => "empty"
    | lines => space_implode "\n" lines
  end

(* scope -> scope -> bool *)
fun scopes_equivalent s1 s2 =
  #datatypes s1 = #datatypes s2 andalso #card_assigns s1 = #card_assigns s2

(* row -> int *)
fun rank_of_row (_, ks) = length ks
(* block -> int *)
fun rank_of_block block = foldl Int.max 1 (map rank_of_row block)
(* int -> typ * int list -> typ * int list *)
fun project_row column (y, ks) = (y, [nth ks (Int.min (column, length ks - 1))])
(* int -> block -> block *)
fun project_block (column, block) = map (project_row column) block

(* theory -> styp * styp -> bool *)
fun const_match thy ((s1, T1), (s2, T2)) =
  s1 = s2 andalso (Sign.typ_match thy (T2, T1) Vartab.empty; true)
  handle Type.TYPE_MATCH => false

(* (''a * ''a -> bool) -> (''a option * int list) list -> ''a -> int list *)
fun lookup_ints_assign eq asgns key =
  case AList.lookup (fn (SOME x, SOME y) => eq (x, y) | _ => false) asgns
                    (SOME key) of
    SOME ks => ks
  | NONE => case AList.lookup (op =) asgns NONE of
              SOME ks => ks
            | NONE => raise ARG ("lookup_ints_assign", "")
(* (typ option * int list) list -> typ -> int list *)
fun lookup_type_ints_assign asgns T =
  map (curry Int.max 1) (lookup_ints_assign (op =) asgns T)
  handle ARG ("lookup_ints_assign", _) =>
         raise TYPE ("lookup_type_ints_assign", [T], [])
(* theory -> (styp option * int list) list -> styp -> int list *)
fun lookup_const_ints_assign thy asgns x =
  lookup_ints_assign (op =) (filter (isSome o fst) asgns) x
  handle ARG ("lookup_ints_assign", _) =>
         lookup_ints_assign (const_match thy) asgns x
         handle ARG ("lookup_ints_assign", _) =>
                raise TERM ("lookup_const_ints_assign", [Const x])

(* theory -> (styp option * int list) list -> styp -> row option *)
fun row_for_ctor thy maxes_asgns ctor =
  SOME (Max ctor, lookup_const_ints_assign thy maxes_asgns ctor)
  handle TERM ("lookup_const_ints_assign", _) => NONE

(* Proof.context -> (typ option * int list) list
   -> (styp option * int list) list -> (styp option * int list) list -> typ
   -> block *)
fun block_for_type ctxt cards_asgns maxes_asgns iters_asgns T =
  let val thy = ProofContext.theory_of ctxt in
    if is_iterator_type T then
      [(Card T, map (curry (op +) 1 o curry Int.max 0)
                    (lookup_const_ints_assign thy iters_asgns
                                              (const_for_iterator_type T)))]
    else
      (Card T, lookup_type_ints_assign cards_asgns T) ::
      (case datatype_ctors thy T of
         [_] => []
       | ctors => map_filter (row_for_ctor thy maxes_asgns) ctors)
  end

(* Proof.context -> (typ option * int list) list
   -> (styp option * int list) list -> (styp option * int list) list -> typ list
   -> typ list -> block list *)
fun blocks_for_types ctxt cards_asgns maxes_asgns iters_asgns mono_Ts
                     nonmono_Ts =
  let
    (* typ -> block *)
    val block_for = block_for_type ctxt cards_asgns maxes_asgns iters_asgns
    val mono_block = maps block_for mono_Ts
    val nonmono_blocks = map block_for nonmono_Ts
  in mono_block :: nonmono_blocks end

(* int list -> int list list *)
val all_combinations_ordered_smartly =
  let
    (* int -> int *)
    fun mono_fn x = x * (x + 1) div 2
    (* (int -> int) -> int list -> int *)
    fun cost f [] = 0
      | cost f (k :: ks) = f k + cost I ks
  in
    all_combinations #> map (`(cost mono_fn)) #> sort (int_ord o pairself fst)
    #> map snd
  end

(* typ -> bool *)
fun is_self_recursive_ctor_type T =
  exists (exists_subtype (equal (body_type T))) (binder_types T)

(* (styp * int) list -> styp -> int *)
val ctor_max = the_default ~1 oo AList.lookup (op =)

type scope_desc = (typ * int) list * (styp * int) list

(* theory -> scope_desc -> typ * int -> bool *)
fun is_surely_inconsistent_card_assign thy (card_asgns, max_asgns) (T, k) =
  case datatype_ctors thy T of
    [] => false
  | xs =>
    let
      val cards =
        map (foldl (op *) 1 o map (precise_card_of_type thy 0 card_asgns)
             o binder_types o snd) xs
      val maxes = map (ctor_max max_asgns) xs
      (* int -> int -> int *)
      fun effective_max 0 ~1 = k
        | effective_max 0 max = max
        | effective_max card ~1 = card
        | effective_max card max = Int.min (card, max)
    in (map2 effective_max cards maxes |> foldl1 (op +)) < k end

(* theory -> scope_desc -> bool *)
fun is_surely_inconsistent_scope_description thy (desc as (card_asgns, _)) =
  exists (is_surely_inconsistent_card_assign thy desc) card_asgns

(* theory -> scope_desc -> (typ * int) list option *)
fun repair_card_assigns thy (card_asgns, max_asgns) =
  let
    (* (typ * int) list -> (typ * int) list -> (typ * int) list option *)
    fun aux seen [] = SOME seen
      | aux seen ((T, 0) :: _) = NONE
      | aux seen ((T, k) :: asgns) =
        (if is_surely_inconsistent_scope_description thy
                ((T, k) :: seen, max_asgns) then
           raise SAME
         else
           case aux ((T, k) :: seen) asgns of
             SOME asgns => SOME asgns
           | NONE => raise SAME)
        handle SAME => aux seen ((T, k - 1) :: asgns)
  in aux [] (rev card_asgns) end

(* (typ * int) list -> typ * int -> typ * int *)
fun repair_iterator_assign asgns (T as Type (s, (T' :: _)), k) =
    (T, if is_iterator_type T then maxed_card_of_type k asgns T' else k)
  | repair_iterator_assign _ asgn = asgn

(* row -> scope_desc -> scope_desc *)
fun add_row_to_scope_descriptor (Card T, ks) = apfst (cons (T, the_single ks))
  | add_row_to_scope_descriptor (Max x, ks) = apsnd (cons (x, the_single ks))
(* block -> scope_desc *)
fun scope_descriptor_from_block block =
  fold_rev add_row_to_scope_descriptor block ([], [])
(* theory -> block list -> int list -> scope_desc option *)
fun scope_descriptor_from_combination thy blocks columns =
  let
    val (card_asgns, max_asgns) =
      maps project_block (columns ~~ blocks) |> scope_descriptor_from_block
    val card_asgns = repair_card_assigns thy (card_asgns, max_asgns) |> the
  in SOME (map (repair_iterator_assign card_asgns) card_asgns, max_asgns) end
  handle Option => NONE

(* theory -> (typ * int) list -> dtype_spec list -> int Typtab.table *)
fun offset_table_for_card_assigns thy asgns dtypes =
  let
    (* int -> (int * int) list -> (typ * int) list -> int Typtab.table
       -> int Typtab.table *)
    fun aux next _ [] = Typtab.update_new (dummyT, next)
      | aux next reusable ((T, k) :: asgns) =
        if k = 1 orelse is_integer_type T then
          aux next reusable asgns
        else if length (these (Option.map #ctors (datatype_spec dtypes T)))
                > 1 then
          Typtab.update_new (T, next) #> aux (next + k) reusable asgns
        else
          case AList.lookup (op =) reusable k of
            SOME j0 => Typtab.update_new (T, j0) #> aux next reusable asgns
          | NONE => Typtab.update_new (T, next)
                    #> aux (next + k) ((k, next) :: reusable) asgns
  in aux 0 [] asgns Typtab.empty end

(* (typ * int) list -> typ -> int *)
fun domain_card card_asgns =
  foldl (op *) 1 o map (card_of_type card_asgns) o binder_types

(* scope_desc -> int -> int -> int -> int -> bool * styp -> ctor_spec list
   -> ctor_spec list *)
fun add_ctor_spec (card_asgns, max_asgns) card sum_dom_cards num_self_recs
                  num_non_self_recs (self_rec, x as (s, T)) ctors =
  let
    val max = ctor_max max_asgns x
    (* int -> int *)
    fun bound k = Int.min (card, (max >= 0 ? curry Int.min max) k)
    (* unit -> int *)
    fun next_delta () = if null ctors then 0 else #epsilon (hd ctors)
    val dom_card = domain_card card_asgns T
    val {delta, epsilon, exclusive, total} =
      if max = 0 then
        let val delta = next_delta () in
          {delta = delta, epsilon = delta, exclusive = true, total = false}
        end
      else if num_self_recs > 0 then
        if self_rec then
          {delta = 1, epsilon = card, exclusive = (s = @{const_name Cons}),
           total = false}
        else if num_non_self_recs = 1 then
          let val epsilon = bound dom_card in
            {delta = 0, epsilon = epsilon, exclusive = (epsilon = 1),
             total = (dom_card = 1)}
          end
        else
          {delta = 0, epsilon = card, exclusive = false, total = false}
      else if card = sum_dom_cards then
        let val delta = next_delta () in
          {delta = delta, epsilon = delta + dom_card, exclusive = true,
           total = true}
        end
      else
        {delta = 0, epsilon = card,
         exclusive = (num_self_recs + num_non_self_recs = 1), total = false}
  in
    {name = s, typ = T, delta = delta, epsilon = epsilon, exclusive = exclusive,
     explicit_max = max, total = total, self_rec = self_rec}
    :: ctors
  end

(* theory -> scope_desc -> typ * int -> dtype_spec *)
fun datatype_spec_from_scope_descriptor thy (desc as (card_asgns, _))
                                        (T, card) =
  let
    val xs = datatype_ctors thy T
    val self_recs = map (is_self_recursive_ctor_type o snd) xs
    val (num_self_recs, num_non_self_recs) =
      List.partition (equal true) self_recs |> pairself length
    val precise = (card = precise_card_of_type thy 0 card_asgns T)
    val sum_dom_cards = map (domain_card card_asgns o snd) xs |> foldl1 (op +)
    val ctors = fold (add_ctor_spec desc card sum_dom_cards num_self_recs
                                    num_non_self_recs)
                     (self_recs ~~ xs) []
  in {typ = T, card = card, precise = precise, ctors = ctors} end

(* Proof.context -> int -> scope_desc -> scope *)
fun scope_from_descriptor ctxt sym_break (desc as (card_asgns, _)) =
  let
    val thy = ProofContext.theory_of ctxt
    val datatypes = map (datatype_spec_from_scope_descriptor thy desc)
                        (filter (is_datatype_like thy o fst) card_asgns)
  in
    {ctxt = ctxt, card_assigns = card_asgns, datatypes = datatypes,
     ofs = if sym_break <= 0 then Typtab.empty
           else offset_table_for_card_assigns thy card_asgns datatypes}
  end

(* Proof.context -> int -> (typ option * int list) list
   -> (styp option * int list) list -> (styp option * int list) list
   -> typ list -> bool -> scope list *)
fun all_scopes ctxt sym_break cards_asgns maxes_asgns iters_asgns mono_Ts
               nonmono_Ts =
  let
    val thy = ProofContext.theory_of ctxt
    val blocks = blocks_for_types ctxt cards_asgns maxes_asgns iters_asgns
                                  mono_Ts nonmono_Ts
    val ranks = map rank_of_block blocks
    val descs = all_combinations_ordered_smartly (map (rpair 0) ranks)
                |> map_filter (scope_descriptor_from_combination thy blocks)
  in descs |> distinct (op =) |> map (scope_from_descriptor ctxt sym_break) end

end;
